/* This is BIGsplines.c, the main C code file for the

   NumericPython BIGsplines toolbox

   Jan Kybic, August 2000

   $Id: BIGsplines.c,v 1.2 2002/01/21 16:33:19 jkybic Exp $
*/

#include "Python.h"
#include "BIGsplines.h"
#include "arrayobject.h"

static PyObject *ErrorObject ; /* Exception object (string) */

/* sets error message for an exception to pass to Python */
/* it is used by WRITE_ERROR used in fiir.c */
void BSerrormsg(char *s) { PyErr_SetString(ErrorObject, s) ; }


/* ------ here the glue functions start ------ */

/* IIR filtering */
/* Python usage: y=IirConvolvePoles(input,poles,bcond) */

static PyObject *BSiirConvolvePoles(PyObject *self, PyObject *args)
{
  PyObject *input,*poles ;
  PyArrayObject *inputarray,*polesarray,*outputarray ;
  int bcond ;
  int dimensions[1] ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OOi",&input,&poles,&bcond)) return NULL ;

  /* convert to arrays (actually, only vectors are accepted) */
  if ((inputarray=(PyArrayObject *)PyArray_ContiguousFromObject(input,
       PyArray_DOUBLE,1,1))==NULL) return NULL ;
  if ((polesarray=(PyArrayObject *)PyArray_ContiguousFromObject(poles,
       PyArray_DOUBLE,1,1))==NULL) {
    Py_DECREF(inputarray) ; return NULL ; }

  /* create output array */
  dimensions[0]=inputarray->dimensions[0] ;
  if ((outputarray=(PyArrayObject *)PyArray_FromDims(1,
       dimensions,PyArray_DOUBLE))==NULL) {
    Py_DECREF(inputarray) ; Py_DECREF(polesarray) ; return NULL ; }

  /* do the real work */  
  if (IirConvolvePoles((double *)inputarray->data,(double *)outputarray->data,
                   dimensions[0],(double *)polesarray->data,
		   polesarray->dimensions[0],bcond,1e-20)==ERROR) {
    Py_DECREF(inputarray) ; Py_DECREF(polesarray) ; Py_DECREF(outputarray) ;
    return NULL ; 
  }

  /* decrement reference counters */
  Py_DECREF(inputarray) ;
  Py_DECREF(polesarray) ;
  /* and return */
  return PyArray_Return(outputarray) ;
}

/* FIR filtering */
/* Python usage: y=FirConvolve(input,kernel,kernelorigin,bcond) */

static PyObject *BSfirConvolve(PyObject *self, PyObject *args)
{
  PyObject *input,*kernel ;
  PyArrayObject *inputarray,*kernelarray,*outputarray ;
  int dimensions[1],kernelorigin,bcond ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OOii",&input,&kernel,
			&kernelorigin,&bcond)) return NULL ;

  /* convert to arrays (actually, only vectors are accepted) */
  if ((inputarray=(PyArrayObject *)PyArray_ContiguousFromObject(input,
       PyArray_DOUBLE,1,1))==NULL) return NULL ;
  if ((kernelarray=(PyArrayObject *)PyArray_ContiguousFromObject(kernel,
       PyArray_DOUBLE,1,1))==NULL) {
    Py_DECREF(inputarray) ; return NULL ; }

  /* create output array */
  dimensions[0]=inputarray->dimensions[0] ;
  if ((outputarray=(PyArrayObject *)PyArray_FromDims(1,
       dimensions,PyArray_DOUBLE))==NULL) {
    Py_DECREF(inputarray) ; Py_DECREF(kernelarray) ; return NULL ; }

  /* do the real work */  
  FirConvolve((double *)inputarray->data,(double *)outputarray->data,
                   (long)dimensions[0],(double *)kernelarray->data,
		   kernelorigin,kernelarray->dimensions[0],bcond) ;

  /* decrement reference counters */
  Py_DECREF(inputarray) ;
  Py_DECREF(kernelarray) ;
  /* and return */
  return PyArray_Return(outputarray) ;
}



/* ----------------------------------------------------------------- */
/* Spline interpolation */
/* Python usage: y=SplineInterpol(coefs,coords,degree,flag,bcond)    */
/*                                                                   */
/* R^n->R^1 function is interpolated by spline of degree degree      */
/* with Bspline coeficients coefs. It is evaluated at points coords  */
/* coefs is always n-dimensional, coords n+1 dimensional, the last   */
/* (fastest changing) dimension contains the coordinates             */
/* where degree 0=constant, 1=linear ...                             */
/*       dflag 1,4,16 - first derivative with respect to x,y,z       */ 
/*             2,8,32 - second der. combinations possible            */

#define BSSIMAXDIM (3) /* maximum number of dimensions */

static PyObject *BSsplineInterpol(PyObject *self, PyObject *args)
{
  PyObject *coefs,*coords ;
  PyArrayObject *coefsarray,*coordsarray,*outputarray ;
  int degree,flag ;
  int nxi,nyi,nzi,nxc,nyc,nzc ;
  int ncoord,ncoef,bcond ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OOiii",&coefs,&coords,
			&degree,&flag,&bcond)) return NULL ;

  /* convert to arrays (actually, only vectors are accepted) */
  if ((coefsarray=(PyArrayObject *)PyArray_ContiguousFromObject(coefs,
       PyArray_DOUBLE,1,3))==NULL) return NULL ;
  if ((coordsarray=(PyArrayObject *)PyArray_ContiguousFromObject(coords,
       PyArray_DOUBLE,2,4))==NULL) {
    Py_DECREF(coefsarray) ; return NULL ; }
  /* The first dimension of coords must be equal to the number of dimensions
     in coefs */
  ncoord=coordsarray->nd ; ncoef=coefsarray->nd ;
  if (coordsarray->dimensions[ncoord-1]!=ncoef) {
    BSerrormsg("last dim of coords must be equal to the no of dim of coefs") ;
  err2:
    Py_DECREF(coefsarray) ; Py_DECREF(coordsarray) ; return NULL ; }
  if (coordsarray->dimensions[ncoord-1]>BSSIMAXDIM) {
    BSerrormsg("last dim of coords is too big") ;
    goto err2 ; }

  /* create output array */
  if ((outputarray=(PyArrayObject *)PyArray_FromDims(coordsarray->nd-1,
       coordsarray->dimensions,PyArray_DOUBLE))==NULL) goto err2 ; 

  nxi=coefsarray->dimensions[ncoef-1] ;
  nyi=ncoef>1 ? coefsarray->dimensions[ncoef-2] : 1 ;
  nzi=ncoef>2 ? coefsarray->dimensions[ncoef-3] : 1 ;
  nxc=coordsarray->dimensions[ncoord-2] ;
  nyc=ncoord>2 ? coordsarray->dimensions[ncoord-3] : 1 ;
  nzc=ncoord>3 ? coordsarray->dimensions[ncoord-4] : 1 ;

  /* do the real work */  
  if (splinterp((double *)coefsarray->data,(double *)coordsarray->data,
                (double *)outputarray->data,nxi,nyi,nzi,nxc,nyc,nzc,
		coefsarray->nd,degree,flag,bcond)==ERROR) {
    Py_DECREF(coefsarray) ; Py_DECREF(coordsarray) ; Py_DECREF(outputarray) ;
    return NULL ; 
  }
  /* decrement reference counters */
  Py_DECREF(coefsarray) ;
  Py_DECREF(coordsarray) ;
  /* and return */
  return PyArray_Return(outputarray) ;
}

/* ----------------------------------------------------------------- */
/* Spline interpolation with mask */
/* Python usage: y=SplineInterpolMask(coefs,coords,degree,flag,mask,bcond) */
/*                                                                   */
/* R^n->R^1 function is interpolated by spline of degree degree      */
/* with Bspline coeficients coefs. It is evaluated at points coords  */
/* coefs is always n-dimensional, coords n+1 dimensional, the last   */
/* (fastest changing) dimension contains the coordinates             */
/* where degree 0=constant, 1=linear ...                             */
/*       dflag 1,4,16 - first derivative with respect to x,y,z       */ 
/*             2,8,32 - second der. combinations possible            */


static PyObject *BSsplineInterpolMask(PyObject *self, PyObject *args)
{
  PyObject *coefs,*coords,*mask ;
  PyArrayObject *coefsarray,*coordsarray,*outputarray,*maskarray ;
  int degree,flag ;
  int nxi,nyi,nzi,nxc,nyc,nzc ;
  int ncoord,ncoef,bcond ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OOiiOi",&coefs,&coords,&degree,
			&flag,&mask,&bcond)) 
    return NULL ;

  /* convert to arrays (actually, only vectors are accepted) */
  if ((coefsarray=(PyArrayObject *)PyArray_ContiguousFromObject(coefs,
       PyArray_DOUBLE,1,3))==NULL) return NULL ;
  if ((coordsarray=(PyArrayObject *)PyArray_ContiguousFromObject(coords,
       PyArray_DOUBLE,2,4))==NULL) {
    Py_DECREF(coefsarray) ; return NULL ; }
  if ((maskarray=(PyArrayObject *)PyArray_ContiguousFromObject(mask,
       PyArray_UBYTE,1,3))==NULL) {
    Py_DECREF(coefsarray) ; Py_DECREF(coordsarray) ; return NULL ; }

  /* The first dimension of coords must be equal to the number of dimensions
     in coefs */
  ncoord=coordsarray->nd ; ncoef=coefsarray->nd ;
  if (coordsarray->dimensions[ncoord-1]!=ncoef) {
    BSerrormsg("last dim of coords must be equal to the no of dim of coefs") ;
  err2:
    Py_DECREF(coefsarray) ; Py_DECREF(coordsarray) ; return NULL ; }
  if (coordsarray->dimensions[ncoord-1]>BSSIMAXDIM) {
    BSerrormsg("last dim of coords is too big") ;
    goto err2 ; }

  /* create output array */
  if ((outputarray=(PyArrayObject *)PyArray_FromDims(coordsarray->nd-1,
       coordsarray->dimensions,PyArray_DOUBLE))==NULL) goto err2 ; 

  nxi=coefsarray->dimensions[ncoef-1] ;
  nyi=ncoef>1 ? coefsarray->dimensions[ncoef-2] : 1 ;
  nzi=ncoef>2 ? coefsarray->dimensions[ncoef-3] : 1 ;
  nxc=coordsarray->dimensions[ncoord-2] ;
  nyc=ncoord>2 ? coordsarray->dimensions[ncoord-3] : 1 ;
  nzc=ncoord>3 ? coordsarray->dimensions[ncoord-4] : 1 ;

  /* do the real work */  
  if (splinterpmask((double *)coefsarray->data,(double *)coordsarray->data,
                (double *)outputarray->data,nxi,nyi,nzi,nxc,nyc,nzc,
		coefsarray->nd,degree,flag,
		(unsigned char *)maskarray->data,bcond)==ERROR) {
    Py_DECREF(coefsarray) ; Py_DECREF(coordsarray) ; Py_DECREF(outputarray) ;
    return NULL ; 
  }

  /* decrement reference counters */
  Py_DECREF(coefsarray) ;
  Py_DECREF(coordsarray) ;Py_DECREF(maskarray) ;
  /* and return */
  return PyArray_Return(outputarray) ;
}


/* ----------------------------------------------------------------- */
/* Spline interpolation with one nonzero B-spline coefficient        */
/* Python usage: y=SplineInterpol(coefind,coefsize,coords,degree,flag,bcond)  */
/*                                                                   */
/* R^n->R^1 function is interpolated by spline of degree degree      */
/* with Bspline coefsize coeficients which are all zero except coefind. */
/* It is evaluated at points coords  */
/* coefind is always n-dimensional, coords n+1 dimensional, the last */
/* (fastest changing) dimension contains the coordinates             */
/* where degree 0=constant, 1=linear ...                             */
/*       dflag 1,4,16 - first derivative with respect to x,y,z       */ 
/*             2,8,32 - second der. combinations possible            */

static PyObject *BSsplineInterpolOne(PyObject *self, PyObject *args)
{
  PyObject *coefind,*coefsize,*coords ;
  PyArrayObject *coefindarray,*coefsizearray,*coordsarray,*outputarray ;
  int degree,flag ;
  int nxi,nyi,nzi,nxc,nyc,nzc ;
  int ncoord,ncoef,bx,by,bz ;
  int bcond ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OOOiii",&coefind,&coefsize,&coords,
			&degree,&flag,&bcond)) 
    return NULL ;

  /* convert to arrays (actually, only vectors are accepted) */
  if ((coefindarray=(PyArrayObject *)PyArray_ContiguousFromObject(coefind,
       PyArray_INT,1,1))==NULL) return NULL ;
  if ((coordsarray=(PyArrayObject *)PyArray_ContiguousFromObject(coords,
       PyArray_DOUBLE,2,4))==NULL) {
    Py_DECREF(coefindarray) ; return NULL ; }
  if ((coefsizearray=(PyArrayObject *)PyArray_ContiguousFromObject(coefsize,
       PyArray_INT,1,1))==NULL) {
    Py_DECREF(coefindarray) ; Py_DECREF(coordsarray) ; return NULL ; }

  /* The first dimension of coords must be equal to the number of dimensions
     in coefs */
  ncoord=coordsarray->nd ; ncoef=coefindarray->dimensions[0] ;
  if (coordsarray->dimensions[ncoord-1]!=ncoef) {
    BSerrormsg("last dim of coords must be equal to the no of dim of coefs") ;
  err2:
    Py_DECREF(coefindarray) ; Py_DECREF(coordsarray) ; 
    Py_DECREF(coefsizearray) ;  return NULL ; }
  if (coordsarray->dimensions[ncoord-1]>BSSIMAXDIM) {
    BSerrormsg("last dim of coords is too big") ;
    goto err2 ; }
  if (coefsizearray->dimensions[0]!=ncoef) {
    BSerrormsg("The size of coefsize must be equal to the no of dim of coefs") ;
    goto err2 ; }

  /* create output array */
  if ((outputarray=(PyArrayObject *)PyArray_FromDims(coordsarray->nd-1,
       coordsarray->dimensions,PyArray_DOUBLE))==NULL) goto err2 ; 

  nxi=((int *)coefsizearray->data)[0] ;
  nyi=ncoef>1 ? ((int *)coefsizearray->data)[1] : 1;
  nzi=ncoef>2 ? ((int *)coefsizearray->data)[2] : 1;
  nxc=coordsarray->dimensions[ncoord-2] ;
  nyc=ncoord>2 ? coordsarray->dimensions[ncoord-3] : 1 ;
  nzc=ncoord>3 ? coordsarray->dimensions[ncoord-4] : 1 ;
  
  bx=((int *)coefindarray->data)[0] ;
  by=ncoef>1 ? ((int *)coefindarray->data)[1] : 0 ;
  bz=ncoef>2 ? ((int *)coefindarray->data)[2] : 0 ;

  /* do the real work */  
  if (splinterpone(bx,by,bz,(double *)coordsarray->data,
                (double *)outputarray->data,nxi,nyi,nzi,nxc,nyc,nzc,
		ncoef,degree,flag,bcond)==ERROR) {
    Py_DECREF(coefindarray) ; Py_DECREF(coefsizearray) ; 
    Py_DECREF(coordsarray) ; Py_DECREF(outputarray) ;
    return NULL ; 
  }

  /* decrement reference counters */
  Py_DECREF(coefsizearray) ; Py_DECREF(coefindarray) ;
  Py_DECREF(coordsarray) ;
  /* and return */
  return PyArray_Return(outputarray) ;
}

/* ----------------------------------------------------------------- */
/* Fast evaluation of the criterion for the registration problem     */
/* and its gradient.                                                 */
/* Python usage: E,gr=Sdcgr(refimg,cx,cy,degc,derw,timgw)                */
/*                                                                   */
/* Only works for 2D. Use Python/Fortran (not Matlab) ordering of images. */

static PyObject *BSsdcgr(PyObject *self, PyObject *args)
{
  PyObject *refimg,*derw,*timgw ;
  PyArrayObject *refimgarr,*derwarr,*timgwarr ;
  int degc,cx,cy ;
  double E ; 
  int dims[3] ;
  PyArrayObject *grarr ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OiiiOO",&refimg,&cx,&cy,&degc,
                        &derw,&timgw)) 
    return NULL ;

  /* convert to arrays  */
  if ((refimgarr=(PyArrayObject *)PyArray_ContiguousFromObject(refimg,
       PyArray_DOUBLE,2,2))==NULL) return NULL ;
  if ((derwarr=(PyArrayObject *)PyArray_ContiguousFromObject(derw,
       PyArray_DOUBLE,3,3))==NULL) {
    Py_DECREF(refimgarr) ; 
    return NULL ; }
  if ((timgwarr=(PyArrayObject *)PyArray_ContiguousFromObject(timgw,
       PyArray_DOUBLE,2,2))==NULL) {
    Py_DECREF(refimgarr) ; 
    Py_DECREF(derwarr) ; return NULL ; }

  /* create output array */
  dims[0]=2 ; dims[1]=cy ; dims[2]=cx ;
  if ((grarr=(PyArrayObject *)PyArray_FromDims(3,dims,PyArray_DOUBLE))==NULL) {
    Py_DECREF(refimgarr) ; 
    Py_DECREF(derwarr) ; 
    Py_DECREF(timgwarr) ; return NULL ; 
  }

  /* do the real work */  
  if (bigsdcgr((double *)refimgarr->data, &E,(double *)grarr->data,degc,
	   refimgarr->dimensions[1],refimgarr->dimensions[0],
	   cx,cy,
	   (double *)derwarr->data,
	   (double *)timgwarr->data)!=0) {
    Py_DECREF(refimgarr) ; 
    Py_DECREF(derwarr) ; Py_DECREF(grarr) ;
    Py_DECREF(timgwarr) ; return NULL ; 
  }

  /* decrement reference counters */
  Py_DECREF(refimgarr) ; 
  Py_DECREF(derwarr) ; 
  Py_DECREF(timgwarr) ;

  /* and return */
  return Py_BuildValue("dO",E,grarr) ;
}

/* ----------------------------------------------------------------- */
/* Fast evaluation of the criterion for the registration problem     */
/* and its gradient, with mask                                       */
/* Python usage: E,gr=SdcgrMask(refimg,cx,cy,degc,derw,timgw,mask)   */
/*                                                                   */
/* Only works for 2D. Use Python/Fortran (not Matlab) ordering of images. */

static PyObject *BSsdcgrMask(PyObject *self, PyObject *args)
{
  PyObject *refimg,*derw,*timgw,*mask ;
  PyArrayObject *refimgarr,*derwarr,*timgwarr,*maskarr ;
  int degc,cx,cy ;
  double E ; 
  int dims[3] ;
  PyArrayObject *grarr ;

  /* parse input */
  if (!PyArg_ParseTuple(args,"OiiiOOO",&refimg,&cx,&cy,&degc,
                        &derw,&timgw,&mask)) 
    return NULL ;

  /* convert to arrays  */
  if ((refimgarr=(PyArrayObject *)PyArray_ContiguousFromObject(refimg,
       PyArray_DOUBLE,2,2))==NULL) return NULL ;
  if ((derwarr=(PyArrayObject *)PyArray_ContiguousFromObject(derw,
       PyArray_DOUBLE,3,3))==NULL) {
    Py_DECREF(refimgarr) ; 
    return NULL ; }
  if ((timgwarr=(PyArrayObject *)PyArray_ContiguousFromObject(timgw,
       PyArray_DOUBLE,2,2))==NULL) {
    Py_DECREF(refimgarr) ; 
    Py_DECREF(derwarr) ; return NULL ; }
  if ((maskarr=(PyArrayObject *)PyArray_ContiguousFromObject(mask,
       PyArray_UBYTE,2,2))==NULL) {
    Py_DECREF(refimgarr) ; Py_DECREF(timgwarr) ;
    Py_DECREF(derwarr) ; return NULL ; }

  /* create output array */
  dims[0]=2 ; dims[1]=cy ; dims[2]=cx ;
  if ((grarr=(PyArrayObject *)PyArray_FromDims(3,dims,PyArray_DOUBLE))==NULL) {
    Py_DECREF(refimgarr) ; 
    Py_DECREF(derwarr) ; Py_DECREF(maskarr) ; 
    Py_DECREF(timgwarr) ; return NULL ; 
  }

  /* do the real work */  
  if (bigsdcgrmask((double *)refimgarr->data, &E,(double *)grarr->data,degc,
	   refimgarr->dimensions[1],refimgarr->dimensions[0],
	   cx,cy,(double *)derwarr->data,
	   (double *)timgwarr->data,(unsigned char *)maskarr->data)!=0) {
    Py_DECREF(refimgarr) ; Py_DECREF(grarr) ;
    Py_DECREF(derwarr) ; Py_DECREF(maskarr) ; 
    Py_DECREF(timgwarr) ; return NULL ; 
  }

  /* decrement reference counters */
  Py_DECREF(refimgarr) ; 
  Py_DECREF(derwarr) ; 
  Py_DECREF(timgwarr) ;

  /* and return */
  return Py_BuildValue("dO",E,grarr) ;
}

/* ----------------------------------------------------------------- */
/* Evaluates B-spline of degree 'degree' at n-points x[0],...,x[n-1] */
/* Python usage: y=BSevalbspln(x,degree)                             */

static PyObject *BSevalbspln(PyObject *self, PyObject *args)
{
  PyObject *x ;
  PyArrayObject *xarr,*yarr ;
  int degree ;

  if (!PyArg_ParseTuple(args,"Oi",&x,&degree)) 
    return NULL ;
  if ((xarr=(PyArrayObject *)PyArray_ContiguousFromObject(x,
       PyArray_DOUBLE,1,1))==NULL) return NULL ;
  if ((yarr=(PyArrayObject *)PyArray_FromDims(xarr->nd,xarr->dimensions,
					     PyArray_DOUBLE))==NULL)
    { err3:
         Py_DECREF(xarr) ; return NULL ; }
  if (evalbspln((double *)xarr->data,(double *)yarr->data,
		xarr->dimensions[0],degree))
    { Py_DECREF(yarr) ; goto err3 ; }

  Py_DECREF(xarr) ;
  return PyArray_Return(yarr) ;
}

/* ConvolveSubsamp performs an equivalent of a convolution */
/* (with reversed kernel!) followed by downsampling. Signal is */
/* assumed to be zero outside the boundaries */
/* Python usage: y=ConvolveSubsamp(x,kernel,yn,ofs,shift) */
/* where ofs is the starting ofset in x, shift is added to the position */
/* for each output point, and yn is the total number of output points */
/* Only works for 1D arrays. */

static PyObject *BSconvolveSubsamp(PyObject *self, PyObject *args)
{
  PyObject *x,*kernel ;
  PyArrayObject *xarr,*yarr,*kernelarr ;
  int yn,ofs,shift ;

  if (!PyArg_ParseTuple(args,"OOiii",&x,&kernel,&yn,&ofs,&shift)) 
    return NULL ;
  if ((xarr=(PyArrayObject *)PyArray_ContiguousFromObject(x,
       PyArray_DOUBLE,1,1))==NULL) return NULL ;
  if ((kernelarr=(PyArrayObject *)PyArray_ContiguousFromObject(kernel,
       PyArray_DOUBLE,1,1))==NULL) 
    { Py_DECREF(xarr) ; return NULL ;}
  if ((yarr=(PyArrayObject *)PyArray_FromDims(1,&yn,
					     PyArray_DOUBLE))==NULL)
    { Py_DECREF(xarr) ; Py_DECREF(kernelarr) ; return NULL ;}

  if (convolve_subsamp((double *)xarr->data,xarr->dimensions[0],
		      (double *)yarr->data,yn,
		      (double *)kernelarr->data,
		       kernelarr->dimensions[0],ofs,shift))
    { Py_DECREF(xarr) ; Py_DECREF(kernelarr) ;
      Py_DECREF(yarr) ; return NULL ;}
  Py_DECREF(xarr) ; Py_DECREF(kernelarr) ;
  return PyArray_Return(yarr) ;
}

/* ConvolveUpsample performs an equivalent of upsampling followed */
/* by convolution (with reversed kernel). The signal is assumed to be zero */
/* outside the boundaries */
/* Python usage: y=ConvolveUpsample(x,kernel,korig,step) */
/* where korig is the center of the kernel and step is the upsampling */
/* factor. Only works for 1D arrays. */

static PyObject *BSconvolveUpsample(PyObject *self, PyObject *args)
{
  PyObject *x,*kernel ;
  PyArrayObject *xarr,*yarr,*kernelarr ;
  int korig,step,yn ;

  if (!PyArg_ParseTuple(args,"OOii",&x,&kernel,&korig,&step)) 
    return NULL ;
  if ((xarr=(PyArrayObject *)PyArray_ContiguousFromObject(x,
       PyArray_DOUBLE,1,1))==NULL) return NULL ;
  if ((kernelarr=(PyArrayObject *)PyArray_ContiguousFromObject(kernel,
       PyArray_DOUBLE,1,1))==NULL) 
    { Py_DECREF(xarr) ; return NULL ;}
  yn=(xarr->dimensions[0]-1)*step+1 ;
  if ((yarr=(PyArrayObject *)PyArray_FromDims(1,&yn,
       PyArray_DOUBLE))==NULL)
    { Py_DECREF(xarr) ; Py_DECREF(kernelarr) ; return NULL ;}

  if (convolve_upsample((double *)xarr->data,xarr->dimensions[0],
		      (double *)yarr->data,
		      (double *)kernelarr->data,
		       kernelarr->dimensions[0],korig,step))
    { Py_DECREF(xarr) ; Py_DECREF(kernelarr) ;
      Py_DECREF(yarr) ; return NULL ;}
  Py_DECREF(xarr) ; Py_DECREF(kernelarr) ;
  return PyArray_Return(yarr) ;
}



/* ---------- table of functions in our module --------------- */

static struct PyMethodDef BIGsplines_methods[]={
  { "IirConvolvePoles",  BSiirConvolvePoles, METH_VARARGS },
  { "SplineInterpol", BSsplineInterpol, METH_VARARGS },
  { "SplineInterpolMask", BSsplineInterpolMask, METH_VARARGS },
  { "SplineInterpolOne", BSsplineInterpolOne, METH_VARARGS },
  { "FirConvolve", BSfirConvolve, METH_VARARGS },
  { "Sdcgr", BSsdcgr, METH_VARARGS },
  { "SdcgrMask", BSsdcgrMask, METH_VARARGS },
  { "Evalbspln", BSevalbspln, METH_VARARGS },
  { "ConvolveSubsamp", BSconvolveSubsamp, METH_VARARGS },
  { "ConvolveUpsample", BSconvolveUpsample, METH_VARARGS },
  { NULL, NULL} /* stop mark */
} ;

/* ----------- module initialization -------------------------- */
void initBIGsplines() {
  PyObject *m,*d ;

  m=Py_InitModule("BIGsplines",BIGsplines_methods) ; /* register methods */

  /* initialize exception object */
  d=PyModule_GetDict(m) ; /* get module dictionary */
  ErrorObject=PyErr_NewException("BIGsplines.error",NULL,NULL) ;
  PyDict_SetItemString(d,"error",ErrorObject) ;

  import_array() ;

  if (PyErr_Occurred()) /* something went wrong ?*/
    Py_FatalError("can't initialize module BIGsplines") ;
}


/* ---------- end of BIGsplines.c ---------- */
