""" Module representing the next generation of
    Fast Elastic Registration Algorithm
    as described in the corresponding chapter of my thesis.

    When this works, I will remove the routines from other modules
    which have thus become superflous.

    Jan Kybic, April 2001
    $Id: bigfelreg.py,v 1.4 2003/11/27 11:35:51 kybic Exp $
"""

import new,time,operator,types
from Numeric import *
import Image,ImageDraw,ImageFont
import MA
import bigsplines
import bigtools
import bigoptimize
import BIGsplines


# ----------------------------------------------------------------------

class WarpableImageS:
    """ contains n-dimensional image to be warped.

    Reimplementation of bigregister.WarpableImage using MSplineSignal

    The ratio between the image size and the control grid size
    is now constrained to be an integer 'crate' (actually a vector)
    The deformation is a list of MSplineSignals, 'defs', one element for
    each component . It coincides with 'imgs' at zero and goes 'crate'
    times slower.

    """

    def __init__(self,imgs,defs=None,par=None,c=None,mask=None):
        """ imgs is supposed to be MSplineSignal or an array,
            defs is either None, or a MSplineSignalList
            If imgs is an array, we convert it.
            The deformation defaults to identity, or it can be given
            by the MSplineSignal defs or the matrix c

            If mask is not None it determines which pixel values are
            going to be evaluated. It should have the same size as
            the image imgs. For the moment, only the calculation
            of the warped image is affected.
            """
        if type(imgs)==bigtools.ArrayType:
            self.dim=len(imgs.shape)
        else:
            self.dim=imgs.getDim()
        self.par=bigtools.Parameters(par,
                   default={'degf':3,'degc':3,
                            'crate': array((8,)*self.dim,Int),
                            'verbose': 0,
                            'cacheidentity': 1})
        self.imgs=imgs
        if type(self.imgs)==bigtools.ArrayType:
            self.imgs=bigsplines.MSplineSignal(x=self.imgs,degree=self.par.degf)
        self.par.crate=asarray(self.par.crate,Int) # make sure it is Int
        self.defs=defs
        self.g=None # the deformation evaluated at all points
        self.w=None # the warped image (an array)
        self.identity=None
        self.xmax=bigtools.float_array(self.imgs.getShape()-1)/self.par.crate
        if c is not None:
            self.setCFromArray(c)
        else:
            self.defs=defs
        if self.defs is None:
            self.setIdentity()
        else:
            self.defs.setXmax(self.xmax)
        self.mask=mask
        self.dim=self.imgs.getDim()


    def setIdentity(self):
        if len(self.par.crate)!=self.dim:
            raise ValueError, 'Crate should be a vector of length %d' % self.cdim
        self.defs=bigsplines.MSplineSignalList().zero(
            self.xmax,self.dim,degree=self.par.degc)
        self.g=None ; self.w=None

    def getCAsArray(self):
        """ Unpacks the coefficients of the deformation to an array """
        return self.defs.toArray()

    def setCFromArray(self,c):
        """ Sets the internal representation of a deformation from an
        array c """
        self.defs=bigsplines.MSplineSignalList().fromArray(
            c,self.imgs.getDim(),self.xmax,degree=self.par.degc)
        self.g=None ; self.w=None 

        
    def getImageS(self):
        """ Return pointer to imgs """
        return self.imgs

    def getDefS(self):
        """ Return pointer the defs  MSplineSignalList """
        return self.defs

    def setDefS(self,defs):
        """ Set a new defs, invalidating the deformation """
        self.w=None ; self.g=None ; self.defs=defs 
        return self

    def getDeformation(self):
        """ Get the deformation function at all points, caching the
            results.

            NOTE: mask could have been used here, but it is not,
            as evalfract does not support it anyway.
            """
        if self.g is None:
            if self.defs is None:
                raise AttributeError, "deformation not initialized"
            if self.identity is None: # identity is now cached
                if self.par.verbose>7:
                    self.identity=bigtools.timeop(lambda self=self:
                                       indices(self.imgs.getShape(),Float),
                                       "indices")
                else:
                    self.identity=indices(self.imgs.getShape(),Float)
            # begin with identity
            if not self.par.cacheidentity:
                self.g=self.identity ; self.identity=None
            else:
                self.g=array(self.identity)
            if self.par.verbose>7:
                time1=time.clock()
            self.g+=self.defs.evalfract(k=self.par.crate)
            if self.par.verbose>7:
                print "evalfract in getDeformation took ",time.clock()-time1, " s"
            self.w=None
        return self.g

    def getDeformationAtPoints(self,pts):
        """ Gets target position of multiple points pts of size (n,dim),
            (which is the transpose of what interpoldef wants)
            It returns the diplacements in the same format (n,dim)"""
        if len(pts.shape)==1: # just one point?
            pts=reshape(pts,(1,-1))
        if pts.shape[1]!=self.dim:
            raise "Incorrect shape for pts, pts.shape[1] should be %d" %self.dim
        pts=bigtools.float_array(reshape(pts,(-1,self.dim)))
        ptsnorm=pts/self.par.crate # grid point coordinates
        # get the dimensions right
        ptsnorm=reshape(transpose(ptsnorm),[self.dim]+[-1]+[1]*(self.dim-1))
        r=reshape(self.defs.evalat(ptsnorm),(self.dim,-1))
        return transpose(r)+pts # add identity deformation 

    def getContrib(self,x):
        """ Returns an array of the size of self.defs.l[0].c which contains
        the spline values of the corresponding spline functions at point x
        """
        return self.getDefS().getContrib(x/self.par.crate)

    def getWarped(self,g=None):
        """ Get the warped image as an array """
        if self.w is None or g is not None:
            if g is None: g=self.getDeformation()
            if self.par.verbose>7:
                self.w=bigtools.timeop(lambda self=self,g=g:
                     self.imgs.evalat(g,mask=self.mask),"evalat in getWarped")            
            else:
                self.w=self.imgs.evalat(g,mask=self.mask)
        # here ATTENTION ! In MaskedArray mask 1 means do not use element 
        #if self.mask is not None:
        #    self.w=MA.MaskedArray(self.w,copy=0,mask=self.mask,fill_value=0.0)
        return self.w

    def getDerivatives(self):
        """ get all first derivatives of the image 
        """
        imgs=self.imgs
        derw=zeros([self.dim]+list(imgs.getShape()),Float) 
        if self.par.verbose>7:
                time1=time.clock()
        g=self.getDeformation()
        if self.par.verbose>7:
                time2=time.clock()
        for i in range(self.dim):
            derw[i]=imgs.evalderat(g,i,mask=self.mask)
        #if self.mask is not None:
        #    dmask=array([self.mask]*self.dim,MA.MaskType)
        #    derw=MA.MaskedArray(derw,copy=0,mask=dmask,fill_value=0.0)
        if self.par.verbose>7:
                print "evalder took ",time.clock()-time2, " s, ",\
                      "getDeform ",time2-time1, " s"
        return derw


    def cRateIsEven(self):
        """ True if all elements of crate are even """
        return bigtools.sumall(self.par.crate % 2)==0

    def getReducedImg(self):
        """ Returns an equivalent version of WarpableImageS with twice
        reduced image and correspondingly modified deformation structure """ 
        if not self.cRateIsEven():
            raise "Cannot reduce, crate is odd"
        rimgs=self.imgs.reduce()
        rdefs=self.getDefS().copy().multby(0.5)
        par=bigtools.Parameters(self.par, override={'crate':self.par.crate/2})
        return WarpableImageS(rimgs,par=par,defs=rdefs)

    def getReducedImgAndDef(self):
        """ Returns an equivalent version of WarpableImageS with twice
        reduced image and twice reduced deformation grid.
        crate stays constant """ 
        rimgs=self.imgs.reduce()
        rdefs=self.getDefS().copy().reduce().multby(0.5)
        # print "getReducedImgAndDef shape", rimgs.getShape()
        return WarpableImageS(rimgs,par=self.par,defs=rdefs)

    def getReducedDef(self):
        """ Returns an equivalent version of WarpableImageS with twice
        reduced deformation grid. crate is multiplied by two """ 
        rdefs=self.getDefS().copy().reduce()
        par=bigtools.Parameters(self.par, override={'crate':self.par.crate*2})
        return WarpableImageS(self.imgs,par=par,defs=rdefs)

    def getExpandedDef(self):
        """ Returns an equivalent version of WarpableImageS with twice
        expanded deformation grid. crate is divided by two """ 
        if not self.cRateIsEven():
            raise "Cannot expand, crate is odd"
        rdefs=self.getDefS().copy().expand()
        par=bigtools.Parameters(self.par, override={'crate':self.par.crate/2})
        return WarpableImageS(self.imgs,par=par,defs=rdefs)

    def getWarpedOther(self,img):
        """ Warp an image img
             with the same deformation as myself, expanding appropriately
        """
        scf=img.shape/self.imgs.getShape() # integer division
        if scf*self.imgs.getShape()!=array(img.shape,Int):
            raise ValueError, "Scaling factor must be an integer."
        c=self.getCAsArray() ; c*=reshape(scf,(-1,)+self.dim*(1,))
        par=bigtools.Parameters(self.par,
                  override={'crate': array(self.par.crate*scf,Int)})
        wi=WarpableImageS(img,par=par,c=c)
        return wi.getWarped()
                                                   


    def getWarpingIndex(self,c,mask=None):
        """ Calculates the warping index (average geometric error) between
        the current deformation and the deformation given by c.
        The mask here is positive, i.e. 1 means consider, 0 do not consider
        """
        c0=self.getCAsArray()    # remember the actual values
        g0=self.getDeformation() # find deformation
        self.setCFromArray(c)    # new deformation
        g1=self.getDeformation()
        if mask is not None:
            mask=mask.astype(MA.MaskType)
            npix=bigtools.sumall(bigtools.int_array(mask))
            mask=repeat(reshape(mask,[1]+list(mask.shape)),array((2,)))
            g0=MA.MaskedArray(g0,mask=1-mask)
            g0=((g0-g1)**2).filled(fill_value=0.0)
            wi=math.sqrt(bigtools.sumall(g0)/npix)
        else:
            npix=bigtools.prodall(g0.shape)
            g0=(g0-g1)**2
        wi=math.sqrt(bigtools.sumall(g0)/npix)
            
        self.setCFromArray(c0)   # give back the old deformation
        return wi

    def getWarpingIndexDefS(self,defs,par,mask=None):
        """ Calculates the warping index (average geometric error) between
        the current deformation and deformation given by defs and
        par (from which crate, degc are taken) """
        g0=self.getDeformation()
        wi=WarpableImageS(self.imgs,defs=defs,par=par,mask=mask)
        g1=wi.getDeformation()
        if mask is not None:
            mask=mask.astype(MA.MaskType)
            npix=bigtools.sumall(bigtools.int_array(mask))
            mask=repeat(reshape(mask,[1]+list(mask.shape)),array((2,)))
            g0=MA.MaskedArray(g0,mask=1-mask)
            g0=((g0-g1)**2).filled(fill_value=0.0)
            wi=math.sqrt(bigtools.sumall(g0)/npix)
        else:
            npix=bigtools.prodall(g0.shape)
            g0=(g0-g1)**2
        wi=math.sqrt(bigtools.sumall(g0)/npix)
        return wi

    def getWarpingIndexG(self,g0,mask=None):
        """ Calculates the warping index (average geometric error) between
        the current deformation and deformation given by g 
        and mask """
        par=bigtools.Parameters(self.par,override=
              {'crate':g0.shape[1:]/self.imgs.getShape()*self.par.crate})
        wi=WarpableImageS(bigsplines.MSplineSignal(xmax=array(g0.shape[1:])-1),
                          defs=self.defs,par=par,mask=mask)
        g1=wi.getDeformation()
        if mask is not None:
            mask=mask.astype(MA.MaskType)
            npix=bigtools.sumall(bigtools.int_array(mask))
            mask=repeat(reshape(mask,[1]+list(mask.shape)),array((2,)))
            g0=MA.MaskedArray(g0,mask=1-mask)
            g0=((g0-g1)**2).filled(fill_value=0.0)
            wi=math.sqrt(bigtools.sumall(g0)/npix)
        else:
            npix=bigtools.prodall(g0.shape)
            g0=(g0-g1)**2
        wi=math.sqrt(bigtools.sumall(g0)/npix)
        return wi

    def setMask(self,mask):
        self.mask=mask

    def setProjection(self,g0,mask=None):
        """ Given a conforming deformation table g0, set coefficients
        of the deformation so that it is as close as possible (in the
        warping index sense) to g0 """
        p=DProjProblem(self,g0,mask)
        #bigoptimize.test_evproblem(p)
        #print "windexbefore=",self.getWarpingIndexG(g0,mask)
        o=bigoptimize.OptimizerGdesQ(p,par=bigtools.Parameters(self.par,
                       override={'xtol':1e-4}))
        o.smoothToConvergence()
        #print "windexafter=",self.getWarpingIndexG(g0,mask)

        
def test_WarpableImageS():
    ima=bigtools.ImageToArray(Image.open('lena.tiff'))
    par=bigtools.Parameters({'verbose':10,'degc':3, 'degf':0 })
    s=bigtools.timeop(lambda ima=ima,d=par.degf:
                      bigsplines.MSplineSignal(ima,degree=d),"Create signal")
    ws=bigtools.timeop(lambda s=s,par=par: WarpableImageS(s,par=par),
                       "Create Warpable")
    w=bigtools.timeop(lambda ws=ws: ws.getWarped(),"getWarped")
    ws.setDefS(ws.getDefS())
    w=bigtools.timeop(lambda ws=ws: ws.getWarped(),"getWarped seconda volta")
    v1=bigtools.view(w)
    defs=ws.getDefS() ; print "cshape=",defs.l[0].c.shape
    defs.l[0].c[30,20]=100 ; ws.setDefS(defs)
    w2=ws.getWarped()
    v2=bigtools.view(w-w2,title='dif')
    raw_input('Press any key')
    
# ----------------------------------------------------------------------

class WarpingProblemS(bigoptimize.EvaluableProblem):
    """ Implements the accelerated variant of the single-scale
    warping problem. Mostly compatible with bigregister.WarpingProblem

    Accepts two images: reference and test. Reference is an array,
    test is a MSplineSignal which will be embedded in
    ReducWarpableImageS.

    The criterion is SSD per pixel.

    Implements the interface of EvaluableProblem.

    """

    def __init__(self,refimg,testimg=None,testw=None,par=None,mask=None):
        """ refimg is an array, testimg a MSplineSignal
            or an array. Alternatively, you can specify testw
            (a WarpableImageS) directly """
        self.refimg=refimg
        self.par=bigtools.Parameters(par,
                   default={'degf':3,'degc':3,'verbose':0})
        if testw is not None:
            self.testw=testw
        else:
            self.testw=WarpableImageS(testimg,par=par)
        if mask is None:
            self.npix=multiply.reduce(self.refimg.shape)
        else:
            self.npix=bigtools.sumall(mask.astype(Int))
        self.dim=len(refimg.shape)
        self.mask=mask ; self.testw.setMask(mask)
        #if self.mask is not None:
        #    self.refimg=MA.MaskedArray(self.refimg,copy=0,mask=self.mask,
        #                               fill_value=0.0)
        
    def getInitialX(self):
        """ Return the initial state of the deformation.
        It is a private copy of the array, it will not change with time """
        return self.testw.getCAsArray()

    def getX(self):
        """ Currently homonyme for getInitialX(). Returns the last value
        used for the calculation """
        return self.testw.getCAsArray()


    def setX(self,x):
        """ Sets the deformation parameters to x. This functions is
        automatically called whenever new calculation of E or g is needed."""
        self.testw.setCFromArray(x)


    def getE(self,c):
        """ Calculate the criterion for a particular c"""
        self.setX(c) # pass c to the WarpableImage object
        self.wt=self.testw.getWarped() # get warped test image
        self.difwtrefimg=self.wt-self.refimg
        #if self.mask is not None:
        #    # WARNING: a hack, apparently Masked arrays do not work with **2
        #    self.difwtrefimg.set_fill_value(v=0.0)
        #    self.difwtrefimg=array(self.difwtrefimg)
        E=bigtools.sumall(self.difwtrefimg**2) # the data part
        E/=self.npix
        return E

    def getEg(self,c):
        """ Calculate the criterion and its gradient for some c """
        E=self.getE(c) # side effects!, many variables get changed
        derw=self.testw.getDerivatives()
        # multiply all directional derivatives at each pixel
        # by F', where F is the criterion function
         #if self.mask is not None:
         #    # Temporary hack: get a regular array out of a masked one
         #    derw=array(derw) 
        derwp=2.0*self.difwtrefimg[NewAxis,:]*derw
        gr=zeros(c.shape,Float)
        # the next step is to convolve each component of derwp by
        # the upsampled B-spline kernel
        # WARNING: This is dependent on the MSplineSignal implementation
        # evaluate B-spline kernels for all dims
        kernsofs=map(lambda kk,d=self.testw.par.degc:
                     bigsplines.bsplnevalfract(kk,d),list(self.testw.par.crate))
        # let us make the executive function
        def convsub(inp,kernsofs=kernsofs,ndim=None,
                    shift=self.testw.defs.l[0].shift,
                    num=self.testw.defs.l[0].c.shape,
                    rate=self.testw.par.crate):
            kernel=kernsofs[ndim][0] ; kernorig=kernsofs[ndim][1]
            y=BIGsplines.ConvolveSubsamp(inp,kernel,
                    num[ndim],-kernorig-shift*rate[ndim],rate[ndim])
            #print "convsub inp=",inp.shape," out=",y.shape, " rate=",rate,\
            #      " ndim=",ndim, " num=",num
            return y
        #time1=time.clock()
        for i in range(self.dim):
            gr[i]=bigsplines.applyseparablewithcopy(derwp[i],convsub,passdim=1)
        #print "evalfract in getEg took ",time.clock()-time1, " s"
        gr/=self.npix
        return E,gr



def test_WarpingProblemS():
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((64,64)))
    par=bigtools.Parameters({'verbose':7, 'crate':(64,64),
                             'stepf':2, 'stepdownf':10 })
    wi=WarpableImageS(ima,par=par)
    c=wi.getCAsArray()
    c[0,1,2]=0.2
    #c+=1
    wi.setCFromArray(c)
    imw=wi.getWarped()
    #v1=bigtools.view(imw-ima.astype(Float))
    wp=WarpingProblemS(imw,testimg=ima,par=par)
    print "initial wi=",wp.testw.getWarpingIndex(c)
    #E,g=bigtools.timeop(lambda wp=wp,c=0*c: wp.getEg(c),'getEg')
    #print E,g
    #E,g=bigtools.timeop(lambda wp=wp,c=c: wp.getEg(c),'getEg')
    #E=bigtools.timeop(lambda wp=wp,c=c: wp.getE(c),'getE')
    #E=bigtools.timeop(lambda wp=wp,c=c: wp.getE(c),'getE')
    #bigoptimize.test_evproblem(wp,x=c*0.0)
    listE=[]
    def cbfun(E,x,list=listE):
        list.append(x)
    o=bigoptimize.OptimizerGdesQ(wp,par=bigtools.Parameters(par,override={'cbf':cbfun,
                                                              'xtol':1e-3}))
                     # par=bigtools.Parameters(par,override={'startx':c}))
    print "E=",o.getE()
    o.smoothToConvergence()
    c1=o.getX()
    print "c1=",c1, "cdif=",bigtools.maxelement(abs(c1-c))
    print "Epost=",o.getE()
    wp.testw.setCFromArray(c1)
    print "wi=",wp.testw.getWarpingIndex(c)
    z=array(listE)
    g=bigtools.plot(z[:,0,1,2],with='lines')
    g.title('Gdes convergence')
    raw_input('Press any key')
    #print c.shape
    #print E
    #bigoptimize.test_evproblem(wp,x=c)

# ----------------------------------------------------------------------


class MultigridableWarpingProblemSDataPart(WarpingProblemS):
    """ Implements the multigridable version of WarpingProblemS.

    The added methods are canReduce(), getReduced(), updateFine()

    When reducing, we will alternatively coarsen the image and
    the deformation grid, or coarsen both at the same time (default).

    Remark: For the moment, we consider only uniform coarsening
    of the deformation grid

    If grmask is nonzero, gradient mask is automatically computed at
    each level with this percentage of active pixels.

    It is not intended to be run independently. Instead it will be
    a part of a WarpingProblemS object, together with other parts
    of the criteria.
    """

    def __init__(self,par=None):
        """ The par members refimg and testimg
            are either an array or a MSplineSignal.
            Alternatively, you can specify testw
            (a WarpableImageS) directly """
        # add minimal image size threshold, flags, etc.
        self.par=bigtools.Parameters(par,
                   default={'minimgsize':16, 'degc':3,
                            'degf':3, 'alternatered':1,
                            'grmask':0,'reducedimg':0,'reduceddef':0,
                            'refimg':None,'testimg':None,'testw':None})
        self.reducedimg=self.par.reducedimg # the flag to know what to do next
        self.reduceddef=self.par.reduceddef # and how to come back
        refimg=self.par.refimg ; testimg=self.par.testimg
        testw=self.par.testw
        if type(refimg)==bigtools.ArrayType:
            self.refimgs=bigsplines.MSplineSignal(x=refimg,degree=self.par.degf)
        else: # refimg is already a MSplineSignal
            self.refimgs=refimg
            refimg=self.refimgs.evalint() # get the image values

        if self.par.grmask>0: # we want to use a gradient mask
            mask=self.getGradientMask(pct=self.par.grmask)
        else:
            mask=None
        # call parent
        WarpingProblemS.__init__(self,refimg,testimg=testimg,
                                 testw=testw,par=self.par,mask=mask)
        if self.par.verbose>3:
            print "MultigridableWarpingProblemS with img size=",\
            self.refimg.shape, " crate=",self.testw.par.crate, " csize=",\
            self.testw.defs.first().c.shape # , " xmax=",self.testw.xmax

    def canReduceImg(self):
        """ True if Img can be reduced """
        r1=bigtools.minelement(self.refimg.shape)>self.par.minimgsize
        r2=not self.testw.cRateIsEven()
        return r1 and (r2 or self.canReduceDef())

    def canReduceDef(self):
        """ Returns true if the deformation can be further coarsened """
        return bigtools.minelement(self.testw.getDefS().first().xmax)>1.0

    def canReduce(self):
        """ True if any coarsening is possible """
        if self.par.alternatered:
            return self.canReduceImg() or self.canReduceDef()
        else:
            return self.canReduceImg() and self.canReduceDef()

    def getReducedAlternate(self):
        """ Returns a reduced version of the MultigridableWarpingProblemS,
            alternatively reducing the image size and the deformation
            grid size.
        """
        if self.reducedimg: # time to reduce the deformation grid
            if self.canReduceDef():
                return self.getReducedDef()
            if self.canReduceImg():
                return self.getReducedImg()
        else: # reduce the image
            if self.canReduceImg():
                return self.getReducedImg()
            if self.canReduceDef():
                return self.getReducedDef()
        raise "This problem cannot be reduced any more"

    def getReduced(self):
        """ Returns a reduced version of the MultigridableWarpingProblemS
            the particular reducing method can be chosen by the parameters """
        if self.par.alternatered:
            return self.getReducedAlternate()
        else:
            return self.getReducedImgAndDef()

    def getReducedImgAndDef(self):
        """ Reduce both image and deformation grid """
        rimgs=self.refimgs.reduce()
        rtestw=self.testw.getReducedImgAndDef()
        par=bigtools.Parameters(self.par,
              override={'refimg':rimgs,
                        'testw':rtestw,'reducedimg':1,'reduceddef':1,
                        'crate':rtestw.par.crate})
        return self.__class__(par=par)
    
    def getReducedImg(self):
        """ Returns a reduced version of the MultigridableWarpingProblemS
            with reduced images and possibly also the deformation grid,
            if needed
        """
        rimgs=self.refimgs.reduce()
        if self.testw.cRateIsEven():
            rtestw=self.testw.getReducedImg()
        elif self.canReduceDef(): # try to reduce both
            rtestw=self.testw.getReducedImgAndDef()
        else:
            raise "The image in this problem cannot be reduced any more"
        par=bigtools.Parameters(self.par,
              override={'refimg':rimgs,
                        'testw':rtestw,'reducedimg':1,'reduceddef':0,
                        'crate':rtestw.par.crate})
        return self.__class__(par=par)


    def getReducedDef(self):
        """ Returns a reduced version of the MultigridableWarpingProblemS
            with reduced deformation grid.
        """
        rtestw=self.testw.getReducedDef()
        par=bigtools.Parameters(self.par,override={'refimg':self.refimgs,
                        'testw':rtestw,'reducedimg':0,'reduceddef':1,
                        'crate':rtestw.par.crate})
        return self.__class__(par=par)


    def getXcopy(self):
        """ Get a copy of the current state, to be used with updateFine() """
        return self.getXstate().copy()

    def getXstate(self):
        """ Get the current state (no copy), to be used with updateFine() """
        return self.testw.getDefS()

    def updateFine(self,xc0,cp):
        """ Update yourself from the result of coarse level computation.
        xc0 is the coarse grid value received from cp.getXcopy()
        (a private copy) before the coarse grid representation.
        It might or might not be an array.
        cp is the coarse level MultigridableWarpingProblemS in its
        final state. The new fine value is
        xforig+Expand(xc1-xc0) """
        # make a copy of the current coarse level deformation
        coarsedefs=cp.getXcopy()
        coarsedefs.addSignal(xc0,factor=-1.0) # calculates xc1-xc0

        if cp.par.reducedimg:
            # image size changed, we need to multiply
            #print "updateFine multiplying"
            coarsedefs.multby(2.0)

        #print self.getInitialX().shape, xcdif.shape

        if cp.par.reduceddef:
            # deformation grid size changed, we need to expand
            coarsedefs=coarsedefs.expand()
            #print "updateFine expanding xmax=",coarsedefs.first().xmax
            
        finedefs=self.getXstate()
        #print "updateFine finedefs xmax=",finedefs.first().xmax

        expdefs=finedefs.addSignal(coarsedefs)
        self.setXstate(expdefs)
        return self

    def setXstate(self,x):
        """ Converse of getXstate() or getXcopy() """
        self.testw.setDefS(x)
        return self

    def getGradientMask(self,pct=10):
        """  Calculates a mask of refimg, such that it contains 'pct'
        pixels with the highest gradient values

        NOTE: This is not very efficient, as we generate the WarpableImageS
        first
        """
        wi=WarpableImageS(self.refimgs,par=self.par)
        derw=wi.getDerivatives()
        dera=sum(derw*derw) # calculate the norm of the gradient (squared)
        sdera=sort(ravel(dera)) # sort the derivative amplitudes
        trsh=sdera[-(sdera.shape[0]*pct)/100] # threshold for 'pct'% of pixels
        mask=(dera>trsh).astype(MA.MaskType)
        return mask

# ----------------------------------------------------------------------

class MultigridableWarpingProblemBase(bigoptimize.EvaluableProblem):
    """ This is the basis versions of a WarpingProblemS.
        Additional plug-in criteria parts can be added.

        The object of a created class supports the following
        interface methods:
        __init__(self,par=None)
        getInitialX,getE,getEg,setX
        canReduce,getReduced,getXstate,getXcopy,updateFine

        These interface methods must be supported by the first (DataPart)
        client object, the auxiliary ones only need to support

        getE,getEg,canReduce,getReduced

        The auxiliary criterion objects will get an additional keyword argument
        datapart for their __init__ and getReduced() methods and
        should access the current state
        through this pointer.

        The possible parameters object par attributes are refimg,
        testimg, testw, and many others
        
        """
    # a list of classes. 
    # It should be enough to change this tuple when plug-in criterion
    # classes are added
    datapartclass=MultigridableWarpingProblemSDataPart
    auxclasses=()

    def __init__(self,par=None):
        # instantiate the first object
        self.datapart=self.datapartclass(par=par)
        self.auxobjects=[ o(par=par,datapart=self.datapart)
                            for o in self.auxclasses ]

    def getInitialX(self):
        """ Gets the initial value of X. We only relay this call
        to the first (data part) member"""
        return self.datapart.getInitialX()

    def getX(self):
        """ Gets the value of X. We only relay this call
        to the first (data part) member"""
        return self.datapart.getX()

    def getE(self,x):
        """ Relay the call to all members and accumulate """
        r=[ self.datapart.getE(x)]+[ o.getE(x) for o in self.auxobjects ]
        return reduce(operator.add,r)

    def getEg(self,x):
        """ Relay the call to all members and accumulate """
        r=[self.datapart.getEg(x)]+[ o.getEg(x) for o in self.auxobjects ]
        return reduce(lambda x,y: (x[0]+y[0],x[1]+y[1]),r)

    def setX(self,x):
        """ Relay the call to the data part """
        self.datapart.setX(x)
        return self

    def setXstate(self,x):
        """ Relay the call to the data part """
        self.datapart.setXstate(x)
        return self

    def canReduce(self):
        """ Relay the call to all members and accumulate """
        r=[ self.datapart.canReduce()]+[ o.canReduce() for o in self.auxobjects ]
        return reduce(operator.and_,r) # bitwise and. Hopefully does not matter

    def getReduced(self):
        """ Returns an object of my own class, containing reduced
        versions of myself and the auxiliary criterion parts. """
        rdatapart=self.datapart.getReduced()
        rauxobjects=[o.getReduced(datapart=rdatapart) for o in self.auxobjects ]
        r=new.instance(self.__class__,{})
        r.datapart=rdatapart
        r.auxobjects=rauxobjects
        return r
    
    def getXstate(self):
        """ Relay the call to the data part """
        return self.datapart.getXstate()

    def getXcopy(self):
        """ Relay the call to the data part """
        return self.datapart.getXcopy()

    def updateFine(self,xc0,cp):
        """ Relay the call to the data part """
        self.datapart.updateFine(xc0,cp.datapart)
        return self
    
    

# ----------------------------------------------------------------------
        
def test_updateFine():
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((256,256)))
    par=bigtools.Parameters({'verbose':4, 'crate':(64,64) })
    wp=MultigridableWarpingProblemS(ima,testimg=ima,par=par)
    while 1:
        wpr=wp.getReduced()
        c0=wpr.getXstate()
        wp.updateFine(c0,wpr)
        wp=wpr
        
# ----------------------------------------------------------------------

class MWPSpringPlugin:
    """ This is a auxiliary plug-in class to be used in derivatives of
    MultigridableWarpingProblemBase. It represents the landmark
    spring constraints.

    New parameters in par are springfactor, springpos, springtarget
    The added part of the criterion is:

    Espring=Sum springfactor(i)*||g(springpos(i))-springtarget(i)||^2

    springpos,springtarget and springfactor should be arrays with first
    dimension equal to N, the number of springs.

    """

    def __init__(self,par=None,datapart=None):
        self.par,self.datapart=par,datapart # save to local variables
        if len(self.par.springfactor.shape)!=1:
            raise "springfactor must be a 1D vector"
        self.nsprings=self.par.springfactor.shape[0]

    def getE(self,x):
        """ Calculate the contribution of the springs. Note that it is
        crucial that the DataPart is evaluated before the spring part,
        so that the deformation is already set. That's why x is not
        used in this function """
        if self.nsprings==0: return 0.0
        p=self.datapart.testw.getDeformationAtPoints(self.par.springpos)
        p=(p-self.par.springtarget)**2
        E=bigtools.sumall(self.par.springfactor[:,NewAxis]*p)
        if self.par.verbose>6:
            print "Spring plugin E=",E
        return E

    def getEg(self,x):
        """ See comment at getE """
        if self.nsprings==0: return (0.0,zeros(x.shape,Float))
        d=self.datapart.testw.getDeformationAtPoints(self.par.springpos)
        d=d-self.par.springtarget # target difference
        E=bigtools.sumall(self.par.springfactor[:,NewAxis]*d*d)
        gr=zeros(x.shape,Float)
        d*=2.0*self.par.springfactor[:,NewAxis]
        for i in range(self.nsprings): # add contributions from all springs
            # NOTE: the contributions returned by getContrib do not change
            # with x. They could be therefore precomputed at the expense of
            # some memory
            contrib=self.datapart.testw.getContrib(self.par.springpos[i])
            for j in range(self.datapart.dim):
                gr[j]+=d[i][j]*contrib
        if self.par.verbose>8:
            print "Spring plugin E=",E," gr=",gr
        elif self.par.verbose>6:
            print "Spring plugin E=",E
        return E,gr

    def canReduce(self):
        return 1

    def getReduced(self,datapart=None):
        """ Returns a reduced version of the spring plug-in. """
        if self.nsprings!=0 and datapart.par.reducedimg:
            # the image size has been reduced
            par=bigtools.Parameters(self.par,override={
              'springpos': 0.5*self.par.springpos,
              'springtarget': 0.5*self.par.springtarget,
              'springfactor': 4.0*self.par.springfactor})  
        else:
            par=self.par # no changes
        return self.__class__(par=par,datapart=datapart)

class MultigridableWarpingProblemSprings(MultigridableWarpingProblemBase):
    """ WarpingProblem including the spring regulatization """
    auxclasses=(MWPSpringPlugin,)

# ----------------------------------------------------------------------
class MWPWarpingRegularizerPlugin:
    """ This is an auxiliary plug-in class to be used in derivatives of
    MultigridableWarpingProblemBase. It adds a regularization part to
    the criterion, based on penalizing the second derivative of the criterion,
    as implemented in WarpingRegularizer (bigregister.py) 

    New parameter is regularizeralpha

    The added part of the criterion is

    Ereg = regularizeralpha/N_c * Sum_dim Norm(finitedif^2_dim c)^2 * nc_dim 

    where nc_dim is the number of coefficients in dimension dim,
    the norm is l2, and  the finite difference kernel is (-1,2,_1).
    The multiplication by nc_dim normalizes by the node spacing in
    a particular dimension and the division by N_c normalizes by the number
    of coefficients. In this way, the regularization is
    effectively reduced at finer levels.

    Note that the implemented regularization is very close to calculating
    the norm of the Laplacian of the continuous deformation, except that
    we do not integrate the derived B-splines as we should. This is however
    well theoretically justified because of the Riesz property of the
    B-splines, is cheaper computationally and does not make significant
    practical difference.
    """
    def __init__(self,par=None,datapart=None):
        self.par,self.datapart=par,datapart # save to local variables
        self.alpha=self.par.ralpha
        self.rk=array((-1.0,2.0,-1.0)) ; self.rkorig=1 # the kernels
        self.rk2=convolve(self.rk,self.rk,2) ; self.rk2orig=2

    def getE(self,x):
        """ Calculate the energy contribution of the regularization  """
        def filtfun(what,kernel=self.rk,kernelorig=self.rkorig):
                f=convolve(what,kernel,2)
                return bigtools.sumall(f**2)*what.shape[0]
        E=0.0
        for i in range(x.shape[0]): # for all components
            E+=bigtools.sumforalldims(x[i],
                   lambda x,oper=filtfun: bigtools.sumforallrows(x,oper))
        E*=self.alpha/multiply.reduce(x.shape) # normalization
        if self.par.verbose>6:
            print "Regularizer plugin E=",E
        return E

    def getEg(self,x):
        """ Calculate the energy and gradient contribution of the
        regularization """
        def filtfun2(what,kernel=self.rk2,
                    kernelorig=self.rk2orig):
            # the chopping might need a revision for other filters
            f=convolve(what,kernel,2)[kernelorig:-kernelorig]
            return (2.0*what.shape[0])*f
        E=self.getE(x)
        gr=zeros(x.shape,Float)
        for i in range(x.shape[0]): # for all components
            gr[i]=bigtools.vectorsumforalldims(x[i],filtfun2)
        gr*=self.alpha/multiply.reduce(x.shape) # normalization
        return E,gr

    def canReduce(self):
        return 1

    def getReduced(self,datapart=None):
        """ Returns a reduced version of this plug-in. No changes """
        return self.__class__(par=self.par,datapart=self.datapart)


class MultigridableWarpingProblemRegularized(MultigridableWarpingProblemBase):
    """ WarpingProblem including the Laplacian-like regulatization """
    auxclasses=(MWPWarpingRegularizerPlugin,)

    


# ----------------------------------------------------------------------

class MWPWarpIndexPlugin:
    """ This is an auxiliary plug-in class to be used in derivatives of
    MultigridableWarpingProblemBase. It returns the trajectory of the
    optimization process: the values of the energy and the warping index.
    It accepts parameters:
    trueg - (the true deformation as returned by getDefS),
    reslist - a list of tuples (imgsize,crate,E,wi)
    wimask  - the mask as returned for example by
    bigregister.calculateThresholdedMask

    When on the parameter 'measuretime' is set, none of the above is
    calculated and the reslist contains the list of times as returned by
    time.clock()
        """

    def __init__(self,par=None,datapart=None):
        par=bigtools.Parameters(par,default={'measuretime':0})
        self.par,self.datapart=par,datapart # save to local variables
        self.first=1 # first evaluation for this object (level)

    def getE(self,x):
        self.evalwi(x) # evaluates the optimization state
        return 0.0

    def getEg(self,x):
        self.evalwi(x) # evaluates the optimization state
        self.first=0
        return 0.0,0.0*x

    def evalwi(self,x):
        """ Evaluates the state of the optimization and appends it to
        the self.par.reslist list

        For second and subsequent evaluation, if the criterion increases,
        values from previous (successful) evaluation are used instead.
        """
        if self.par.measuretime:
            self.par.reslist.append(time.clock())
            return

        imgsize=self.datapart.refimgs.getShape()
        crate=self.datapart.par.crate
        E=self.datapart.getE(x)
        wi=self.datapart.testw.getWarpingIndexG(self.par.trueg,mask=self.par.wimask)
        if self.par.verbose>6:
            print "evalwi: first=%d, E=%g, wi=%g" % (self.first,E,wi)
        if not self.first and E>self.lastE: # skip value after failure
            E=self.lastE ; wi=self.lastwi
        self.par.reslist.append((imgsize,crate,E,wi))
        self.lastE=E ; self.lastwi=wi
        
    def canReduce(self):
        return 1

    def getReduced(self,datapart=None):
        """ Returns a reduced version of the warpindex plugin. """
        #if datapart.par.reduceddef:
        #    rdefs=par.truedefs.copy().reduce()
        #    par=bigtools.Parameters(par,override={'truedefs':rdefs})
        return self.__class__(par=self.par,datapart=datapart)

    def reduceMask(self,m):
        """ Take a binary mask m and reduce it by half """
        m=array(m) # copy
        m=bigsplines.mconvrect(m,1)
        m=bigtools.mdownsample(m,repeat([2.0],(len(m.shape),)))
        m=(m>0.5).astype(UnsignedInt8) 
        return m

class MultigridableWarpingProblemWI(MultigridableWarpingProblemBase):
    """ WarpingProblem including the warping index calculation """
    auxclasses=(MWPWarpIndexPlugin,)


# ----------------------------------------------------------------------


class MWPSequencePlugin:
    """ This is an auxiliary plug-in class to be used in derivatives of
    MultigridableWarpingProblemBase.
    It returns a sequence of images illustrating the optimization.
    It accepts parameters:
    fileseqmask - the full filename including a printf-style placeholder
               for image number, for example 'regseq%04d.tiff'
    imgsize - target size of a single image, we currently make
              a sequence of images about 2*imgsize+10
        """

    def __init__(self,par=None,datapart=None):
        par=bigtools.Parameters(par,default={'imgsize': (256,256),
                                             'seqnol': [0]})
        # the parameter seqnol is a hack to share sequence numbers
        # between instances. It should be rewritten passing an output
        # function to the plugin
        self.par,self.datapart=par,datapart # save to local variables
        # first evaluation for this object (level), to skip failures
        self.first=1 

    def getE(self,x):
        self.makeimg(x) # evaluates the optimization state
        return 0.0

    def getEg(self,x):
        self.makeimg(x) # evaluates the optimization state
        return 0.0,0.0*x

    def makeimg(self,x):
        """ Creates an image corresponding to current state.

        For second and subsequent evaluation, if the criterion increases,
        previous images are being used (to reconsider)
        """

        E=self.datapart.getE(x)
        if not self.first and E>self.lastE: # skip value after failure
            if self.par.verbose>6:
                print "makeimg: NOT first=%d, E=%g, size=(%d,%d)" \
                  % (self.first,E,self.datapart.difwtrefimg.shape[0],
                     self.datapart.difwtrefimg.shape[1])
            imgtot=self.lastimg
        else:
            if self.par.verbose>6:
                print "makeimg: YES first=%d, E=%g, size=(%d,%d), seqno=%d" \
                  % (self.first,E,self.datapart.difwtrefimg.shape[0],
                     self.datapart.difwtrefimg.shape[1],self.par.seqnol[0])

            # paste the small difference image
            maxel=bigtools.maxelement(self.datapart.refimg)
            imgscl=(self.datapart.difwtrefimg+maxel)/(2.0*maxel)*255.0
            imgscl=clip(imgscl,0,255.0)
            imgdiff=bigtools.ArrayToImage(imgscl)
            imgtot=Image.new('RGB',2*array(self.par.imgsize)+(20,20),
                             (255,255,255)) # all white
            imgtot.paste(imgdiff,(5,5,5+imgdiff.size[0],5+imgdiff.size[1]))
            # big, two colour difference image
            imgdiffbig=bigtools.showDif(self.datapart.wt,self.datapart.refimg)
            imgdiffbig=bigtools.ArrayToImage(imgdiffbig)
            imgdiffbig=imgdiffbig.resize(self.par.imgsize)
                                         #resample=Image.BILINEAR)
            imgtot.paste(imgdiffbig,(15+self.par.imgsize[0],5,
                         15+2*self.par.imgsize[0],5+self.par.imgsize[1]))
            grid=255.0-bigtools.makeGrid(self.par.imgsize)*255.0
            grid=self.datapart.testw.getWarpedOther(grid)
            grid=clip(grid,0,255.0)
            grid=bigtools.ArrayToImage(grid)
            imgtot.paste(grid,(15+self.par.imgsize[0],15+self.par.imgsize[1],
                         15+2*self.par.imgsize[0],15+2*self.par.imgsize[1]))
            
            self.lastE=E 
            self.lastimg=imgtot


        imgcpy=imgtot.copy()  # Make a copy of the image
        draw=ImageDraw.Draw(imgcpy)
        font=ImageFont.load_path('9x15.pil')
        draw.setfont(font) ; textcol=(0,0,255)
        incr=array((0,int(font.getsize('A')[1]*1.3)),Int)
        tpos=(5,15+self.par.imgsize[1])
        draw.text(tpos,'Number: %d' % self.par.seqnol[0],fill=textcol)
        draw.text(tpos+incr,'Image: %dx%d' % self.datapart.difwtrefimg.shape,
                  fill=textcol)
        draw.text(tpos+2*incr,'Pix/knot: %dx%d' %
                  tuple(self.datapart.par.crate),fill=textcol)
        draw.text(tpos+3*incr,'E: %.3f' % self.lastE,fill=textcol)
        
        # Saving
        imgcpy.save(self.par.fileseqmask % self.par.seqnol[0])
        #v=bigtools.viewimg(imgcpy)
        #raw_input('Press any key')
        self.par.seqnol[0]+=1 ; self.first=0

        
    def canReduce(self):
        return 1

    def getReduced(self,datapart=None):
        """ Returns a reduced version of the plugin. """
        return self.__class__(par=self.par,datapart=datapart)


class MultigridableWarpingProblemSQ(MultigridableWarpingProblemBase):
    """ WarpingProblem including the sequence plugin """
    auxclasses=(MWPSequencePlugin,)


# ----------------------------------------------------------------------


def test_MultigridableWarpingProblem():
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((256,256)))
    par=bigtools.Parameters({'verbose':4, 'crate':(64,64) })
    wi=WarpableImageS(ima,par=par)
    c=wi.getCAsArray()
    c[0,1,2]=1
    wi.setCFromArray(c)
    imw=wi.getWarped()
    par=bigtools.Parameters({'verbose':7, 'crate':(9,9), 'refimg':imw,
                             'testimg':ima })
    wp=MultigridableWarpingProblemSprings(par=par)
    for i in range(10):
        try:
            wp=wp.getReduced()
        except:
            break
    c=wp.getInitialX()
    E,g=wp.getEg(c)
    print E
    o=bigoptimize.OptimizerGdesQ(wp,par=par)
    while not o.hasConverged():
        o.makeStep()
    print o.getE()

# ----------------------------------------------------------------------




def test_MROptimizerState():
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((64,64)))
    par=bigtools.Parameters({'xtol':0.1,'verbose':3, 'crate':(32,32) })
    wi=WarpableImageS(ima,par=par)
    c=wi.getCAsArray() ; print "c.shape=",c.shape
    c[0,1,2]=0.2
    c+=1.0
    wi.setCFromArray(c)
    imw=wi.getWarped()
    par=bigtools.Parameters(par,default={'testimg':ima,'refimg':imw})
    wp=MultigridableWarpingProblemBase(par=par)
    print "initial wi=",wp.datapart.testw.getWarpingIndex(c)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
    cf=mrs.getOptimizer().getX()
    #print cf
    print "cdif=", bigtools.maxelement(abs(c-cf))
    wp.datapart.testw.setCFromArray(cf)
    print "wi=",wp.datapart.testw.getWarpingIndex(c)

def test_Springs():
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((64,64)))
    #ima=bigtools.ImageToArray(Image.open('/home/kybic/data/images/lena.png').resize((64,64)))
    par=bigtools.Parameters({'xtol':0.001,'verbose':5, 'crate':(32,32) })
    wi=WarpableImageS(ima,par=par)
    c=wi.getCAsArray() ; print "c.shape=",c.shape
    c[0,1,2]=0.2
    c+=1.0
    wi.setCFromArray(c)
    imw=wi.getWarped()
    tpoint=array(((16,32),),Float)
    truepoint=wi.getDeformationAtPoints(tpoint)
    par=bigtools.Parameters(par,{'springfactor':array((1e8,)),
                                 'springpos':tpoint,
                                 'springtarget':truepoint})
    par=bigtools.Parameters(par,default={'testimg':ima,'refimg':imw})
    wp=MultigridableWarpingProblemSprings(par=par)
    print "initial wi=",wp.datapart.testw.getWarpingIndex(c)
    print "True position of %s is %s" % (repr(tpoint),
               repr(truepoint))
    print "Initial position of %s is %s" % (repr(tpoint),
               repr(wp.datapart.testw.getDeformationAtPoints(tpoint)))
    #bigoptimize.test_evproblem(wp,h=1e-4) ; return
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    bigtools.timeop(lambda mrs=mrs: mrs.solveMR(),'solveMR')
    cf=mrs.getOptimizer().getX()
    #print cf
    print "cdif=", bigtools.maxelement(abs(c-cf))
    wp.datapart.testw.setCFromArray(cf)
    print "wi=",wp.datapart.testw.getWarpingIndex(c)
    print "Final position of %s is %s" % (repr(tpoint),
               repr(wp.datapart.testw.getDeformationAtPoints(tpoint)))


def test_registerUltrasound():
    # read images
    imgn1='images_chus/Imn-Chus-4C-Clip0009/Clip0009-768x576-7_s.tiff'
    imgn2='images_chus/Imn-Chus-4C-Clip0009/Clip0009-768x576-10_s.tiff'
    img1=bigtools.ImageToArray(Image.open(imgn1))
    img2=bigtools.ImageToArray(Image.open(imgn2))
    imgdif=abs(bigtools.int_array(img1)-img2)
    #v1=bigtools.view(img1)
    #v2=bigtools.view(img2)
    difscale=(0,190)
    v3=bigtools.view(imgdif,title='difference before',scale=difscale)
    #print "difbefore", bigtools.maxelement(abs(imgdif))
    # These parameters work well, register in 23s
    #par=bigtools.Parameters({'verbose':7, 'crate':(16,16),
    #                         'xtol':0.5, 'degc':3, 'degf':1 ,
    #                         'alternatered':1})
    par=bigtools.Parameters({'verbose':6, 'crate':(16,16),
                             'xtol':0.5, 'degc':3, 'degf':1 ,
                             'alternatered':1, 'grmask':0,
                             'testimg':img2, 'refimg':img1,
                             'springfactor': array((1000.0,1000.0)),
                             'springpos': array(((140,100),(140,300)),Float),
                             'springtarget': array(((140,50),(140,350)),Float)
                             })
    wp=MultigridableWarpingProblemSprings(par=par)
    print "Starting position of %s is %s" % (repr(par.springpos),
               repr(wp.datapart.testw.getDeformationAtPoints(par.springpos)))
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    mrs.solveMR()
    img2w=wp.datapart.testw.getWarped()
    imgdifaft=abs(img1-img2w)
    defs=mrs.p.getXstate()
    grid=bigtools.makeGrid(img2.shape)
    wg=WarpableImageS(grid,defs=defs,par=par)
    dg=wg.getWarped()
    v5=bigtools.view(dg,title='deformed grid')
    v4=bigtools.view(imgdifaft,title='difference after',scale=difscale)
    print "Final position of %s is %s" % (repr(par.springpos),
               repr(wp.datapart.testw.getDeformationAtPoints(par.springpos)))
    raw_input('Press any key')

    

def test_preserve():
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((16,16)))
    par=bigtools.Parameters({'verbose':5, 'crate':(8,8) })
    wi=WarpableImageS(ima,par=par)
    c=wi.getCAsArray() ; c[0,1,2]=3 ; wi.setCFromArray(c)
    imw=wi.getWarped()
    #v1=bigtools.view(ima,size=(256,256),title='test')
    #v2=bigtools.view(imw,size=(256,256),title='reference')
    imas=bigsplines.MSplineSignal(x=ima).expand().expand()
    imws=bigsplines.MSplineSignal(x=imw).expand().expand()
    #v1=bigtools.view(imas.evalint(),size=(256,256),title='test')
    #v2=bigtools.view(imws.evalint(),size=(256,256),title='reference')
    #v3=bigtools.view(imws.evalint()-imas.evalint(),size=(256,256),
    #                 title='difference')
    par=bigtools.Parameters({'verbose':5, 'crate':(32,32) })
    wp=MultigridableWarpingProblemS(imws,testimg=imas,par=par)
    c=wp.getInitialX()
    print "Eorig=",wp.getE(c)
    wr=wp.getReduced() ; cr=wr.getInitialX()
    print "E1=",wr.getE(cr)
    wrr=wr.getReduced() ; crr=wrr.getInitialX()
    print "E2=",wrr.getE(cr)
    
    #raw_input('Press any key')
    
# ----------------------------------------------------------------------

class DProjProblem(bigoptimize.EvaluableProblem):
    """ Implements a task of evaluating the distance between two
    deformations, one given by a WarpableImageS object and second
    by a table of values. Suitable for minimization using optimizerGdesQ
    and similar.

    Implements the interface of EvaluableProblem.
    """
    
    def __init__(self,wi,tbl,mask=None):
        """ wi is a WarpableImageS, tbl a table as returned from
        WarpableImageS.getDeformation """
        self.w,self.tbl=wi,tbl
        self.par,self.mask=wi.par,mask
        if mask is not None:
            mask=mask.astype(MA.MaskType)
            self.npix=bigtools.sumall(bigtools.int_array(mask))
        else:
            self.npix=bigtools.prodall(g0.shape)
        
    def getInitialX(self):
        """ Return the initial state of the deformation.
        It is a private copy of the array, it will not change with time """
        return self.w.getCAsArray()

    def getInitialX(self):
        """ Return the initial state of the deformation.
        It is a private copy of the array, it will not change with time """
        return self.w.getCAsArray()

    def getX(self):
        """ Currently synonyme for getInitialX(). Returns the last value
        used for the calculation """
        return self.w.getCAsArray()


    def setX(self,x):
        """ Sets the deformation parameters to x. This functions is
        automatically called whenever new calculation of E or g is needed."""
        self.w.setCFromArray(x)


    def getE(self,c):
        """ Calculate the criterion for a particular c"""
        self.setX(c) # pass c to the WarpableImage object
        E=self.w.getWarpingIndexG(self.tbl,self.mask)
        return E**2

    def getEg(self,c):
        E=self.getE(c) # side effects!, many variables get changed
        gr=zeros(c.shape,Float)
        mask=repeat(reshape(self.mask,[1]+list(self.mask.shape)),array((2,)))
        tbl=MA.MaskedArray(self.tbl,mask=1-mask)
        difg=2.0*(self.w.getDeformation()-tbl).filled(fill_value=0.0)
        # the next step is to convolve each component of derwp by
        # the upsampled B-spline kernel
        # WARNING: This is dependent on the MSplineSignal implementation
        # evaluate B-spline kernels for all dims
        kernsofs=map(lambda kk,d=self.w.par.degc:
                     bigsplines.bsplnevalfract(kk,d),list(self.w.par.crate))
        # let us make the executive function
        def convsub(inp,kernsofs=kernsofs,ndim=None,
                    shift=self.w.defs.first().shift,
                    num=self.w.defs.first().c.shape,
                    rate=self.w.par.crate):
            kernel=kernsofs[ndim][0] ; kernorig=kernsofs[ndim][1]
            return BIGsplines.ConvolveSubsamp(inp,kernel,
                    num[ndim],-kernorig-shift*rate[ndim],rate[ndim])
        #time1=time.clock()
        for i in range(self.w.dim):
            gr[i]=bigsplines.applyseparablewithcopy(difg[i],convsub,passdim=1)
        return E,gr/self.npix
    

# ----------------------------------------------------------------------

if __name__=='__main__': # running interactively
    test_MultigridableWarpingProblemS()

# ----------------------------------------------------------------------
