""" This module contains a Python reimplementation of
    Thierry Blu's Matlab code to calculate
    fractional Wavelet transform

    See http://bigwww.epfl.ch/

    Jan Kybic, July 2001 (and indirectly, Thierry Blu, October 1999)
    $Id: bigfract.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
"""

from Numeric import *
import MA
import FFT
import RandomArray
import bigtools
import Image
import bigsplines


datadir='data_felreg/'

# ----------------------------------------------------------------------

def getFFTfractsplinefilters(M,alpha,type):
    """ Returns two pairs of filters:
    (analysisfilters,synthesisfilters), each being pair containing
    a (lowpass,highpass) filter
    M is the signal length and should be a power of two
    alpha is a spline degree and should be grater than -0.5

    The only type implemented is currently '*ortho' (symmetric orthogonal)
    """

    if type!='*ortho':
        raise ValueError, "Type '%s' not implemented." % type

    nu=arange(0,1.0,1.0/M) # the frequencies

    A=fractsplineautocorr(alpha,nu)
    A2=concatenate((A,A))[0::2]
    lowa=sqrt(2.0)*abs((1+exp(-2.0j*pi*nu))/2.0)**(alpha+1.0)*sqrt(A/A2)
    higha=exp(-2.0j*pi*nu)*lowa
    higha=conjugate(concatenate((higha[M/2:],higha[:M/2])))
    filters=((lowa,higha))
    # since we only implement the orthogonal case so far, the analysis
    # and synthesis filters are identical
    return (filters,filters)

# ----------------------------------------------------------------------

def FFTwaveletanalysis(x,filters,J):
    """ computes the wavelet transform of a signal x using a Fourier
    method.  It uses periodic boundary conditions.  The length of the
    signal must be a power of two and the frequency responses of the
    filters specified in 'filters'. 'filters' can be obtained from
    getFFTfractsplinefilters(M,alpha,type)[0]

    J is the depth of the decomposition
    """
    M=len(x) ; (G,H)=filters
    
    if G.shape!=(M,) or H.shape!=(M,):
        raise ValueError,\
               "The size of the input signal and of the filters must match."
    X=FFT.fft(x) # Fourier transform
    w=zeros((0,),Float)
    for j in xrange(J):
        Y=G*X ; Z=H*X # lowpass, highpass filtering
        Y=0.5*(Y[:M/2]+Y[M/2:]); # amounts to up/down sampling
        Z=0.5*(Z[:M/2]+Z[M/2:]);
        if len(Z)<1:
            raise ValueError, " Depth of composition J=%d is too big." % J
        z=FFT.inverse_fft(Z).real;
        w=concatenate((w,z));
        M/=2; X=Y;
        G=G[::2]; H=H[::2] # downsampling filters
    z=FFT.inverse_fft(X).real;
    w=concatenate((w,z));
    return w


# ----------------------------------------------------------------------

def mFFTwaveletanalysis(x,filters,J):
    """ Multidimensional extension of FFTwaveletanalysis

    WARNING: x must by square, otherwise we have a problem
             to be corrected later
    """
    return bigsplines.applyseparable(x, lambda inp,f=filters,J=J:
                          FFTwaveletanalysis(inp,f,J))

# ----------------------------------------------------------------------

def mFFTwaveletsynthesis(x,filters,J):
    """ Multidimensional extension of FFTwaveletsynthesis

    WARNING: x must by square, otherwise we have a problem
             to be corrected later 
    """
    return bigsplines.applyseparable(x, lambda inp,f=filters,J=J:
                          FFTwaveletsynthesis(inp,f,J))

# ----------------------------------------------------------------------

def FFTwaveletsynthesis(w,filters,J):
    """ computes the inverse wavelet transform of coefficients w using a Fourier
    method.  It uses periodic boundary conditions.  The length of the
    signal must be a power of two and the frequency responses of the
    filters specified in 'filters'. 'filters' can be obtained from
    getFFTfractsplinefilters(M,alpha,type)[0]

    J is the depth of the decomposition
    """
    M=len(w) ; (G,H)=filters
    
    if G.shape!=(M,) or H.shape!=(M,):
        raise ValueError,\
               "The size of the input signal and of the filters must match."

    G=conjugate(G) ; H=conjugate(H) ;
    M/=2**J;
    y=w[-M:] # extract low-pass
    w=w[:-M]
    Y=FFT.fft(y) # Fourier transform

    for j in xrange(J-1,-1,-1): # cycle from J-1 to 0
        z=w[-M:] ; w=w[:-M]
        Z=FFT.fft(z)
        H1=H[::2**j] ; G1=G[::2**j]
	Y0=G1[:M]*Y+H1[:M]*Z ; Y1=G1[M:]*Y+H1[M:]*Z
        M*=2
	Y=concatenate((Y0,Y1))

    x=FFT.inverse_fft(Y).real
    return x

# ----------------------------------------------------------------------


def fractsplineautocorr(alpha,nu):
    """ Frequency domain computation of fractional spline 
 	autocorrelation.
    """

    N=100 # number of terms to sum
    S=zeros(nu.shape,Float)

    for n in xrange(-N,N+1):
        S+=abs(sinc(nu+n))**(2.0*alpha+2.0)
    U=2.0/(2.0*alpha+1)/N**(2.0*alpha+1.0);
    U-=1.0/N**(2*alpha+2.0);
    U+=(alpha+1.0)*(1.0/3.0+2.0*nu*nu)/N**(2.0*alpha+3.0);
    U-=(alpha+1.0)*(2.0*alpha+3.0)*nu*nu/N**(2.0*alpha+4.0);
    U*=abs(sin(pi*nu)/pi)**(2.0*alpha+2.0);
    return S+U


# ----------------------------------------------------------------------

def sinc(x):
    """ Evaluates sin(x*pi)/(x*pi) for a vector x """
    pix=MA.masked_values(pi*x,0.0)
    pix=MA.sin(pix)/pix
    return MA.filled(pix,value=1.0)

# ----------------------------------------------------------------------

def genfractrand(r,J):
    """ Generate random 1D signal corresponding to
    Sobolev exponent r. J is the decomposition depth.
    Returns a signal of length 2**J
    """
    
    alpha=r-0.5  # spline degree, beta=2*r
    # print "alpha=",alpha

    M=2**J # total length
    w=zeros((M,),Float)
    for j in xrange(0,J):
        s=M/2**(j+1) ; ofs=M-2*s
        w[ofs:ofs+s]+=RandomArray.normal(0.0,1.0*2.0**(r*(j-J+1)),(s,))
    (af,sf)=getFFTfractsplinefilters(M,alpha,'*ortho')
    y=FFTwaveletsynthesis(w,sf,J)
    return y




def test_genfractrand(r1=0.6,r2=1.6):
    """ generate signals of two different smothnesses """
    RandomArray.seed(123111,14125)
    y=genfractrand(r1,10) ;
    g1=bigtools.plot(y,title='signal',with='lines',xlabel='time')
    g1("set terminal postscript 20")
    g1.hardcopy(datadir+"genfractrand1sig.eps")
    Y=20*log10(abs(FFT.fft(y)))
    g2=bigtools.plot(clip(Y[1:len(Y)/2],-60.0,20.0),
                     xlabel='frequency',ylabel='amplitude [dB]',
                     title='signal spectrum',with='lines',
                     cmd='set yrange [-60.0:20.0] ; set logscale x ')

    g2("set terminal postscript 20")
    g2.hardcopy(datadir+"genfractrand1spect.eps")
    y=genfractrand(r2,10) ;
    g3=bigtools.plot(y,title='signal',with='lines',xlabel='time')
    g3("set terminal postscript 20")
    g3.hardcopy(datadir+"genfractrand2sig.eps")
    Y=20*log10(abs(FFT.fft(y,len(Y))))
    g4=bigtools.plot(clip(Y[1:len(Y)/2],-60.0,20.0),
                     xlabel='frequency',ylabel='amplitude [dB]',
                     title='signal spectrum',with='lines',
                     cmd='set yrange [-60.0:20.0] ; set logscale x  ')
    g4("set terminal postscript 20")
    g4.hardcopy(datadir+"genfractrand2spect.eps")
    
    raw_input('Press any key')


# ----------------------------------------------------------------------

def genfractrand2D(r,J):
    """ Generate random 2D signal corresponding to
    Sobolev exponent r. J is the decomposition depth.
    Returns an image of size (2**J,2**J)
    """
    
    alpha=r-0.5  # spline degree, beta=2*r

    M=2**J # total length
    w=zeros((M,M),Float)
    for j1 in xrange(0,J):
        s1=M/2**(j1+1) ; ofs1=M-2*s1
        for j2 in xrange(0,J):
            s2=M/2**(j2+1) ; ofs2=M-2*s2
            w[ofs1:ofs1+s1,ofs2:ofs2+s2]+=RandomArray.normal(
                0.0,1.0*2.0**(r*(j1+j2-2*(J-1))),(s1,s2))
    (af,sf)=getFFTfractsplinefilters(M,alpha,'*ortho')
    y=mFFTwaveletsynthesis(w,sf,J)
    return y

# ----------------------------------------------------------------------

def test_fract():
    alpha=1.0 ; J=4
    x=arange(16)
    (af,sf)=getFFTfractsplinefilters(len(x),alpha,'*ortho')
    w=FFTwaveletanalysis(x,af,J)
    print w
    y=FFTwaveletsynthesis(w,sf,J)
    print y
    print "peak SNR=",20*log10(bigtools.maxelement(abs(y-x))/
                               bigtools.maxelement(abs(x)))

def test_fractimg():
    ima=bigtools.ImageToArray(Image.open('lena.tiff'))
    alpha=0.7 ; J=8
    (af,sf)=getFFTfractsplinefilters(len(ima),alpha,'*ortho')
    v1=bigtools.view(ima,title='Lena')
    w=bigtools.timeop(lambda x=ima,af=af,J=J:
                      mFFTwaveletanalysis(x,af,J), 'Analysis')
    v2=bigtools.view(w,title='coefs')
    y=bigtools.timeop(lambda x=w,sf=sf,J=J:
                      mFFTwaveletsynthesis(x,sf,J), 'Synthesis')
    print "peak SNR=",20*log10(bigtools.maxelement(abs(y-ima))/
                               bigtools.maxelement(abs(ima)))
    raw_input('Press any key')
                      



# ----------------------------------------------------------------------

if __name__=='__main__': # running interactively
    test_fract()
    test_fractimg()

    
        
