""" Module bigmultigrid implements multidimensional nonlinear
    variational optimization algorithms

    Jan Kybic, November 2000
    $Id: bigmultigrid.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
    """

import math
from Numeric import *
import bigtools
import bigpyramid
import bigoptimize

# --------------------------------------------------------------------

class MultigridOptimizerState:
    """ A class representing a multigrid optimizer and its state
    """

    def __init__(self,problem,optimizer=bigoptimize.OptimizerGdes,
                 verbose=0,maxiter=100,abstol=1e-50,reltol=1e-5,
                 xtol=1e-3,
                 smoothcount=1,callbackE=None,level=0,
                 maxlevel=1000,itercounts=None):
        # copy parameters to local variables
        self.p=problem ; self.optclass=optimizer ;
        self.verbose,self.maxiter,self.smoothcount=verbose,maxiter,smoothcount
        self.abstol,self.reltol=abstol,reltol
        self.callbackE=callbackE # function to be called after each evaluation
        self.level,self.xtol=level,xtol
        self.maxlevel=maxlevel # maximum level we will go up to
        # create optimizer
        self.o=self.optclass(self.p,verbose=self.verbose>4,
                             reltol=self.reltol,abstol=self.abstol,
                             xtol=self.xtol)
        self.E=self.o.getE() # remember the criterion value
        self.oldE=None       # previous value
        if itercounts==None:
            self.itercounts={} # total number of iterations
        else:
            self.itercounts=itercounts
        if self.verbose>4:
            print "itercounts=",self.itercounts
            print "MultigridOptimizerState at level ",level, " with xsize=",\
                  self.o.getX().shape


    def setX(self,x):
        self.o.setX(x)
        
    def getReducedOptimizer(self,correct=0):
        """ If possible, get a reduced version of the problem
        and return an MultigridOptimizerState ready to solve it"""
        # get reduced version of the problem
        if not self.canReduce():
            raise ValueError, "Cannot reduce"
        if self.p.__dict__.has_key('doCorrect'): # set correction, if supported
           self.p.doCorrect=correct 
        pc,xc=self.p.getReduced(self.o.getX())
        # get reduced version of the MultigridOptimizerState
        r=MultigridOptimizerState(pc,self.optclass,self.verbose,
                                  self.maxiter,self.abstol,self.reltol,
                                  self.xtol,self.smoothcount,self.callbackE,
                                  self.level+1,self.maxlevel,self.itercounts)
        # set optimizer using his parents data
        r.setX(xc)
        try:  # if supported, set the step size from the last iteration
            r.o.setStep(self.o.getStep())
        except:
            print "Setting step size unsupported for this optimizer"
        return r

    def addToIterCount(self,x):
        """ adds x to an iteration count at given level """
        if self.itercounts.has_key(self.level):
            self.itercounts[self.level]+=x
        else:
            self.itercounts[self.level]=x
        
    def smooth(self):
        """ Apply the smoothing operation (optimizer) smoothcount times"""
        oldit=self.o.i
        for i in range(self.smoothcount):
            self.o.makeSuccessfulStep()
            if self.callbackE:
                self.callbackE(self.level,self.o.getE())
        self.addToIterCount(self.o.i-oldit)
        self.oldE=self.E
        self.E=self.o.getE()

    def smoothToConvergence(self):
        """ Apply the smoothing operation until convergence"""
        oldit=self.o.i
        while not self.o.hasConverged():
            self.o.makeStep()
            if self.callbackE:
                self.callbackE(self.level,self.o.getE())
        self.addToIterCount(self.o.i-oldit)
        self.oldE=self.E
        self.E=self.o.getE()

    def hasConvergedE(self):
        """ This function decides if we have converged or not, based on E """
        if self.oldE==None: # we have not even started
            return 0
        dif=abs(self.E-self.oldE)
        return dif<=self.abstol or dif<=self.reltol*self.E

    def solveMR(self):
        """ Return a solution using a multiresolution approach """
        xf0=self.o.getX()
        if self.canReduce(): # not the coarsest level
            r=self.getReducedOptimizer()
            xc0=array(r.o.getX()) # copy the inital value
            xc1=r.solveMR()   # solve recursively
            x=self.p.updateFine(xf0,xc0,xc1)
            self.o.setX(x) # update fine level solution
        self.smoothToConvergence()
        return self.o.getX()

    def tCycleCGC(self,correct=0):
        if correct:
            name='Corrected'
        else:
            name='Simple'
        xf0=self.o.getX() ; E0=self.o.getE()
        r=self.getReducedOptimizer(correct)
        xc0=array(r.o.getX()) # copy initial value at coarse grid
        xc1=r.makeTcycle()    # recursive call
        xf1=self.p.updateFine(xf0,xc0,xc1)
        E1=self.p.getE(xf1)
        result=0
        if E1<E0:                 # criterion has improved
            self.setX(xf1)
            result=1
            if self.verbose>3:
                print self.level," "+name+" coarse grid correction helped E1=",\
                      E1," E0=",E0
        else:
            if self.verbose>3:
                print self.level," "+name+" coarse grid correction failed E1=",\
                      E1," E0=",E0
        return result

    def canReduce(self):
        return self.level<self.maxlevel and self.p.canReduce()

    def makeTcycle(self):
        """ Makes one iteration of the T cycle. This consists of
        trying to improve the criterion by solving a reduced
        problem (we skip this now), then a corrected reduced
        problem, or finally, if all else fails, to smooth at the
        current level.
        
        Returns a reference to the resulting x.
        """
        if self.verbose>1:
            print "Entering T cycle at level=",\
                  self.level," E=",self.o.getE()
        # if not self.canReduce() or (
        #    not self.tCycleCGC() and not self.tCycleCGC(correct=1)):
        if not self.canReduce() or not self.tCycleCGC(correct=1):
            # both CGC failed or cannot reduce any more
            E0=self.o.getE()
            self.smooth()
            E1=self.o.getE()
            if self.verbose>3:
                print self.level, " smoothing improved E1=",E1," E0=",E0
        return self.o.getX()


    def makeTcyclesToConvergence(self):
        """ Keep executing T cycles until we converge

        Returns a reference to a resulting x.
        """
        while not self.hasConverged():
            self.makeTcycle()
        return self.o.getX()

    def solveFullT(self):
        """ Return a solution using a full multigrid algorithm using T
        cycles """
        # first find the best starting estimate by recursive call
        if self.verbose>1:
            print "Entering solveFullT at level=",\
                  self.level," E=",self.o.getE()

        if self.canReduce(): # if reduction is still possible
            r=self.getReducedOptimizer()
            xf0=self.o.getX() # initial value
            xc0=array(r.o.getX())
            xc1=r.solveFullT() # recursive call
            x=self.p.updateFine(xf0,xc0,xc1)
            self.setX(x)
            # initialization finished, make T cycles until convergence
            if self.verbose>1:
                print "Initialization finished in solveFullT at level=",\
                  self.level," E=",self.o.getE()
            self.maxlevel=self.level+1
            self.makeTcyclesToConvergence()
        else:
            # if no reduction is possible, iterate until convergence
            self.smoothToConvergence() # this automatically updates x

        if self.verbose>1:
            print "solveFullT finished  at level=",\
                  self.level," E=",self.o.getE()
        return self.o.getX()

            
# --------------------------------------------------------------------
# Testing routines

import Gnuplot
import bigregister
import Image
import bigsplines
import bigpyramid

def showEevol(listE):
    z=array(listE)
    t=arange(z.shape[0])
    g=Gnuplot.Gnuplot()
    g1=Gnuplot.Gnuplot()
    d1=Gnuplot.Data(t,z[:,0],with='lines',title='level') # level data
    g('set logscale y')
    d2=Gnuplot.Data(t,z[:,1],with='lines',title='criterion')
    g.plot(d2) ; g1.plot(d1)
    # g.hardcopy('fmgres.eps',enhanced=1)
    raw_input('Press any key')


def testMOS(p=None):
    if not p:
        p=bigoptimize.MultigridableProblem(xshape=(32,))
    listE=[]
    def cbfun(E,l,list=listE):  list.append((E,l))

    m=MultigridOptimizerState(p,verbose=5,callbackE=cbfun,xtol=1e-3,
                              optimizer=bigoptimize.OptimizerGdes)
    x=bigtools.timeop(lambda opt=m: opt.solveMR(),'Multiresolution')
    #print x
    print "Iterations in testMOS ",m.itercounts
    #g=bigtools.plot(x)
    showEevol(listE)
    return x

def testTcycle():
    listE=[]
    def cbfun(E,l,list=listE):  list.append((E,l))


    p=bigoptimize.MultigridableProblem(xshape=(32,))
    p=bigoptimize.CorrectedMultigridableProblem(p)
    m=MultigridOptimizerState(p,verbose=4,callbackE=cbfun)
    #for i in range(10):
    #    x=m.makeTcycle()
    x=m.makeTcyclesToConvergence()    
    #print x
    print "Iterations in testTcycle ",m.itercounts
    g=bigtools.plot(x)
    showEevol(listE)


def testFullT(p=None):
    listE=[]
    def cbfun(E,l,list=listE):  list.append((E,l))

    if not p:
        p=bigoptimize.MultigridableProblem(xshape=(32,))
        p=bigoptimize.CorrectedMultigridableProblem(p)
    m=MultigridOptimizerState(p,verbose=5,callbackE=cbfun)
    x=bigtools.timeop(lambda opt=m: opt.solveFullT(),'FullT')
    #print x
    print "Iterations in testTcycle ",m.itercounts
    #g=bigtools.plot(x)
    showEevol(listE)
    return x

def getWarpedLena(degf=3,degc=3,size=(64,64)):
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize(size))
    #ima=bigtools.ImageToArray(Image.open('lena.tiff'))
    print "Initialization finished"
    imb=bigtools.timeop(lambda x=ima,y=degf:
                        bigsplines.mfspline(x,y),'Bspline transform')
    w=bigregister.WarpableImage(ima,degf,degc)
    w.setIdentity((2,4,4))
    #w.setIdentity((2,8,8))
    c0=array(w.getC())
    #c0[0,2,1]=0.1 # deform
    #c0[0,4,2]=10 # deform
    c0[0,2,1]=30 # deform
    c0[1,3,2]=-20 # deform
    w.setC(c0) 
    wi=w.getWarped() 
    print "Lena warped"
    #wp=bigregister.WarpingProblem(wi,imb,c0.shape,degf,degc)
    refp=bigpyramid.BSPyramid(wi,degree=degf)
    testp=bigpyramid.BSPyramid(ima,degree=degf)
    m=bigregister.calculateMask(ima,10)
    wp=bigregister.MultigridableWarpingProblem(refp,testp,c0.shape,degf,degc,
                                               mask=m,verbose=0)
    return wp,w

def test_lena():
    degf=3 ; degc=2 ;
    wp,w=getWarpedLena(degf,degc,size=(256,256))
    #bigtools.view(wi)
    ima=wp.testp.getimage(0)
    wi=wp.refp.getimage(0)
    x=testMOS(wp)
    x=wp.denormalizeC(x)
    print array2string(x,precision=3,suppress_small=1)
    w.setC(x)
    wd=w.getWarped()
    print "Lena warped"
    a1=bigtools.view(wd)
    a2=bigtools.view(abs(ima-wi),size=(512,512),scale=(0,185),title='before')
    a3=bigtools.view(abs(wd-wi),size=(512,512),scale=(0,185),title='after')
    raw_input('Press any key')	

# --------------------------------------------------------------------

if __name__=='__main__': # running interactively
    #testTcycle()
    #testMOS()
    #testFullT()
    test_lena()
