""" Module bigoptimize implements multidimensional nonlinear optimization
    technique and some prototype problems

    Jan Kybic, October 2000
    $Id: bigoptimize.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
    """

import math
from Numeric import *
import BIGsplines
import bigtools
import bigpyramid
import time
import Gnuplot
import bigsplines
import LinearAlgebra

# --------------------------------------------------------------------

class EvaluableProblem:
    """ Prototype multidimensional optimization problem

        E=(x0-1)^4+(x1-2)^4+(x2-3)^4+(x0-1)^2*(x1-2)^2*(x2-3)^2

        The minimum is clearly at (1,2,3)
    """

    def getInitialX(self): # returns a suitable first guess of the solution
        return array((0,0,0),Float) # just start with zeros

    def getE(self,x):      # return the scalar criterion to be minimized
        return (pow(x[0]-1,4)+pow(x[1]-2,4)+pow(x[2]-3,4)+pow(x[0]-1,2)*
                pow(x[1]-2,2)*pow(x[2]-3,2))

    def getEg(self,x):     # returns the criterion and its gradient
        return self.getE(x),array((
            4*pow(x[0]-1,3)+2*(x[0]-1)*pow(x[1]-2,2)*pow(x[2]-3,2),
            4*pow(x[1]-2,3)+2*(x[1]-2)*pow(x[0]-1,2)*pow(x[2]-3,2),
            4*pow(x[2]-3,3)+2*(x[2]-3)*pow(x[1]-2,2)*pow(x[0]-1,2)))

    def getEgDh(self,x):   # returns E,gr, and the diagonal of the Hessian
        r=self.getEg(x)
        return r[0],r[1],array((
            12*pow(x[0]-1,2)+2*pow(x[1]-2,2)*pow(x[2]-3,2),
            12*pow(x[1]-2,2)+2*pow(x[0]-1,2)*pow(x[2]-3,2),
            12*pow(x[2]-3,2)+2*pow(x[1]-2,2)*pow(x[0]-1,2)))

    # the following methods are ment mainly for testing

    def numg(self,x,h=1e-6): # calculate the gradient at x numericaly
        g=zeros(x.shape,Float) # here we store the gradient
        E=self.getE(x)
        for i in bigtools.getindices(x):
            xf=x+bigtools.basisv(x.shape,i)*h
            g[i]=self.getE(xf)-E # forward differencing
        return g/h
        

    def gerr(self,x,h=1e-6): # calculates the linf error of gradient computation
        return bigtools.maxelement(abs(self.getEg(x)[1]-self.numg(x,h)))

    def numDh(self,x,h=1e-6): # calculate the Hessian diagonal at x numericaly
        H=zeros(x.shape,Float) # here we store the gradient
        g=self.getEg(x)[1]
        for i in bigtools.getindices(x):
            xf=x+bigtools.basisv(x.shape,i)*h 
            H[i]=self.getEg(xf)[1][i]-g[i] # forward differencing
        return H/h

    def dHerr(self,x,h=1e-6): # calculates the linf error of Hessian computation
        return bigtools.maxelement(abs(self.getEgDh(x)[2]-self.numDh(x,h)))

    
# --------------------------------------------------------------------


class HarmonicProblem(EvaluableProblem):
    """ Larger 1D prototype problem, whose solution is a harmonic function

        E=Int  f'^2+T f^2  dx

        evaluated at x=0..1 on a N point grid. Fix f[0]=1
        Use MirrorOnBounds boundary conditions
    """

    def __init__(self,T=1.0,xshape=(10,)):
        self.T,self.N=T,xshape[0]
        self.kernel,self.kernelorig=array((1.0,-1.0)),0
        # calculate k2=kernel*kernel
        self.k2,self.k2orig=array((-1.0,2.0,-1.0)),1

    def getInitialX(self):
        x=zeros((self.N,),Float) ; 
        return x

    def newx(self,xp):
        x=zeros((xp.shape[0]+1,),Float) 
        x[1:]=xp ; x[0]=1.0
        return x

    def getE(self,xp):
        x=self.newx(xp)
        #f=BIGsplines.FirConvolve(x,self.kernel,self.kernelorig)
        f=convolve(x,self.kernel,2) # zero padding of x
        E=dot(f,f)*self.N+self.T/self.N*dot(x,x)
        return E

    def getEg(self,xp):
        x=self.newx(xp)
        #f=BIGsplines.FirConvolve(x,self.k2,self.k2orig)
        f=convolve(x,self.k2,2)[1:-1] # chopping is filter dependent
        return self.getE(xp),2.0*(f[1:]*self.N+self.T/self.N*x[1:])

    def getEgDh(self,xp):
        x=self.newx(xp)
        r=self.getEg(xp) ;
        k=self.k2[self.k2orig]
        y=k*ones(x.shape)
        return r[0],r[1],2.0*(y[1:]*self.N+self.T/self.N)
        
        
        
# --------------------------------------------------------------------

class SimpleReducer:
      """ The SimpleReducer class assembles the expand and reduce operations
      based on linear averaging and linear interpolation. It works
      for matrices with arbitrary dimensions. """

      def simplereduce(self,x):
          """ reduce 1D vector by 3pt averaging """
          c=BIGsplines.FirConvolve(x,array((0.25,0.5,0.25)),1)
          return bigtools.downsample(c,2)

      def simpleexpand(self,x,targetlen=None):
          """ expand 1D vector by 3pt averaging """
          x=bigtools.upsample(bigtools.addmirrored(x,1),2)
          x=convolve(x,array((0.5,1,0.5)))
          xlen=x.shape[0]
          tlen=targetlen or xlen-2
          if tlen==xlen: return x
          if tlen==xlen-1: return x[:-1]
          if tlen==xlen-2: return x[1:-1]
          raise BIGsplines.error, 'Target length %d cannot be obtained' % tlen

      def reduce(self,x):
          """ reduce arbitrary vector/matrix x """
          #print "Simple reducing"
          # return bigpyramid.bsreduce(x,1)
          return bigsplines.applyseparablewithcopy(x,
                 lambda inp,s=self: s.simplereduce(inp) )

      def reduceexceptfirst(self,x):
          """ reduce x except the first dimension """
          y1=self.reduce(x[0])
          y=zeros((x.shape[0],)+y1.shape,Float)
          y[0]=y1
          for i in range(1,x.shape[0]):
              y[i]=self.reduce(x[i])
          return y

      def expandexceptfirst(self,x,targetlen=None):
          """ expand x except the first dimension """
          #print "expandexceptfirst"
          if targetlen is not None:
              t=targetlen[1:]
          else:
              t=None
          y1=self.expand(x[0],targetlen=t)
          y=zeros((x.shape[0],)+y1.shape,Float)
          y[0]=y1
          # print 'expandexceptfirst y.shape=',y.shape, ' x.shape=',x.shape
          for i in range(1,y.shape[0]):
              y[i]=self.expand(x[i],targetlen=t)
          return y

      def expand(self,x,targetlen=None):
          """ expand arbitrary vector x """
          #print "Simple expanding"
          # return bigpyramid.bsexpand(x,1)
          t=targetlen
          return bigsplines.applyseparablewithcopy(x,
                 lambda inp,s=self,tlen=None:
                 s.simpleexpand(inp,targetlen=tlen), targetlen=t )
          
# --------------------------------------------------------------------
class SamplingReducer(SimpleReducer):
      """ The SimpleReducer class assembles the expand and reduce operations
      based on sampling and linear interpolation. It works
      for matrices with arbitrary dimensions. """

      def simplereduce(self,x):
          """ reduce 1D vector by 3pt averaging """
          return bigtools.downsample(x,2)

      def simpleexpand(self,x,targetlen=None):
          """ expand 1D vector by 3pt averaging """
          x=bigtools.upsample(bigtools.addmirrored(x,1),2)
          x=convolve(x,array((0.5,1,0.5)))
          xlen=x.shape[0]
          tlen=targetlen or xlen-2
          if tlen==xlen: return x
          if tlen==xlen-1: return x[:-1]
          if tlen==xlen-2: return x[1:-1]
          raise BIGsplines.error, 'Target length %d cannot be obtained' % tlen
# --------------------------------------------------------------------


class SplineReducer(SimpleReducer):

    def getsplinedegree(self):
        deg=3
        try:
            deg=self.degc
            # print "getsplinedegree=",deg
        except:
            pass
        return deg

    def reduce(self,x):
        #print "Spline reducing"
        rx=bigpyramid.mbsreduce(x,self.getsplinedegree())
        return rx

    def expand(self,x,targetlen=None):
        #print "Spline expanding"
        y=bigpyramid.mbsexpand(x,self.getsplinedegree())
        if targetlen:
            targetlen=asarray(targetlen,Int)-1
            s=bigtools.rangetoslice((zeros(targetlen.shape),targetlen))
            y=y[s]
        return y

# --------------------------------------------------------------------

class MultigridableProblem(HarmonicProblem,SamplingReducer):
#class MultigridableProblem(HarmonicProblem,SimpleReducer):
#class MultigridableProblem(HarmonicProblem,SplineReducer):
      """ An extension of HarmonicProblem to show the multigrid protocol.

          The following methods are added:
          getReduced(x) which gets a reduced version of the problem and
                        the parameters x
          updateFine(xf,xc0,xc1)  which updates xf at fine level with the
                                  difference xc1-xc0 at the coarse level
          canReduce()   returns true if further reduction is possible
                                  
          The following implementation is only prototypical and will
          probably have to be modified for a specific problem.
      """

      def reduceX(self,x):
          return self.reduce(x)

      def getReduced(self,x):
            """ returns (pc,xc), the coarsened problem and the coarsened x
            """
            xc=self.reduceX(x) 
            pc=MultigridableProblem(T=self.T,xshape=xc.shape)
            return pc,xc

      def updateFine(self,xf,xc0,xc1):
            """ returns an updated array xf based on the coarse x before (xc0)
            and after (xc1) coarse level smoothing """
            return xf+self.expand(xc1-xc0,targetlen=xf.shape)

      def canReduce(self):
            return self.N>=8 # heuristic limit
        
# --------------------------------------------------------------------

class CorrectedMultigridableProblem(EvaluableProblem):
    """ Take an MultigridableProblem and make another multigridable problem
    out of it. A correction term is added so that the fixed point of the fine
    grid solution is a fixed point of the coarse grid solution, too.

    The correction comes from Bouman ICIP99 paper and is:

    E'=E-<r,x>
    where r= grad_x1 E1(R(x0)) - R grad_x0 E0(x0)

    This implementation inherits: updateFine
    It encapsulates the rest of the EvaluableProblem routines
    """

    def __init__(self,p):
        """ p is an MultigridableProblem instance which we will encapsulate
        """
        self.p=p            # instance of the inner problem
        self.r=zeros(self.p.getInitialX().shape,Float) # the correction vector
        self.doCorrect=0    # do not correct by default
        self.Eofs=0.0
        
    def getReduced(self,x):
        """ returns (pc,xc), the coarsened problem and the coarsened x
        """
        spc,xc=self.p.getReduced(x)
        # create smaller Corrected problem encapsulating spc
        pc=CorrectedMultigridableProblem(spc)          
        if self.doCorrect:
            gr0=self.p.getEg(x)[1]     # fine-grid gradient
            gr1=pc.p.getEg(xc)[1]      # get true coarse-grid gradient
            if gr1.shape!=gr0.shape:
                gr0=self.p.reduceX(gr0)
            pc.r=gr1-gr0               # set correction term in the child
            pc.doCorrect=1
            pc.Eofs=bigtools.sumall(pc.r*xc)
        #    print "corrected reduction"
        #else:
        #    print "uncorrected reduction"
        return pc,xc

    def getInitialX(self):
        return self.p.getInitialX()

    def getE(self,x):
        """ Return corrected criterion """
        return self.p.getE(x)-bigtools.sumall(self.r*x)+self.Eofs

    def getEg(self,x):
        """ Return corrected criterion and gradient """
        E,g=self.p.getEg(x)
        return E-bigtools.sumall(self.r*x)+self.Eofs, g-self.r

    def getEgDh(self,x):
        E,g,H=self.p.getEgDh(x)
        return E-bigtools.sumall(self.r*x), g-self.r, H
        
    def canReduce(self):
        return self.p.canReduce()

    def updateFine(self,xf,xc0,xc1):
        return self.p.updateFine(xf,xc0,xc1)

    def reduceX(self,x):
        return self.p.reduceX(x)

# --------------------------------------------------------------------

class QCorrectedMultigridableProblem(CorrectedMultigridableProblem):
    """ Quadratic extension to Bouman correction
        The corrected problem is:

        E'=E-betacorr*<r,x>+<r,x>^2

        This assures that E' does no tend to -infty in any direction
        and consequently the existence of the minimum. 
    """


    def __init__(self,p):
        """ p is an MultigridableProblem instance which we will encapsulate
        """
        CorrectedMultigridableProblem.__init__(self,p) # call father
        self.betacorr=0.0

    def getReduced(self,x):
        """ returns (pc,xc), the coarsened problem and the coarsened x
        """
        spc,xc=self.p.getReduced(x)
        # create smaller Corrected problem encapsulating spc
        pc=QCorrectedMultigridableProblem(spc)          
        gr0=self.p.getEg(x)[1]     # fine-grid gradient
        gr1=pc.p.getEg(xc)[1]      # get true coarse-grid gradient
        if gr1.shape!=gr0.shape:
            gr0=self.p.reduceX(gr0)
        z=gr0-gr1               # set correction term in the child
        A=2.0*bigtools.sumall(z*xc)
        if A<0:
            sA=sqrt(A) ; pc.betacorr=-2.0*sA
            alpha=1.0/sA
        else:
            alpha=1.0
            pc.betacorr=1.0-A
        pc.r=alpha*z                      # set correction term in the child
        print "A=",A," alpha=",alpha," beta=",pc.betacorr, "z=",z
        #test_evproblem(pc,xc)
        return pc,xc

    def getE(self,x):
        """ Return corrected criterion """
        q=bigtools.sumall(self.r*x)
        Eorig=self.p.getE(x)
        #print "Eorig=",Eorig," q=",q," q2=",q*q
        return Eorig+self.betacorr*q+q*q

    def getEg(self,x):
        """ Return corrected criterion and gradient """
        E,g=self.p.getEg(x)
        q=bigtools.sumall(self.r*x)
        #print "Eorig=",E," q=",q," q2=",q*q
        return E+self.betacorr*q+q*q,g+self.r*(self.betacorr+2.0*q)

    def getEgDh(self,x):
        """ Return corrected criterion and gradient """
        E,g,H=self.p.getEgDh(x)
        q=bigtools.sumall(self.r*x)
        #print "Eorig=",E," q=",q," q2=",q*q
        return (E+self.betacorr*q+q*q,g+self.r*(self.betacorr+2.0*q),
                H+2.0*self.r*self.r)

# --------------------------------------------------------------------

class WCorrectedMultigridableProblem(CorrectedMultigridableProblem):
    """ Windowed extension to Bouman correction


        The corrected problem is:

        E'=E-<r,x>*W(<r,x>)

        where W(z)=exp(-(z/sigma)^2) is a Gaussian shaped window

        where the window size 'sigma^2=20 zm^2' is calculated from

        zm=| I R x- x | * |r| /2

        This assures that E' coincides with E except a small neighborhood
        of the original point.

    """


    def __init__(self,p):
        """ p is an MultigridableProblem instance which we will encapsulate
        """
        CorrectedMultigridableProblem.__init__(self,p) # call father
        self.x0=zeros(self.p.getInitialX().shape,Float)
        self.r=zeros(self.p.getInitialX().shape,Float)
        self.sigma2=1e-10    # squared sigma
        self.minsigma2=1e-30 # minimum acceptable sigma

    def getReduced(self,x):
        """ returns (pc,xc), the coarsened problem and the coarsened x
        """
        spc,xc=self.p.getReduced(x)
        # create smaller Corrected problem encapsulating spc
        pc=WCorrectedMultigridableProblem(spc)          
        gr0=self.p.getEg(x)[1]     # fine-grid gradient
        gr1=pc.p.getEg(xc)[1]      # get true coarse-grid gradient
        if gr1.shape!=gr0.shape:
            gr0=self.p.reduceX(gr0)
        pc.r=gr0-gr1           # set correction term in the child
        pc.x0=xc               # origin
            # discretization error
            # errd=self.p.expand(xc,targetlen=x.shape)-x
            # the square of a typical distance
            # normprod=0.25*bigtools.sumall(errd**2)*bigtools.sumall(pc.r**2)
            # pc.sigma2=max(20.0*normprod,self.minsigma2) # to avoid zero div.
        print "r=",pc.r
        pc.sigma2=1e-5
            #print "sigma2=",pc.sigma2
        #test_evproblem(pc,xc,h=1e-8)
        return pc,xc

    def getTau(self,x):
        """ Calculate the scalar product tau2=<r,x-x0> """
        return bigtools.sumall(self.r*(x-self.x0))

    def getWindow(self,tau2):
        """ Calculate the window value """
        return math.exp(-(tau2/self.sigma2))

    def getE(self,x):
        """ Return corrected criterion """
        tau=self.getTau(x)
        Eorig=self.p.getE(x)
        #print "Eorig=",Eorig," tau=",tau," E=",Eorig-tau*self.getWindow(tau**2)
        return Eorig-tau*self.getWindow(tau**2)

    def getEg(self,x):
        """ Return corrected criterion and gradient """
        E,g=self.p.getEg(x)
        Eorig=self.p.getE(x)
        tau=self.getTau(x) ; tau2=tau**2
        w=self.getWindow(tau2)
        #print "Eorig=",Eorig," tau=",tau," E=",Eorig-tau*w
        return E-tau*w,g-tau*w+2.0*w/self.sigma2*tau2

    def getEgDh(self,x):
        """ Return corrected criterion, gradient and Hessian"""
        #return self.p.getEgDh(x)
        raise "Not implemented"


# --------------------------------------------------------------------

class OptimizerMLdH:
    """ Prototype optimizer for EvaluableProblem-like classes. It uses
    Marquardt-Levenberg like method using the gradient and the diagonal of
    the Hessian. Usage:

    o=OptimizerMLdH(evaluableProblemInstance)
    while not o.hasConverged():
          o.makeStep()
    print o.getX()

    optional parameters to the constructor include the starting point
    'startx', tolerances 'abstol', 'reltol', verbosity flag 'verbose',
    lambda multiplication factor 'lambdaf', maximum and minimum
    values for lambda 'maxlambda' and 'minlambda'. Callback function 'cbf'
    can also be specified.
    """

    def __init__(self,problem,startx=None,abstol=1e-50,reltol=1e-6,
                 xtol=1e-6,verbose=0,
                 lambdaf=10.0,maxlambda=1e100,minlambda=1e-50,cbf=None):
        self.p=problem # save the problem instance
        self.lambdaf,self.maxlambda=lambdaf,maxlambda # save parameters
        self.minlambda,self.cbf=minlambda,cbf
        self.abstol,self.reltol,self.verbose=abstol,reltol,verbose
        self.xtol,self.dx=xtol, None
        self.l=1.0     # regularization factor lambda
        self.setX(startx or problem.getInitialX()) # get initial X        
        self.E,self.gr,self.H=problem.getEgDh(self.x) # most current values
        self.i=1       # the iteration number (number of evaluations)
        if self.verbose:
            print "Initial i=",self.i," lambda=",self.l," E=",self.E
        if self.cbf:
            cbf(self.E)

    def hasConvergedE(self): # return true if we have converged already
        if self.l>self.maxlambda: return 1  # lambda too big, no hope
        if not self.success: return 0        # finish only after going down
        dif=abs(self.E-self.oldE)
        return dif<=self.abstol or dif<=self.reltol*self.E

    def hasConverged(self): # convergence criterion based on the change of x
        # maximum absolute change of x must be smaller than xtol
        if self.dx is None:
            return 0 # we have not yet tried
        #print "converged crit=",bigtools.maxelement(abs(self.dx))
        return bigtools.maxelement(abs(self.dx))<=self.xtol
                               
    def setX(self,x):
        self.x=x 
        self.E,self.gr,self.H=self.p.getEgDh(self.x) # most current values
        self.success=0  # was the last step successful

    def getE(self):
        return self.E

    def getX(self):
        return self.x

    def makeStep(self): 
        # calculate next trial point and evaluate there
        self.dx=-self.gr/(self.H+self.l)
        tx=self.x+self.dx
        tE=self.p.getE(tx)
        self.i=self.i+1
        if tE<=self.E: # successful step
            self.oldE=self.E
            self.l=self.l/self.lambdaf # decrease lambda
            self.x,self.E,self.gr,self.H=(tx,)+self.p.getEgDh(tx)
            self.success=1
            if self.verbose:
                print "Success: i=",self.i," lambda=",self.l," E=",self.E
        else:
            # the 1.5 factor is a heuristic to avoid cycles
            self.l=max(1.5*self.l*self.lambdaf,self.minlambda) # increase lambda
            self.success=0 # we do not want to stop in the next step
            if self.verbose:
                print "Failure: i=",self.i," lambda=",self.l," E=",tE
        if self.cbf:
            self.cbf(self.E)


    def makeSuccessfulStep(self):
        # make steps until one is successful
        while 1:
            self.makeStep()
            if self.success or self.l>self.maxlambda:
                return

# --------------------------------------------------------------------

class OptimizerMLH:
    """ Optimizer for EvaluableProblem-like classes that support the
    'getEgH' method. Marquardt-Levenberg like method using the gradient
    and the full Hessian approximated using first derivatives. Usage:

    o=OptimizerMLH(evaluableProblemInstance)
    while not o.hasConverged():
          o.makeStep()
    print o.getX()

    optional parameters to the constructor include the starting point
    'startx', tolerances 'abstol', 'reltol', verbosity flag 'verbose',
    lambda multiplication factor 'lambdaf', maximum and minimum
    values for lambda 'maxlambda' and 'minlambda'. Callback function 'cbf'
    can also be specified.
    """

    def __init__(self,problem,startx=None,abstol=1e-50,reltol=1e-6,
                 xtol=1e-6,verbose=0,
                 lambdaf=10.0,maxlambda=1e100,minlambda=1e-50,cbf=None):
        self.p=problem # save the problem instance
        self.lambdaf,self.maxlambda=lambdaf,maxlambda # save parameters
        self.minlambda,self.cbf=minlambda,cbf
        self.abstol,self.reltol,self.verbose=abstol,reltol,verbose
        self.xtol,self.dx=xtol,None
        self.l=1.0     # regularization factor lambda
        self.setX(startx or problem.getInitialX()) # get initial X        
        self.xshape=self.x.shape
        self.i=1       # the iteration number (number of evaluations)
        if self.verbose:
            print "Initial i=",self.i," lambda=",self.l," E=",self.E
        if self.cbf:
            cbf(self.E)

    def hasConvergedE(self): # return true if we have converged already
        if self.l>self.maxlambda: return 1  # lambda too big, no hope
        if not self.success: return 0        # finish only after going down
        dif=abs(self.E-self.oldE)
        return dif<=self.abstol or dif<=self.reltol*self.E

    def hasConverged(self): # convergence criterion based on the change of x
        # maximum absolute change of x must be smaller than xtol
        if self.dx is None:
            return 0 # we have not yet tried
        #print "converged crit=",bigtools.maxelement(abs(self.dx))
        return bigtools.maxelement(abs(self.dx))<=self.xtol

    def setX(self,x):
        self.x=x 
        self.E,self.gr,self.H=self.p.getEgH(self.x) # most current values
        self.success=0  # was the last step successful


    def makeStep(self): 
        # calculate next trial point and evaluate there
        lx=reshape(self.x,(-1,))   # linearize x
        lgr=reshape(self.gr,(-1,)) # linearize gr
        sizex=lx.shape[0]
        lH=reshape(self.H,(sizex,sizex))+self.l*identity(sizex)
        self.dx=-LinearAlgebra.solve_linear_equations(lH,lgr) # solve
        tx=lx+self.dx
        tx=reshape(tx,self.xshape) # reshape back
        tE=self.p.getE(tx)
        self.i+=1
        if tE<=self.E: # successful step
            self.oldE=self.E
            self.l=self.l/self.lambdaf # decrease lambda
            self.x,self.E,self.gr,self.H=(tx,)+self.p.getEgH(tx)
            self.success=1
            if self.verbose:
                print "Success: i=",self.i," lambda=",self.l," E=",self.E
        else:
            # the 1.5 factor is a heuristic to avoid cycles
            self.l=max(1.5*self.l*self.lambdaf,self.minlambda) # increase lambda
            self.success=0 # we do not want to stop in the next step
            if self.verbose:
                print "Failure: i=",self.i," lambda=",self.l," E=",tE
        if self.cbf:
            self.cbf(self.E)


    def makeSuccessfulStep(self):
        # make steps until one is successful
        while 1:
            self.makeStep()
            if self.success or self.l>self.maxlambda:
                return


# --------------------------------------------------------------------

class OptimizerML:
    """ Prototype optimizer for EvaluableProblem-like classes. It uses
    Marquardt-Levenberg like method using the gradient but does not use the
    Hessian. Usage:

    o=OptimizerML(evaluableProblemInstance)
    while not o.hasConverged():
          o.makeStep()
    print o.getX()

    optional parameters to the constructor include the starting point
    'startx', tolerances 'abstol', 'reltol', verbosity flag 'verbose',
    lambda multiplication factor 'lambdaf', maximum and minimum
    values for lambda 'maxlambda' and 'minlambda'. Callback function 'cbf'
    can also be specified.
    """

    def __init__(self,problem,startx=None,abstol=1e-50,reltol=1e-6,verbose=0,
                 xtol=1e-6,lambdaf=10.0,maxlambda=1e100,minlambda=1e-50,cbf=None):
        self.p=problem # save the problem instance
        self.lambdaf,self.maxlambda=lambdaf,maxlambda # save parameters
        self.minlambda,self.cbf=minlambda,cbf
        self.abstol,self.reltol,self.verbose=abstol,reltol,verbose
        self.xtol,self.dx=xtol,None
        self.l=1.0     # regularization factor lambda
        self.setX(startx or problem.getInitialX()) # get initial X        
        self.E,self.gr=problem.getEg(self.x) # most current values
        self.i=1       # the iteration number (number of evaluations)
        if self.verbose:
            print "Initial i=",self.i," lambda=",self.l," E=",self.E
        if self.cbf:
            cbf(self.E)

    def hasConvergedE(self): # return true if we have converged already
        if self.l>self.maxlambda: return 1  # lambda too big, no hope
        if not self.success: return 0        # finish only after going down
        dif=abs(self.E-self.oldE)
        return dif<=self.abstol or dif<=self.reltol*self.E

    def hasConverged(self): # convergence criterion based on the change of x
        # maximum absolute change of x must be smaller than xtol
        if self.dx is None:
            return 0 # we have not yet tried
        #print "converged crit=",bigtools.maxelement(abs(self.dx))
        return bigtools.maxelement(abs(self.dx))<=self.xtol


    def setX(self,x):
        self.x=x 
        self.E,self.gr=self.p.getEg(self.x) # most current values
        self.success=0  # was the last step successful

    def getX(self):
        return self.x

    def getE(self):
        return self.E

    def makeStep(self): 
        # calculate next trial point and evaluate there
        self.dx=-self.gr*(self.E/(bigtools.sumall(self.gr)+self.l))
        tx=self.x+self.dx
        tE,tgr,tH=self.p.getEgDh(tx)
        self.i=self.i+1
        if tE<=self.E: # successful step
            self.oldE=self.E
            self.l=self.l/self.lambdaf # decrease lambda
            self.x,self.E,self.gr=tx,tE,tgr
            self.success=1
            if self.verbose:
                print "Success: i=",self.i," lambda=",self.l," E=",self.E
        else:
            # the 1.5 factor is a heuristic to avoid cycles
            self.l=max(1.5*self.l*self.lambdaf,self.minlambda) # increase lambda
            self.success=0 # we do not want to stop in the next step
            if self.verbose:
                print "Failure: i=",self.i," lambda=",self.l," E=",tE
        if self.cbf:
            self.cbf(self.E)


    def makeSuccessfulStep(self):
        # make steps until one is successful
        while 1:
            self.makeStep()
            if self.success or self.l>self.maxlambda:
                return
            

# --------------------------------------------------------------------

class OptimizerGdes:
    """ Gradient descent optimizer for EvaluableProblem-like classes.
    Just go down along the gradient, adapting the step size as needed.
    Usage:

      o=OptimizerGdes(evaluableProblemInstance)
    while not o.hasConverged():
          o.makeStep()
    print o.getX()

    optional parameters to the constructor include the starting point
    'startx', tolerances 'abstol', 'reltol', verbosity flag 'verbose'.
    
    Callback function 'cbf' can also be specified.
    """

    def __init__(self,problem,startx=None,abstol=1e-50,reltol=1e-6,verbose=0,
                 xtol=1e-6,
                 startstep=1.0,minstep=1e-50,stepf=10.0,cbf=None):
        self.p=problem # save the problem instance
        self.minstep,self.cbf=minstep,cbf
        self.abstol,self.reltol,self.verbose=abstol,reltol,verbose
        self.step,self.stepf=startstep,stepf
        self.xtol,self.dx=xtol,None
        self.setX(startx or problem.getInitialX()) # get initial X        
        # self.E,self.gr now contain the most current values
        self.i=1       # the iteration number (number of evaluations)
        if self.verbose:
            print "Initial i=",self.i," step=",self.step," E=",self.E
        if self.cbf:
            cbf(self.E)

    def hasConvergedE(self): # return true if we have converged already
        if self.step<=self.minstep: return 1
        if not self.success: return 0        # finish only after going down
        dif=abs(self.E-self.oldE)
        return dif<=self.abstol or dif<=self.reltol*self.E

    def hasConverged(self): # convergence criterion based on the change of x
        # maximum absolute change of x must be smaller than xtol
        if self.dx is None:
            return 0 # we have not yet tried
        if self.verbose>1:
            print "linf norm of dx=",bigtools.maxelement(abs(self.dx))
        return bigtools.maxelement(abs(self.dx))<=self.xtol


    def setX(self,x):
        self.x=x 
        self.E,self.gr=self.p.getEg(self.x) # most current values
        self.success=0  # was the last step successful

    def getX(self):
        return self.x

    def getE(self):
        return self.E

    def getStep(self):
        return self.step

    def setStep(self,step):
        self.step=step

    def makeStep(self): 
        # calculate next trial point and evaluate there
        self.dx=-self.gr*self.step # /sqrt(bigtools.sumall(self.gr**2))
        tx=self.x+self.dx
        tE,tgr=self.p.getEg(tx)
        self.i=self.i+1
        if tE<=self.E: # successful step
            self.oldE=self.E
            self.step=self.step*self.stepf 
            self.x,self.E,self.gr=tx,tE,tgr
            self.success=1
            if self.verbose:
                print "Success: i=",self.i," step=",self.step," E=",self.E
        else:
            # the 1.5 factor is a heuristic to avoid cycles
            self.step=max(self.step/self.stepf/1.5,
                          self.minstep) # increase lambda
            self.success=0 # we do not want to stop in the next step
            if self.verbose:
                print "Failure: i=",self.i," step=",self.step," E=",tE
        if self.cbf:
            self.cbf(self.E)


    def makeSuccessfulStep(self):
        # make steps until one is successful
        while 1:
            self.makeStep()
            if self.success or self.step<=self.minstep:
                return

    def showEnvironment(self,title=None,grad=None,step=None):
        # show the profile along the gradient
        stept=0.1 ; t=arange(-2,2+stept,stept)
        y=zeros(t.shape,Float)
        if not grad:
            grad=self.gr
        grad/=sqrt(bigtools.sumall(grad**2))
        if not step:
            step=self.step
        for i in range(t.shape[0]):
            y[i]=self.p.getE(self.x+grad*t[i]*step)
        tit=title or 'showEnvironment'
        g=bigtools.plot(t,y,title=tit)
        return g
    
# --------------------------------------------------------------------

def twoGrid(p,verb=1,presmooth=2,smooth=30,twogrid=1):
    """ Test of the multigrid concept. Solve problem p on two grids
    and return the solution.
    """

    o=OptimizerMLdH(p,verbose=verb)
    iter=0
    while iter<presmooth:
          o.makeStep()
          iter+=1
    # now go to the finer grid if needed
    if twogrid:
          print "Coarsening"
          pc,xc=p.getReduced(o.getX())
          oc=OptimizerMLdH(pc,verbose=verb,startx=array(xc)) # copy needed
          while iter<smooth:
                o.makeStep()
                iter+=1
          print "Refining"
          o=OptimizerMLdH(p,verbose=verb,
                          startx=p.updateFine(o.getX(),xc,oc.getX()))
    # continue with the optimizing
    while not o.hasConverged():
          o.makeStep()
    return o.getX()      

# --------------------------------------------------------------------

class MultigridOptimizer:
    """ A class representing a multigrid optimizer. Usage:

        m=MultigridOptimizer(OptimizerMLdH,
                             inismooth=3,presmooth=1,
                             postsmooth=2,intsmooth=2,verbose=0,
                             abstol=1e-50,reltol=1e-5)
        x=m.optimizeFMG(problem)
    """

    def __init__(self,optimizer,presmooth=3,postsmooth=3,
                 intsmooth=2,verbose=0,maxiter=100,abstol=1e-50,reltol=1e-5
                 ):
        self.optcls,self.presm=optimizer,presmooth
        self.postsm,self.intsm,self.verbose=postsmooth,intsmooth,verbose
        self.abstol,self.reltol=abstol,reltol
        self.failedlevel=10000

    def makeTcycle(self,p,x,level=0,callbackE=None):
        """ Returns a value improved by a multigrid T cycle
            Its main characteristics are not to iterate at fine level
            when iteration at coarse level improved also the fine level
            criterion.
        """
        l=level ; cbE=callbackE
        o=self.optcls(p,startx=x,verbose=self.verbose>4,
                      reltol=self.reltol) # create optimizer
        if self.verbose>1:
            print l," Entering makeTcycle E=",o.getE(), " maxlevel=",ml
        if callbackE: callbackE(l,o.getE())
        try:  # if supported, set the step size from the last iteration
            o.setStep(self.laststep)
        except:  pass
        

    def makeVcycle(self,p,x,level=0,callbackE=None,maxlevel=1000,
                   parentopt=None):
        """ Returns a value improved by a multigrid V cycle """
        l,ml=level,maxlevel ; cbE=callbackE
        o=self.optcls(p,startx=x,verbose=self.verbose>4,
                      reltol=self.reltol) # create optimizer
        if self.verbose>1:
            print l," Entering makeVcycle E=",o.getE(), " maxlevel=",ml
        if callbackE:
            callbackE(l,o.getE())

        if 0: # parentopt and p.doCorrect:
            gr=parentopt.gr
            if gr.shape!=x.shape:
                gr=p.reduceX(gr)
            g1=o.showEnvironment(title='coarse level',
                                 grad=gr,step=parentopt.step)
            g2=parentopt.showEnvironment(title='fine level')
            raw_input('showEnvironment: press any key')

        try:  # if supported, set the step size from the last iteration
            o.setStep(self.laststep)
            #print "Step set successfully to ",o.getStep()
        except:  pass

        # presmoothing    
        for i in range(self.presm):
            o.makeStep()
            if callbackE:
                callbackE(l,o.getE())
        E0=o.getE()    # criterion before coarse grid correction
        if self.verbose>4:
            print l," Before coarse grid correction E0=",E0


        #gn=Gnuplot.Gnuplot(); #p.getE(x,visualize=1,gnupl=gn)
        shouldReduce=l<maxlevel and p.canReduce() # not the coarsest level    
        if shouldReduce: 
            pc,xc0=p.getReduced(o.getX()) # get reduced version
            xc1=self.makeVcycle(pc,array(xc0),level=l+1,callbackE=cbE,
                                maxlevel=ml,parentopt=o) # recursive

            if 0: # show the coarse->fine transition
                p.getE(xc1)
                bigtools.plot(arange(p.w.refimg.shape[0]),p.w.refimg,title='reference fine',gnupl=gn)
                bigtools.plot(arange(p.w.wt.shape[0]),p.w.wt,title='warped fine',gnupl=gn)
                pc.getE(xc1)
                bigtools.plot(arange(pc.w.refimg.shape[0])*2.0,pc.w.refimg,
                          title='reference coarse',gnupl=gn)
                bigtools.plot(arange(pc.w.wt.shape[0])*2.0,pc.w.wt,title='warped coarse',gnupl=gn)
                raw_input("Press any key - makeVcycle")
            
            
            x=p.updateFine(x,xc0,xc1) # update fine grid solution
            # E1=p.getE(x,visualize=1,gnupl=gn)
            E1=p.getE(x) # criterion after coarse grid correction
            if E1<E0:                 # criterion has improved
                if self.verbose>3:
                    print l," Coarse grid correction helped E1=",E1, " E0=",E0
                o.setX(x) # update x
            else:
                self.failedlevel=l # remember so that we do not go up any more
                if self.verbose>3:
                    print l," Coarse grid correction failed E1=",E1," E0=",E0
        # postsmoothing
        if callbackE:
            callbackE(l,o.getE())

        if self.postsm>=0: # and shouldReduce:
            # iterate certain number of times unless at the coarsest level
            for i in range(self.postsm):
                o.makeStep()
                if callbackE:
                    callbackE(l,o.getE())

        else:            # iterate until convergence
            while not o.hasConverged():
                o.makeSuccessfulStep()
                if callbackE:
                    callbackE(l,o.getE())

        if self.verbose>3:
            print l," After postsmoothing E",o.getE()
        #p.getE(o.getX(),visualize=1,gnupl=gn)
        try: self.laststep=o.getStep()
        except: pass
        return o.getX()

    def makeWcycle(self,p,x,level=0,callbackE=None):
        """ Returns a value of x improved by a multigrid W cycle """
        l=level ; cbE=callbackE
        o=self.optcls(p,startx=x,reltol=self.reltol,
                      verbose=self.verbose>3) # create optimizer
        if self.verbose>1:
            print level," Entering makeWcycle E=",o.getE()
        if callbackE:
            callbackE(l,o.getE())
        # presmoothing    
        for i in range(self.presm):
            o.makeSuccessfulStep()
            if callbackE:
                callbackE(l,o.getE())
        E0=o.getE()    # criterion before coarse grid correction
        if self.verbose>3:
            print level," Before coarse grid correction E0=",E0
        if p.canReduce(): # not the coarsest level    
            pc,xc0=p.getReduced(o.getX()) # get reduced version
            xc1=array(xc0)
            for i in range(self.intsm):
                if self.verbose>3:
                    print level," Internal W call",E0
                xc1=self.makeWcycle(pc,xc1,level=l+1,callbackE=cbE) # recursive 
            x=p.updateFine(x,xc0,xc1) # update fine grid solution
            E1=p.getE(x)              # criterion after coarse grid correction
            if E1<E0:                 # criterion has improved
                if self.verbose>3:
                    print level," Coarse grid correction helped E1=",E1," E0=",E0
                o.setX(x) # update x
            else:
                if self.verbose>3:
                    print level," Coarse grid correction failed E1=",E1," E0=",E0
        # postsmoothing
        if callbackE:
            callbackE(l,o.getE())
        
        if self.postsm>=0:  # iterate certain number of times
            for i in range(self.postsm):
                o.makeSuccessfulStep()
                if callbackE:
                    callbackE(l,o.getE())

        else:            # iterate until convergence
            while not o.hasConverged():
                o.makeSucessfulStep()
                if callbackE:
                    callbackE(l,o.getE())

        if self.verbose>3:
            print level, " After postsmoothing E",o.getE()
        return o.getX()

    def optimizeFMG(self,p,level=0,maxiter=500,callbackE=None):
        """ Full multigrid optimizer of problem p. Returns the optimal value
            x found."""
        cbE,maxit=callbackE,maxiter
        if self.verbose:
            print level, "Entering optimizeFMG"
        x=p.getInitialX() ; l=level
        self.failedlevel=10000
        # The initialization part
        if p.canReduce(): # not the coarsest level
            pc,xc0=p.getReduced(x)
            # recursively get initial estimate
            xc1=self.optimizeFMG(pc,level=l+1,maxiter=maxit,
                                 callbackE=cbE) 
            x=p.updateFine(x,xc0,xc1)
        else: # the coarsest level - iterate until convergence and return
            o=self.optcls(p,startx=x,verbose=self.verbose>4,
                      reltol=self.reltol) # create optimizer
            while not o.hasConverged():
                o.makeSuccessfulStep()
                if callbackE:
                    callbackE(l,o.getE())
            return x
        
        # now just smooth until convergence or maxiter
        E0=p.getE(x) ; iter=0
        if self.verbose>1:
            print l," optimizeFMG iter=",iter," E=",E0
            
        try: # from now on, start Boumann correcting, if supported
            p.doCorrect=1 
        except AttributeError: pass

        while 1:
            x=self.makeVcycle(p,x,level=l,
#                         callbackE=cbE,maxlevel=min(l+2,self.failedlevel))
                          callbackE=cbE,maxlevel=l+2)
            E=p.getE(x) ; iter+=1
            if self.verbose>1:
                print l," optimizeFMG iter=",iter," E=",E
            difE=abs(E-E0)    
            if iter>=maxiter or difE<self.abstol or difE<E*self.reltol:
                break
            E0=E 
        return x    
        



# --------------------------------------------------------------------


class OptimizerGdesP:
    """ Gradient descent optimizer for EvaluableProblem-like classes.
    Just go down along the gradient, adapting the step size as needed.

    It works almost exactly as bigoptimize.OptimizerGdes, except that
    parameters are given as a parameter object
    Usage:

    o=OptimizerGdesP(evaluableProblemInstance,parameters=None)
    while not o.hasConverged():
          o.makeStep()
    print o.getX()

    optional parameters include the starting point
    'startx', tolerances 'xtol', starting step 'startstep',
    step mult. factors 'stepf' and 'stepdownf', verbosity flag 'verbose'.
    
    Callback function 'cbf' can also be specified, which is then
    called whenever the criterion is evaluated.

    Some optimization has been performed to minimize the number
    of evaluations needed. 
    """

    def __init__(self,problem,par=None): # par are the Parameters
        self.p=problem
        self.par=bigtools.Parameters(par,
           default={ 'xtol':1e-3, 'verbose':0, 'minstep':1e-20,
                     'stepf': 9, 'stepdownf':10, 'startstep': 1.0 })
        self.i=0       # the iteration number (number of evaluations)
        self.x=None ; self.E=None ; self.dx=None ; self.gr=None
        self.step=self.par.startstep

    def hasConverged(self): # convergence criterion based on the change of x
        # maximum absolute change of x must be smaller than xtol
        if self.dx is None:
            return 0 # we have not yet tried
        if self.par.verbose>7:
            print "linf norm of dx=",bigtools.maxelement(abs(self.dx))
        return bigtools.maxelement(abs(self.dx))<=self.par.xtol

    def setX(self,x):
        self.x=x 
        self.E=None

    def getX(self):
        if self.x is None:
            if self.par.isdef('startx'):
                self.x=self.par.startx
            else:
                self.x=self.p.getInitialX() 
        return self.x

    def getEg(self):
        if self.E is None or self.gr is None:
            self.E,self.gr=self.p.getEg(self.getX())
            self.i+=1
        if self.par.isdef('cbf'):
            self.par.cbf(self.E,self.getX())
        return self.E, self.gr

    def getE(self):
        if self.E is None:
            self.E=self.p.getE(self.getX())
            self.i+=1
        if self.par.isdef('cbf'):
            self.par.cbf(self.E,self.getX())
        return self.E

    def makeStep(self):
        E,gr=self.getEg() # recover old ones if not known already
        self.dx=-gr*self.step
        if self.par.verbose>8:
                print "makeStep: i=%d step=%.3g dx=%.3g E=%.3g" % (self.i,
              self.step,bigtools.maxelement(abs(self.dx)),E)
        tx=self.x+self.dx
        tE,tgr=self.p.getEg(tx)
        self.i+=1
        if tE<=E: # successful step
            self.step*=self.par.stepf 
            self.x,self.E,self.gr=tx,tE,tgr
            self.success=1
            if self.par.verbose>5:
                print "Success: i=",self.i," step=",self.step," E=",self.E
        else:
            self.step=self.step/self.par.stepdownf # decrease step
            self.success=0
            if self.par.verbose>5:
                print "Failure: i=",self.i," step=",self.step," E=",tE
  
    def makeSuccessfulStep(self):
        # make steps until one is successful or until step-size drops
        # until a predefined threshold
        while 1:
            self.makeStep()
            if self.success or self.step<=self.par.minstep:
                return

    def getIterCount(self):
        return self.i

    def smoothToConvergence(self):
        """ Iterate until convergence is reached. By virtue of the
        stopping criterion, the last evaluation is always the best """
        while not self.hasConverged():
            self.makeStep()
        return self


def test_optimizerGdes(p=None):
    import Gnuplot, bigoptimize
    if p is None:
        p=bigoptimize.EvaluableProblem()
    listE=[]
    def cbfun(E,x,list=listE):
        list.append(E)
    par=bigtools.Parameters({'verbose':10, 'cbf': cbfun, 'xtol': 1e-10})
    o=OptimizerGdesP(p,par=par)
    g=Gnuplot.Gnuplot()
    while not o.hasConverged():
        o.makeStep()
    x=o.getX() ;  print x
    z=array(listE)
    g=bigtools.plot(z,with='lines',cmd='set logscale y')
    g.title('Gdes convergence')
    #g.hardcopy('mlconv.eps',enhanced=1)
    raw_input('Press any key')

# ----------------------------------------------------------------------

class OptimizerGdesQ(OptimizerGdesP):
    """ Adds automatic step-size calculation to OptimizerGdesP """
    def makeStep(self):
        E,gr=self.getEg() # recover old ones if not known already
        self.dx=-gr*self.step
        if self.par.verbose>8:
                print "makeStep: i=%d step=%.3g dx=%.3g E=%.3g" % (self.i,
              self.step,bigtools.maxelement(abs(self.dx)),E)
        tx=self.x+self.dx
        tE,tgr=self.p.getEg(tx)
        self.i+=1
        if tE<=E: # successful step
            # Parabolic approximation
            try:
                candstep=0.5*(bigtools.sumall(self.dx*self.dx)/
                           (E-tE+bigtools.sumall(self.dx*tgr)))
            except ZeroDivisionError:
                print "division by zero calculating candstep"
                candstep=-1.0
            if candstep<=0.0:
                self.step*=self.par.stepf
            else:
                self.step=candstep
            self.x,self.E,self.gr=tx,tE,tgr
            self.success=1
            if self.par.verbose>5:
                print "Success: i=%d step=%.3g dx=%.4g E=%.6g" % (self.i,
              self.step,bigtools.maxelement(abs(self.dx)),self.E)
        else:
            self.step=self.step/self.par.stepdownf # decrease step
            self.success=0
            if self.par.verbose>5:
                print "Failure: i=%d step=%.3g dx=%.4g E=%.6g" % (self.i,
              self.step,bigtools.maxelement(abs(self.dx)),tE)
  
# --------------------------------------------------------------------

class MROptimizerState:
    """ Encapsulates a state of a one level of a multiresolution optimizer.
    After instantiation, solveMR() is called to solve the problem.
    Then, the final state of the problem can be queried using
    getProblem(), or getOptimizer().

    New parameters include: optimizerClass (defaults to OptimizerGdesQ)
    and minlevel, minimum level at which we optimize (default is 0, optimize
    at all levels). If xtollast is set, it replaces xtol for level 0.
    maxlevel is the maximum level at which we descend.
    """

    def __init__(self,problem,par=None,level=0):
        """ Accepts an instance of an EvaluablemProblem problem,
        implementing the Multigridable protocol (functions canReduce(),
        getReduced(), and updateFine() )"""
        self.par=bigtools.Parameters(par,
               default={'optimizerClass': OptimizerGdesQ,
                        'minlevel':0, 'prcbf': None, 'xtollast':None,
                        'maxlevel':1000,'verbose':3 })
        self.p=problem
        # instantiate the optimizer
        if level==0 and self.par.xtollast is not None:
            paro=bigtools.Parameters(par,override={'xtol':self.par.xtollast})
        else:
            paro=self.par
        self.o=self.par.optimizerClass(self.p,paro)
        self.level=level
        if self.par.prcbf is not None:
            self.par.prcbf(self.p)

    def getProblem(self):
        return self.p

    def getOptimizer(self):
        return self.o

    def canReduce(self):
        return self.p.canReduce()

    def getReduced(self):
        """ Returns a reduced version of itself, passing the current
        optimization state along """
        rp=self.p.getReduced() # reduced problem
        # lower tolerance
        # NOTE: this assumes the absolute values of the solution
        # are also reduced
        # par=bigtools.Parameters(self.par,override={'xtol': self.par.xtol/2.0})
        return MROptimizerState(rp,self.par,level=self.level+1)

    def smoothToConvergence(self):
        """ Optimizes until convergence is reached.
            Returns a reference to self.
        """
        self.o.smoothToConvergence()
        return self

    def solveMR(self):
        """ Finds a solution to the problem using multiresolution approach.
        Returns a reference to self.
        """
        if self.par.verbose>3: time1=time.clock()
        if self.par.verbose>4:
            origE=self.o.getE()
        if self.level<self.par.maxlevel and self.canReduce():
            r=self.getReduced()           # get a reduced version of myself
            xc0=r.p.getXcopy()            # save the initial coarse level x
            r.solveMR()                   # solve at coarse level
            self.p.updateFine(xc0,r.p)    # update yourself
            self.o.setX(self.p.getInitialX()) # update optimizer state
        if self.par.verbose>4:
            befE=self.o.getE()
        if self.level>=self.par.minlevel:
            self.smoothToConvergence()        # iterate at the fine level
        if self.par.verbose>4:
            print "solveMR at level %d E=%.3g > %.3g > %.3g " % (self.level,
                  origE,befE,self.o.getE())," iter: ",self.o.getIterCount()
        if self.par.verbose>3:
            print "solveMR at level ", self.level, " took ", \
                  time.clock()-time1, " s"
        return self

# --------------------------------------------------------------------

import bigregister
import bigmultigrid

def testVcycle(p=None):
    #p=MultigridableProblem(xshape=(16,))
    #import bigregister
    if not p:
        #p=CorrectedMultigridableProblem(MultigridableProblem(xshape=(16,)))
        p=MultigridableProblem(xshape=(32,))
        p=CorrectedMultigridableProblem(p)
    #m=MultigridOptimizer(OptimizerMLdH,verbose=4)
    m=MultigridOptimizer(OptimizerMLdH,verbose=4)
    x=p.getInitialX()
    listEl=[]
    def cbfun(E,l,list=listEl):
        list.append((E,l))
    for i in range(5):
        x=m.makeVcycle(p,x,callbackE=cbfun)
    z=array(listEl)
    t=arange(z.shape[0])
    g=Gnuplot.Gnuplot()
    d1=Gnuplot.Data(t,z[:,0]*1.0+8,with='lines',title='level') # level data
    d2=Gnuplot.Data(t,z[:,1],with='lines',title='criterion')
    g.plot(d1,d2)
    # g.hardcopy('vcycleres.eps',enhanced=1)
    raw_input('Press any key')
    g=bigtools.plot(x)
    g.title('Solution')
    # g.hardcopy('harmonicsol.eps',enhanced=1)
    raw_input('Press any key')


def testMR(p=None):
    if not p:
        p=MultigridableProblem(xshape=(32,))
    #    m=MultigridOptimizer(OptimizerMLdH,verbose=4,presmooth=0,
    #                         postsmooth=-1,reltol=1e-6)
    m=MultigridOptimizer(OptimizerGdes,verbose=5,presmooth=0,
                             postsmooth=-1,reltol=1e-4)
    #m=MultigridOptimizer(OptimizerMLdH,verbose=5,presmooth=0,
    #                         postsmooth=-1,reltol=1e-4)
    x=p.getInitialX()
    listEl=[]
    def cbfun(E,l,list=listEl):
        list.append((E,l))
    x=m.makeVcycle(p,x,callbackE=cbfun)
    z=array(listEl)
    t=arange(z.shape[0])
    g=Gnuplot.Gnuplot()
    #d1=Gnuplot.Data(t,z[:,0]*0.2+1,with='lines',title='level') # level data
    d2=Gnuplot.Data(t,z[:,1],with='lines',title='criterion')
    g('set logscale y')
    g.plot(d2)
    g.title('Multiresolution')
    g.hardcopy('mrres.eps',enhanced=1)
    raw_input('Press any key')
    return x

def testWcycle():
    #p=MultigridableProblem(xshape=(16,))
    p=CorrectedMultigridableProblem(MultigridableProblem(xshape=(16,)))
    m=MultigridOptimizer(OptimizerMLdH,verbose=2)
    x=p.getInitialX()
    listEl=[]
    def cbfun(E,l,list=listEl):
        list.append((E,l))
    for i in range(5):
        x=m.makeWcycle(p,x,callbackE=cbfun)
    z=array(listEl)
    t=arange(z.shape[0])
    g=Gnuplot.Gnuplot()
    d1=Gnuplot.Data(t,z[:,0]*1+8,with='lines',title='level') # level data
    d2=Gnuplot.Data(t,z[:,1],with='lines',title='criterion')
    g.plot(d1,d2)
    # g.hardcopy('wcycleres.eps',enhanced=1)
    raw_input('Press any key')
    g=bigtools.plot(x)
    g.title('Solution')
    # g.hardcopy('harmonicsol.eps',enhanced=1)
    raw_input('Press any key')



#def testWcycle():
#    p=MultigridableProblem(xshape=(32,))
#    m=MultigridOptimizer(OptimizerMLdH,verbose=1)
#    x=p.getInitialX()
#    x=m.makeWcycle(p,x)

def testFMG(p=None):
    #p=MultigridableProblem(xshape=(320,))
    if not p:
        #p=CorrectedMultigridableProblem(MultigridableProblem(xshape=(32,)))
        p=MultigridableProblem(xshape=(32,))
        p=CorrectedMultigridableProblem(p)
    # m=MultigridOptimizer(OptimizerMLdH,verbose=2,reltol=1e-3)
    # m=MultigridOptimizer(OptimizerMLdH,verbose=3,reltol=1e-3)
    #opt=OptimizerMLdH
    opt=OptimizerGdes
    m=MultigridOptimizer(opt,verbose=4,presmooth=0,
                             postsmooth=5,reltol=1e-4)

    listEl=[]
    def cbfun(E,l,list=listEl):
        list.append((E,l))
    x=m.optimizeFMG(p,callbackE=cbfun)
    z=array(listEl)
    t=arange(z.shape[0])
    g=Gnuplot.Gnuplot()
    g1=Gnuplot.Gnuplot()
    d1=Gnuplot.Data(t,z[:,0],with='lines',title='level') # level data
    g('set logscale y')
    d2=Gnuplot.Data(t,z[:,1],with='lines',title='criterion')
    #g.plot(d1,d2)
    g.plot(d2) ; g1.plot(d1)
    # g.hardcopy('fmgres.eps',enhanced=1)
    raw_input('Press any key')
    #g1=bigtools.plot(x)
    #g1.title('Solution')
    # g1.hardcopy('harmonicsol.eps',enhanced=1)
    #raw_input('Press any key')
    return x

def test_twoGrid():
    t1=time.time()
    # x=twoGrid(MultigridableProblem(xshape=(32,)),twogrid=0)
    x=twoGrid(MultigridableProblem(xshape=(32,)),verb=0,twogrid=0)
    t2=time.time()
    print "twoGrid took ",t2-t1, " s"
    g=bigtools.plot(x)
    raw_input('Press any key')


def test_optimizerMLdH(p=None,x=None,progress=0,solution=0):
    #o=OptimizerMLdH(EvaluableProblem(),verbose=1)
    listE=[]
    def cbfun(E,list=listE):
        list.append(E)
    if not p:
        p=MultigridableProblem(xshape=(320,),T=2.0)
    o=OptimizerMLdH(p,verbose=1,reltol=1e-6,cbf=cbfun,startx=x)
    if progress: g=Gnuplot.Gnuplot()
    while not o.hasConverged():
          o.makeStep()
          if progress:
              x=o.getX()[0]
              d1=Gnuplot.Data(arange(x.shape[0]),x,with='lines')
              g.plot(d1)
              raw_input('Press any key')
    x=o.getX() ; # print x
    if solution:
        gx=bigtools.plot(x)
        gx.title('Solution')
    # g.hardcopy('harmonicsol.eps',enhanced=1)
    #raw_input('Press any key')
    z=array(listE)
    g=bigtools.plot(z,with='lines',cmd='set logscale y')
    g.title('ML convergence')
    #g.hardcopy('mlconv.eps',enhanced=1)
    raw_input('Press any key')
    return x

def test_optimizerGdes(p=None,x=None,progress=0,solution=0):
    #o=OptimizerMLdH(EvaluableProblem(),verbose=1)
    listE=[]
    def cbfun(E,list=listE):
        list.append(E)
    if not p:
        p=MultigridableProblem(xshape=(320,),T=2.0)
    o=OptimizerGdes(p,verbose=1,reltol=1e-6,cbf=cbfun,xtol=1e-3,startx=x)
    if progress: g=Gnuplot.Gnuplot()
    while not o.hasConverged():
          o.makeStep()
          if progress:
              x=o.getX()[0]
              d1=Gnuplot.Data(arange(x.shape[0]),x,with='lines')
              g.plot(d1)
              raw_input('Press any key')
    x=o.getX() ; # print x
    if solution:
        gx=bigtools.plot(x)
        gx.title('Solution')
    # g.hardcopy('harmonicsol.eps',enhanced=1)
    #raw_input('Press any key')
    z=array(listE)
    g=bigtools.plot(z,with='lines',cmd='set logscale y')
    g.title('Gdes convergence')
    #g.hardcopy('mlconv.eps',enhanced=1)
    raw_input('Press any key')
    return x

def test_optimizerML(p=None,x=None,progress=0,solution=0):
    #o=OptimizerMLdH(EvaluableProblem(),verbose=1)
    listE=[]
    def cbfun(E,list=listE):
        list.append(E)
    if not p:
        p=MultigridableProblem(xshape=(320,),T=2.0)
    o=OptimizerML(p,verbose=1,reltol=1e-12,cbf=cbfun,startx=x)
    if progress: g=Gnuplot.Gnuplot()
    while not o.hasConverged():
          o.makeStep()
          if progress:
              x=o.getX()[0]
              d1=Gnuplot.Data(arange(x.shape[0]),x,with='lines')
              g.plot(d1)
              raw_input('Press any key')
    x=o.getX() ; # print x
    if solution:
        gx=bigtools.plot(x)
        gx.title('Solution')
    # g.hardcopy('harmonicsol.eps',enhanced=1)
    #raw_input('Press any key')
    z=array(listE)
    g=bigtools.plot(z,with='lines',cmd='set logscale y')
    g.title('ML convergence')
    #g.hardcopy('mlconv.eps',enhanced=1)
    raw_input('Press any key')
    return x

    
def test_evproblem(p,x=None,h=1e-6):
    import sys
    sys.float_output_precision=3
    #sys.float_output_suppress_small=1
    if not x: x=p.getInitialX()
    print "Criterion=",p.getE(x)
    print "Gradient analytic=",p.getEg(x)[1]
    print "Gradient numeric=",p.numg(x,h)
    print "Gradient evaluation error=", p.gerr(x,h)
    #print "Hessian analytic=",p.getEgDh(x)[2]
    #print "Hessian numeric=",p.numDh(x,h)
    #print "Hessian evaluation error=", p.dHerr(x,h)

def compare_optimizers(p=None):
    class mylist:
        l=[]
    listE=mylist()
    def cbfun(E,list=listE):
            list.l.append(E)
    def evalopt(opt,list=listE):
        list.l=[]
        t1=time.clock()
        while not opt.hasConverged():
            opt.makeStep()
        t2=time.clock()
        print "Optimization took ",t2-t1, "s and ",opt.i, " iterations, E=",\
              opt.E
        return array(list.l)

    if 0:
           degf=3 ; degc=3
           N=200
           xref=fromfunction(lambda x,NN=float(N): exp(-((x/NN-0.5)*10)**2),(N,))
           xtest=fromfunction(lambda x,NN=float(N): exp(-((x/NN-0.6)*10)**2)+
                              0.01,(N,))
           xtestc=bigsplines.mfspline(xtest,degf)
           x0=zeros((1,16),Float)
           p=bigregister.WarpingProblem(xref,xtestc,x0.shape,alpha=0.0)

    if not p:
        p=bigmultigrid.getWarpedLena()[0]
        
    
    xt=1e-3
    zMLdH=evalopt(OptimizerMLdH(p,verbose=0,xtol=xt,cbf=cbfun))
    zMLH=evalopt(OptimizerMLH(p,verbose=0,xtol=xt,cbf=cbfun))
    zGdes=evalopt(OptimizerGdes(p,verbose=0,xtol=xt,cbf=cbfun))
    g=Gnuplot.Gnuplot()
    # g('set logscale y')
    ln=min([zMLdH.shape[0],zMLH.shape[0],zGdes.shape[0]])
    d1=Gnuplot.Data(arange(ln),zMLdH[0:ln],with='lines',title='MLdH')
    d2=Gnuplot.Data(arange(ln),zMLH[0:ln],with='lines',title='MLH')
    d3=Gnuplot.Data(arange(ln),zGdes[0:ln],with='lines',title='Gdes')
    g.plot(d1,d2,d3)
    #g.plot(d1)
    g.hardcopy('compare_optimizers.eps',enhanced=1)
    raw_input('Criteria plotted. Press any key')


if __name__=='__main__': # running interactively
    test_optimizerMLdH()
    # test_evproblem(EvaluableProblem())
    # test_evproblem(HarmonicProblem(xshape=(3,)),x=array((1,3,3)))
