""" Module bigregister contains functions for performing
    registration between two images representing the deformation using
    B-splines

    Jan Kybic, August-November 2000
    $Id: bigregister.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
"""

import bigsplines     
from Numeric import *
import Image
#import NumTut
import bigoptimize
import bigtools
import BIGsplines
import Gnuplot
import time
import bigmultigrid
import bigpyramid
import ImageFilter

# ------------------------------------------------------------

class WarpableImage:
    """ contains an n-dimensional image which can be warped
        if mask is defined, only pixels set in the mask will be considered
        """

    # f ... image to be warped
    # degf ... spline degree for the image
    # degc ... spline degree for the deformation
    # b ... Bspline coefficients
    # g ... deformation table
    # w ... warped image
    c=None ; g=None
    w=None 
    
    def __init__(self,image,degf=3,degc=2,imageiscoefs=0,verbose=0,mask=None):
        self.f=image ; self.degf=degf ; self.degc=degc
        self.verbose,self.mask=verbose,mask
        self.dim=len(image.shape)
        if imageiscoefs:
            self.b=image
        else:
            self.b=bigsplines.mfspline(image,degf)    

    def setIdentity(self,shape):   
        """ initializes deformation as identity with a given shape
            the shape should be (n,x1,x2,..,xn) """
        if shape[0]!=self.dim or len(shape)!=self.dim+1:
            raise ValueError, "incorrect shape"
        self.c=zeros(shape,Float)
        self.g=None 
        self.w=None

    def getDeformation(self):
        if self.g is None:
            if self.c is None:
                raise AttributeError, "deformation not initialized"
            
            tblf=indices(self.f.shape,'d')
            self.steps=((array(self.c.shape[1:])-1.0)/(array(self.f.shape)-1.0))
            d=[self.dim]+self.dim*[1]
            tbl=tblf * reshape(self.steps,d)
            self.tbl=tbl
            if self.verbose:
                 print "getting table"
            self.g=tblf # begin with identity
            for i in range(self.dim): # for all axes
                 self.g[i]=(self.g[i]+self.interpoldef(i,self.mask))
            if self.verbose:
                 print "done"
        return self.g

    def getDeformationDerivative(self,flag=0):
        if self.c is None:
            raise AttributeError, "deformation not initialized"
            
        tblf=indices(self.f.shape,'d')
        self.steps=((array(self.c.shape[1:])-1.0)/(array(self.f.shape)-1.0))
        d=[self.dim]+self.dim*[1]
        tbl=tblf * reshape(self.steps,d)
        self.tbl=tbl
        if self.verbose:
            print "getting table"
        self.g=zeros(tblf.shape,Float) # begin with identity
        for i in range(self.dim): # for all axes
            self.g[i]=self.interpoldef(i,self.mask,flag)
        if self.verbose:
            print "done"
        return self.g

    def setDeformation(self, tbl):
        self.g=tbl
        self.w=None
        self.c=None  #This is not extrictly necessary.
        
    def getDisplacementAtPoint(self,point,flag=0):
        """ Gets displacement at 'point' """
        if self.c is None:
            raise AttributeError, "deformation not initialized"
        self.steps=((array(self.c.shape[1:])-1.0)/(array(self.f.shape)-1.0))
        d=[self.dim]+self.dim*[1]
        point=asarray(point,Float)
        point=reshape(point*self.steps,d)
        displ=zeros(self.dim,Float) # resulting displacement
        for i in range(self.dim): # for all axes
           displ[i]=bigsplines.splineinterpol(self.c[i],point,self.degc,flag).flat[0]
        return displ

    def getDisplacementAtPoints(self,points,flag=0):
        """ Gets displacement at multiple 'points' of size (n,dim),
            (which is the transpose of what interpoldef wants)
            It returns the diplacements in the same format (n,dim)"""
        if self.c is None:
            raise AttributeError, "deformation not initialized"
        self.steps=((array(self.c.shape[1:])-1.0)/(array(self.f.shape)-1.0))
        d=[self.dim]+(self.dim-1)*[1]+[-1]
        npoints=points.shape[0]
        points=asarray(transpose(points),Float)
        points=reshape(points,d)*reshape(self.steps,d)
        displ=zeros(points.shape,Float) # resulting displacement
        for i in range(self.dim): # for all axes
           displ[i]=reshape(
               bigsplines.splineinterpol(self.c[i],points,self.degc,flag),
               displ.shape[1:])
        return transpose(reshape(displ,(self.dim,npoints)))


    def interpoldef(self,i,m,flag=0):
        if self.verbose:
            time1=time.clock()
        r=bigsplines.splineinterpol(self.c[i],self.tbl,self.degc,
                                    mask=m,flag=flag)
        if self.verbose:
            print "interpoldef took ",time.clock()-time1, " s"
            if m is None: print "m=None"
            else: print "mask=",bigtools.sumall(m)
            
        return r

    def getC(self):
        """ returns a reference to the c coefficient matrix describing
            the deformation. The matrix should not be modified """
        return self.c

    def setC(self,c):
        """ remembers the reference to a matrix c, which should not be
            modified while needed. """
        self.g=None 
        self.w=None
        self.c=c
        return self.c


    def getWarped(self):
        if not(self.w):
            g=self.getDeformation()
            if self.verbose:
                print "deforming"
                time1=time.clock()
            self.w=bigsplines.splineinterpol(self.b,self.g,self.degf,
                                             mask=self.mask)
            if self.verbose:
                print "deformation in getWarped took ",time.clock()-time1, " s"
                print "done"
        return self.w    

# ------------------------------------------------------------

class WarpingRegularizer:
    """ A helper class for WarpingProblem. Evaluates the regularization
    part of the criterion. """

    def initRegularizer(self):
        self.rk=array((-1.0,2.0,-1.0)) ; self.rkorig=1
        self.rk2=convolve(self.rk,self.rk,2) ; self.rk2orig=2

    def getEr(self,c):
        def filtfun(what,kernel=self.rk,kernelorig=self.rkorig):
                f=convolve(what,kernel,2)
                return bigtools.sumall(f**2)*what.shape[0]
        E=0.0
        for i in range(c.shape[0]): # for all components
            E+=bigtools.sumforalldims(c[i],
                   lambda x,oper=filtfun: bigtools.sumforallrows(x,oper))
        E/=multiply.reduce(c.shape) # normalization
        #print "getEr=",E
        return E

    def getEgr(self,c):
        def filtfun2(what,kernel=self.rk2,
                    kernelorig=self.rk2orig):
            # the chopping might need a revision for other filters
            f=convolve(what,kernel,2)[kernelorig:-kernelorig]
            return (2.0*what.shape[0])*f
        E=self.getEr(c)
        gr=zeros(c.shape,Float)
        for i in range(c.shape[0]): # for all components
            gr[i]=bigtools.vectorsumforalldims(c[i],filtfun2)
        gr/=multiply.reduce(c.shape) # normalization
        return E,gr

    def getEgDhr(self,c):
        E,gr=self.getEgr(c)
        H=(2.0*self.rk2[self.rk2orig]*bigtools.sumall(c.shape[1:]))*ones(c.shape)
        H/=multiply.reduce(c.shape) # normalization
        return E,gr,H
    


# ------------------------------------------------------------

class ZeroRegularizer:
    """ A helper class for WarpingProblem. Evaluates the regularization
    part of the criterion as zero. """

    def initRegularizer(self):
        pass

    def getEr(self,c):
        return 0.0

    def getEgr(self,c):
        return 0.0,zeros(c.shape)

    def getEgDhr(self,c):
        return 0.0,zeros(c.shape),zeros(c.shape)
    


# ------------------------------------------------------------


#class WarpingProblem(bigoptimize.EvaluableProblem,WarpingRegularizer):
class WarpingProblem(bigoptimize.EvaluableProblem,ZeroRegularizer):
    """ Accepts two images, reference and test, the test image in the
    form of  its B-spline coefficients, as well as the suggested size
    of the coefficients c describing the deformation and the degree
    of B-splines to represent the deformation. Evaluates the SSD
    criterion and its first and second derivatives.
    
    Alpha is the factor in the regularization part E=Ed+alpha Er
    where Er is whatever is returned by WarpingRegularizer

    The coefficients c are normalized, i.e. multiplied by the size of the
    image in the specific direction. This way, even if the image size
    changes, the appropriate c coefficients will not.

    If mask is defined, only pixels in the mask will be taken into account

    implements the interface of EvaluableProblem
    """

    def __init__(self,refimg,testimgc,csize=None,degf=3,degc=3,
                 alpha=0.0,verbose=0,mask=None):
        self.refimg,self.testimgc=refimg,testimgc
        self.degc,self.degf,self.verbose=degc,degf,verbose
        self.alpha=alpha # regularization factor
        self.dim=len(refimg.shape)
        self.mask=mask
        # suggested size of coefs is 4 for each dimension
        self.csize= csize or array([self.dim]+[4]*self.dim,Int) 
        self.w=WarpableImage(self.testimgc,degf,degc,imageiscoefs=1,
                             verbose=self.verbose,mask=self.mask)
        self.npix=multiply.reduce(self.testimgc.shape)
        if self.alpha:
            self.initRegularizer() 
        self.cnormcoef=reshape(asarray(self.refimg.shape,Float),
                               (self.dim,)+(1,)*self.dim)
        
    def getInitialX(self):
        return zeros(self.csize,Float)

    def denormalizeC(self,c):
        """ Multiply components of c responsible for movements with respect
        to each axis by the number of pixels in this axis. Also usable for
        gradient and Hessian diagonal normalization """
        return c*self.cnormcoef

    def normalizeC(self,c):
        """ Inverse of denormalizeC """
        return c/self.cnormcoef


    def getE(self,cc,visualize=0,gnupl=None): # evaluate the criterion at point c
        c=self.denormalizeC(cc)
        self.w.setC(c) # pass c to the WarpableImage object
        self.wt=self.w.getWarped() # get warped test image
        self.difwtrefimg=self.wt-self.refimg
        if self.mask is not None:
            self.difwtrefimg*=self.mask
        E=bigtools.sumall(self.difwtrefimg**2) # the data part
        E/=self.npix
        if self.alpha: E+=self.getEr(c)*self.alpha
        if visualize:
            g=gnupl or Gnuplot.Gnuplot()
            d1=Gnuplot.Data(range(self.refimg.shape[0]),self.refimg,
                            with='lines',title='ref')
            d2=Gnuplot.Data(range(self.wt.shape[0]),self.wt,
                            with='lines',title='warped')
            g.replot(d1,d2)
            g.hardcopy('gete.eps')
            raw_input('Press any key')
        return E

    def getDerivatives(self):
        """ get first derivatives of the image described by coefs
        self.w.b along the coordinate axis """
        derw=zeros([self.dim]+list(self.w.b.shape),Float) # initialize
        g=self.w.getDeformation()
        for i in range(self.dim):
            derw[i]=bigsplines.splineinterpol(self.w.b,g,self.degf,
                                              flag=4**i,mask=self.mask)
        #print 'derw=',array2string(derw,precision=3,suppress_small=1)
        return derw

    def getSecondDerivatives(self):
        """ get second derivatives of the image described by coefs
        self.w.b along the coordinate axis """
        dderw=zeros([self.dim]+list(self.w.b.shape),Float) # initialize
        g=self.w.getDeformation()
        for i in range(self.dim):
            dderw[i]=bigsplines.splineinterpol(self.w.b,g,self.degf,
                                              flag=2*(4**i),mask=self.mask)
        return dderw

    def getEg(self,cc):
        if self.dim==2:
            return self.getEgC(cc)
        else:
            return self.getEgPython(cc)

    def getEgC(self,cc):
        """ The fast version of getEgPython, using a C implementation.
            Limitation: only works in 2D """
        
        if self.verbose:
            time1=time.clock()
            print "start getEgC"
        c=self.denormalizeC(cc)
        self.w.setC(c) # pass c to the WarpableImage object
        self.wt=self.w.getWarped() # get warped test image
        if self.verbose:
            time2=time.clock()
            print "getWarped took ",time2-time1, " s"
        derw=self.getDerivatives()
        if self.verbose:
            time3=time.clock()
            print "getDerivatives took ",time3-time2, " s"
        E,gr=bigsplines.sdcgr(self.refimg,c.shape[2],c.shape[1],
                              self.degc,derw,self.wt,self.mask)    
        if self.verbose:
            print "Sdcgr took ",time.clock()-time3, " s"
        gr/=self.npix
        if self.alpha:
            Er,grr=self.getEgr(c)
            E+=Er*self.alpha ; gr+=self.alpha*grr 
        return E/self.npix,self.denormalizeC(gr)
        
    def getEgPython(self,cc): # evaluate the criterion and the gradient
        c=self.denormalizeC(cc)
        E=self.getE(cc) # side effects!, many variables get changed
        derw=self.getDerivatives()
        # multiply all directional derivatives at each pixel
        # by F', where F is the criterion function
        derwp=2.0*self.difwtrefimg[NewAxis,:]*derw
        gr=zeros(c.shape,Float)
        # for all positions of knots
        clshape=c.shape[1:]
        for i in bigtools.getindices(shape=clshape): 
            # evaluate the contribution of a B-spline at position i
            beta=bigsplines.splineinterpolone(i,clshape,self.w.tbl,
                                              self.degc)
            # find the range of non-zero elements of beta
            brange=bigsplines.splinerange(i,clshape,self.refimg.shape,
                                          0.0,self.w.steps,
                                          self.degc)
            brange=bigtools.rangetoslice(brange) # convert to slices
            for j in range(self.dim): # all components
                gr[[j]+i]=bigtools.dotrange(derwp[j],beta,brange)
        gr/=self.npix
        if self.alpha:
            Er,grr=self.getEgr(c)
            E+=Er*self.alpha ; gr+=self.alpha*grr 
        return E,self.denormalizeC(gr)

    def getEgDh(self,cc): # evaluate the criterion, gradient, and Hessian diag.
        c=self.denormalizeC(cc)
        E=self.getE(cc) # side effects!, many variables get changed
        derw=self.getDerivatives()
        dderw=self.getSecondDerivatives()
        # multiply all directional derivatives at each pixel
        # by F', where F is the criterion function
        derwp=2.0*self.difwtrefimg[NewAxis,:]*derw
        # calculate F''*derw^2+F'*dderw
        dderwp=2.0*(derw**2+ self.difwtrefimg[NewAxis,:]*dderw)
        gr=zeros(c.shape,Float)
        H=zeros(c.shape,Float)
        # for all positions of knots
        clshape=c.shape[1:]
        for i in bigtools.getindices(shape=clshape): 
            # evaluate the contribution of a B-spline at position i
            beta=bigsplines.splineinterpolone(i,clshape,self.w.tbl,
                                              self.degc)
            # find the range of non-zero elements of beta
            brange=bigsplines.splinerange(i,clshape,self.refimg.shape,
                                          0.0,self.w.steps,
                                          self.degc)
            brange=bigtools.rangetoslice(brange) # convert to slices
            for j in range(self.dim): # all components
                gr[[j]+i]=bigtools.dotrange(derwp[j],beta,brange)
                H[[j]+i]=bigtools.dotrange(dderwp[j],beta,brange,square=1)
        gr/=self.npix
        H/=self.npix
        if self.alpha:
            Er,grr,Hr=self.getEgDhr(c)
            E+=Er*self.alpha ; gr+=self.alpha*grr 
            H+=Hr*self.alpha
        return E,self.denormalizeC(gr),self.denormalizeC(H)

    def getEgH(self,cc):
        """ evaluate the criterion, gradient, and the true Hessian
            slow implementation for the moment. Does not support
            regularization.
        """
        c=self.denormalizeC(cc)
        E=self.getE(cc) # side effects!, many variables get changed
        derw=self.getDerivatives()
        # multiply all directional derivatives at each pixel
        # by F', where F is the criterion function
        derwp=2.0*self.difwtrefimg[NewAxis,:]*derw
        gr=zeros(c.shape,Float)
        H=zeros(c.shape+c.shape,Float) # square matrix
        # for all positions of knots
        clshape=c.shape[1:]
        for i in bigtools.getindices(shape=clshape): 
            # find the range of non-zero elements of beta
            brange=bigsplines.splinerange(i,clshape,self.refimg.shape,
                                          0.0,self.w.steps,
                                          self.degc)
            bslice=bigtools.rangetoslice(brange) # convert to slices
            # evaluate the contribution of a B-spline at position i
            beta=bigsplines.splineinterpolone(i,clshape,self.w.tbl,
                                              self.degc)[bslice]
            for j in range(self.dim): # all components
                gr[[j]+i]=bigtools.sumall(derwp[j][bslice]*beta)
            # Here starts the Hessian calculation
            for k in bigtools.getindices(shape=clshape): 
                # find the range of non-zero elements of beta2
                brange2=bigsplines.splinerange(k,clshape,self.refimg.shape,
                                               0.0,self.w.steps,
                                               self.degc)
                # find the common part and terminate if empty
                brangecomm=bigtools.rangesintersect(brange,brange2)
                if bigtools.rangeisempty(brangecomm):
                    continue
                # relative and absolute slices
                bslice2rel=bigtools.rangetoslice((brangecomm[0]-brange[0],
                                                brangecomm[1]-brange[0]))
                bslice2abs=bigtools.rangetoslice(brangecomm)
                # evaluate the contribution of a B-spline at position k
                beta2=bigsplines.splineinterpolone(k,clshape,self.w.tbl,
                                              self.degc)[bslice2abs]
                betaprod=beta2*beta[bslice2rel]
                for j in range(self.dim): # all components
                    for l in range(self.dim):
                        H[[j]+i+[l]+k]=2.0*bigtools.sumall(betaprod*\
                                      derw[j][bslice2abs]*derw[l][bslice2abs])
        gr/=self.npix
        H/=self.npix
        return E,self.denormalizeC(gr),self.denormalizeC(H)


# ------------------------------------------------------------

#class MultigridableWarpingProblem(bigoptimize.SimpleReducer):
class MultigridableWarpingProblem(bigoptimize.SplineReducer):
    """ Generate a multigriddable variant of the WarpingProblem.
        An instance is generated given a reference and test pyramids,
        a class of the WarpingProblem, the size of the control grid,
        and degrees of splines to interpolate the deformation and
        the image.
    """

    def __init__(self,refp,testp,xshape=None,degf=3,degc=3,
                 wp=WarpingProblem,level=0,reducedimg=0,alpha=0.0,
                 verbose=0,mask=None):
        self.refp,self.testp=refp,testp
        self.degf,self.degc=degf,degc
        self.wp,self.level,self.mask=wp,level,mask
        self.alpha,self.verbose=alpha,verbose
        rimg=self.refp.getimage(0) # just to find out the dimensionality
        self.dim=len(rimg.shape)   
        # suggested size of coefs is 4 for each dimension
        self.csize= xshape or array([self.dim]+[4]*self.dim,Int)
        self.w=wp(self.refp.getimage(level),self.testp.getcoefs(level),
                  self.csize,self.degf,self.degc,alpha,verbose,mask)
        # reducedimg is true if image was reduced in the last call to getReduced
        self.reducedimg=reducedimg
        #print ('MultigridableWarpingProblem with xshape=',xshape,
        #           ' level=',level,' imsize=',self.refp.getimage(level).shape)

    def getInitialX(self):
        return self.w.getInitialX()

    def getE(self,c,visualize=0,gnupl=None):
        return self.w.getE(c,visualize,gnupl)
    
    def getEg(self,c):
        return self.w.getEg(c)

    def getEgDh(self,c):
        return self.w.getEgDh(c)

    def getEgH(self,c):
        return self.w.getEgH(c)

    def canReduceImg(self):
        # return true if the image should be further reduced
        r=bigtools.minelement(self.w.refimg.shape)>16
        # print "canReduceImg returning", r, ' ', self.w.refimg.shape
        return r

    def canReduceC(self):
        # return true if the control grid should be further reduced
        r=bigtools.minelement(self.csize[1:])>4
        # print "canReduceC returning", r, ' ', self.csize[1:]
        return r

    def canReduce(self):
        # return true if anything can be reduced
        return self.canReduceImg() or self.canReduceC()

    def reduceX(self,x):
        # a proper method to reduce the parameter x
        return self.reduceexceptfirst(x)

    def denormalizeC(self,c):
        return self.w.denormalizeC(c)

    def getReducedMask(self,mask):
        """ Given a bitmap mask, reduce it to half the size """
        mask=choose(greater(bigpyramid.mbsreduce(mask,0),0.2),(0,1))
        return mask.astype(UnsignedInt8)
    
    def getReduced(self,c):
        """ gets a reduced version of the problem. Here, the strategy is
        to alternatively reduce the size of the images and of the control
        points """
        # time to reduce the control size
        if (self.reducedimg or not self.canReduceImg()) and self.canReduceC(): 
            xc=self.reduceX(c)
            pc=MultigridableWarpingProblem(self.refp,self.testp,
                 xc.shape, self.degf,self.degc,self.wp,self.level,
                 alpha=self.alpha,verbose=self.verbose,mask=self.mask)
            return pc,xc
        # otherwise, reduce the image sizes
        if self.canReduceImg():
            xc=array(c) # a copy, it might be superflous??
            if self.mask is not None: # there is a mask
                self.mask=self.getReducedMask(self.mask)
            pc=MultigridableWarpingProblem(self.refp,self.testp,
                 xc.shape, self.degf,self.degc,self.wp,self.level+1,
                 reducedimg=1,alpha=self.alpha,verbose=self.verbose,mask=self.mask)
            return pc,xc
        raise BIGsplines.error, 'Problem cannot be further reduced.'

    def updateFine(self,xf,xc0,xc1):
        #print "Warping updateFine"
        xdif=xc1-xc0
        if xdif.shape!=xf.shape: # control size changed
            xdif=self.expandexceptfirst(xdif,targetlen=xf.shape)
        return xf+xdif
    
        
# ------------------------------------------------------------

class SpringProblem:
    """ This serves as a regularizer for the WarpingProblem.
    It evaluates

    Es=sum_q omega_q | g(x_i)/N-z_i |^2

    (where N is the number of pixels in the specific direction)
    and its derivatives with respect to c(describing g).
    x and z should have values between 0..1, x will be later
    multiplied by the image size in the given direction and
    converted to integer pixel coordinates to improve performance.

    For the moment it only works in 2D.

    """

    def __init__(self,wp,x,z,omega):
        """ Takes as parameters: the asociated WarpingProblem(wp),
        and the parameters (x,z,omega) of the springs
        """
        self.wp,self.x,self.z,self.omega=wp,x,z,omega 
        self.nump=x.shape[0]

    def getE(self,cc):
        return self.getEg(cc)[0]

    def getEg(self,cc):
        c=self.wp.denormalizeC(cc)
        E=0.0 ; gr=zeros(c.shape,Float)
        self.wp.w.setC(c) # this is maybe superflous 
        g=self.wp.w.getDeformation()
        x=(self.x*array(self.wp.refimg.shape,Float)).astype(Int)
        dx=zeros((self.nump,),Float)
        dy=zeros((self.nump,),Float)
        for q in range(self.nump):
            dx[q]=g[0][x[q]]/self.wp.refimg.shape[0]-self.z[q,0]
            dy[q]=g[1][x[q]]/self.wp.refimg.shape[1]-self.z[q,1]
            E+=self.omega[q]*(dx[q]*dx[q]+dy[q]*dy[q])
         # for all positions of knots
        clshape=c.shape[1:]
        # This is not very efficient ... to rewrite in C if needed
        for i in bigtools.getindices(shape=clshape): 
            # evaluate the contribution of a B-spline at position i
            beta=bigsplines.splineinterpolone(i,clshape,self.wp.w.tbl,
                                              self.wp.degc)
            for q in range(self.nump):
                gr[0][i]+=self.omega[q]*dx[q]*beta[x[q]]/self.wp.refimg.shape[0]
                gr[1][i]+=self.omega[q]*dy[q]*beta[x[q]]/self.wp.refimg.shape[1]
        gr*=2.0
        return E,self.wp.denormalizeC(gr)

# ------------------------------------------------------------

class SpringRegularizedWarpingProblem(bigoptimize.EvaluableProblem):
    """ Combines WarpingProblem and SpringProblem """

    def __init__(self,refimg,testimgc,csize=None,degf=3,degc=3,
                 alpha=0.0,verbose=0,mask=None,
                 springx=array(((0.0,0.0),)),
                 springz=array(((0.0,0.0),)),
                 springomega=array((0.0,))):
        self.wp=WarpingProblem(refimg,testimgc,csize,degf,degc,
                               alpha,verbose,mask)
        self.sp=SpringProblem(self.wp,springx,springz,springomega)
        # some variables need to be copied - ugly.... to redesign
        self.w=self.wp.w

    def getInitialX(self):
        return self.wp.getInitialX()

    def denormalizeC(self,c):
        return self.wp.denormalizeC()

    def normalizeC(self,c):
        return self.wp.normalizeC()

    def getE(self,cc,visualize=0,gnupl=None):
        E=self.wp.getE(cc,visualize,gnupl)+self.sp.getE(cc)
        return E
    
    def getEg(self,cc):
        Ew,gw=self.wp.getEg(cc)
        Es,gs=self.sp.getEg(cc)
        return Ew+Es,gw+gs

        
# ------------------------------------------------------------

def calculateMask(img,pct,degf=3):
    """ Calculates mask for img, such that it contains 'pct'
    pixels with the highest gradient values """
    # all what follows is an inefficient way to get the gradient
    imgw=bigsplines.mfspline(img,degf)
    wp=WarpingProblem(img,imgw,degf=degf)
    wp.w.setIdentity((2,1,1))
    derw=wp.getDerivatives()
    # derw**2 resulted in an overflow :-(, replaced by derw*derw
    dera=sum(derw*derw) # calculate the norm of the gradient (squared)
    sdera=sort(ravel(dera)) # sort the derivative amplitudes
    trsh=sdera[-(sdera.shape[0]*pct)/100] # threshold for 'pct'% of the pixels
    mask=choose(greater(dera,trsh),(0,1)).astype(UnsignedInt8)
    return mask
    
# ------------------------------------------------------------

def calculateThresholdedMask(img,threshold=30,degs=5):
    """ Calculates mask for img, such that it contains pixels
    with values higher than threshold.
    degs is roughly the half size of the holes to be filled in,
    similar to morphological closing. When degs=0, no filling
    is performed. """
    mask=(img>threshold).astype(Float)
    # convolve with a cubic spline to plug the holes
    mask=bigsplines.mconvrect(mask,degs)
    # threshold again
    mask=(mask>0.5).astype(UnsignedInt8) 
    return mask
    
# ------------------------------------------------------------

def test_WarpableImage():
    import NumTut
    im=Image.open('lena.tiff')
    # print "scaling image"
    #im=im.resize((64,64))
    # im=im.transpose(Image.ROTATE_90)
    # print "converting image"
    ima=bigtools.ImageToArray(im)
    # print "done"
    # print ima.shape
    #NumTut.view(ima)
    print "Initializing"
    w=WarpableImage(ima)
    print "done"
    w.setIdentity((2,4,4))
    w.getC()[0,2,2]=100 # deform
    wi=w.getWarped()
    #wim=bigsplines.ArrayToImage(wi)
    #wim=wim.resize((512,512))
    #wib=bigsplines.ImageToArray(wim)
    NumTut.view(wi)
    print wi.shape, ima.shape
    dif=abs(wi-ima)
    maxdif=maximum.reduce(maximum.reduce(dif))
    mindif=minimum.reduce(minimum.reduce(dif))
    print maxdif, mindif
    dif=(dif-mindif)*255.0/(maxdif-mindif)
    #difim=bigsplines.ArrayToImage(dif)
    #difim=difim.resize((512,512))
    #NumTut.view(bigsplines.ImageToArray(difim))
    NumTut.view(dif)

# ------------------------------------------------------------

def test_WarpingProblem():
    im=Image.open('lena.tiff')
    print "scaling image"
    im=im.resize((64,64))
    # im=im.transpose(Image.ROTATE_90)
    print "converting image"
    ima=bigtools.ImageToArray(im)
    print "done"
    # print ima.shape
    #NumTut.view(ima)
    print "Initializing"
    w=WarpableImage(ima)
    print "warping"
    w.setIdentity((2,4,4))
    c0=array(w.getC())
    # c0[0,2,2]=10 # deform
    w.setC(c0) 
    wi=w.getWarped()
    print "warped"
    #wim=bigtools.ArrayToImage(wi-ima)
    #wim=wim.resize((512,512))
    # wim.save("lenadif.jpg","JPEG")
    #wib=bigtools.ImageToArray(wim)
    #NumTut.view(wi-ima)
    #print bigtools.maxelement(abs(wi-ima))
    #wp=WarpingProblem(wi,w.b,csize=(2,2,2)) # reference image, test image
    wp=SpringRegularizedWarpingProblem(wi,w.b,csize=(2,2,2),
           springx=array(((0.0,0.0),(0.5,0.25))),
           springz=array(((1.0,0.0),(0.6,0.35))),
           springomega=array((1e6,1e6)))
    import bigoptimize
    bigoptimize.test_evproblem(wp)

def test_1dWP():
    degf=3 ; degc=3
    #xref=array((0,0.1,0,1,2,3,2,1))
    #xtest=array((0,0,1,2,3,2,1,0))
    N=200
    xref=fromfunction(lambda x,NN=float(N): exp(-((x/NN-0.5)*10)**2),(N,))
    xtest=fromfunction(lambda x,NN=float(N): exp(-((x/NN-0.6)*10)**2)+0.01,(N,))
    #xtest=xref+0.5*xtest
    #xtest=fromfunction(lambda x,NN=float(N): exp(-((x/NN-0.5)*10)**2),(N,))
    xtestc=bigsplines.mfspline(xtest,degf)
    #x0=array((1,1,0,-1))
    #x0=reshape(x0,(1,4))
    x0=zeros((1,8),Float)
    x1=array(x0) ; x1[0,2]+=0.1
    w=WarpableImage(xtestc,degf,degc,imageiscoefs=1)
    wp=WarpingProblem(xref,xtestc,x0.shape,alpha=0.0)
    import bigoptimize
    E,g,dH=wp.getEgDh(x0)
    print "E=",E,"g=",g,"dH=",dH
    E,g,H=wp.getEgH(x0)
    print "E=",E,"g=",g,"H=",bigtools.superdiagonal(H)
    #print H.shape
    return

    #print wp.getEg(x0)
    #x=x0
    #bigoptimize.test_evproblem(wp,x=x1)
    #return
    #x=bigoptimize.test_optimizerMLdH(wp,x1,progress=0)
    #x=bigoptimize.test_optimizerML(wp,x1,progress=1)
    #x=bigoptimize.test_optimizerGdes(wp,x0,progress=0)

    import bigpyramid
    refp=bigpyramid.BSPyramid(xref,degree=degf)
    testp=bigpyramid.BSPyramid(xtest,degree=degf)
    #g1=refp.show1d(5,3) ; g2=testp.show1d(5,3) 
    p=MultigridableWarpingProblem(refp,testp,x0.shape)
    p=bigoptimize.CorrectedMultigridableProblem(p)
    #bigoptimize.test_optimizerMLdH(p)
    #x=bigoptimize.testFMG(p)
    x=bigmultigrid.testFullT(p)
    #x=bigoptimize.testMR(p)
    print x
    w.setC(p.p.denormalizeC(x))
    wd=w.getWarped()
    g=Gnuplot.Gnuplot()
    t=arange(xref.shape[0])
    d1=Gnuplot.Data(t,xref,with='lines',title='ref') 
    d2=Gnuplot.Data(t,xtest,with='lines',title='test')
    d3=Gnuplot.Data(t,wd,with='lines',title='def') 
    g.plot(d1,d2,d3)
    raw_input('Press any key')

    



def test_2dWP():
    degf=3 ; degc=1
    xref=reshape(array((0,0.1,0,1,2,3,2,1)*10),(10,-1))
    xtest=reshape(array((0,0,1,2,3,2,1,0)*10),(10,-1))
    #xtest=xref
    #xtest[9,2]=0.7
    xtestc=bigsplines.mfspline(xtest,degf)
    #x0=0.1*reshape(array((1,1,0,-1)*8),(2,4,4))
    #x0[0,2,1]=0.3
    #x0=zeros((2,8,8))
    #x0[0,3:5,4:6]=1.0
    #x0=zeros((2,4,4)) ; x0[0,1,2]=1.0
    x0=zeros((2,4,4))
    #x0[0,3:6,4:7]=1.0
    wp=WarpingProblem(xref,xtestc,x0.shape,degf,degc,alpha=0.0)
    import bigoptimize
    #wp.getEg(zeros((2,4,4)))
    import sys
    sys.float_output_precision=3
    sys.float_output_suppress_small=1

    E,g,dH=wp.getEgDh(x0)
    print "E=",E,"g=",g,"dH=",dH
    E,g,H=wp.getEgH(x0)
    print "E=",E,"g=",g,"H=",bigtools.superdiagonal(H)
    return


    #bigoptimize.test_evproblem(wp,x=x0,h=1e-9)
    #return
    x=bigoptimize.test_optimizerMLdH(p=wp)
    
    import bigpyramid
    refp=bigpyramid.BSPyramid(xref,degree=degf)
    testp=bigpyramid.BSPyramid(xtest,degree=degf)
    p=MultigridableWarpingProblem(refp,testp,x0.shape)
    #p=bigoptimize.CorrectedMultigridableProblem(p)
    #bigoptimize.testFMG(p)
    #bigoptimize.testMR(p)
    #bigoptimize.testVcycle(p)
    print x

def test_3dWP():
    degf=3 
    xref=reshape(array((0,0.1,0,1,2,3,2,1)*4),(2,2,-1))
    xtest=reshape(array((0,0,1,2,3,2,1,0)*4),(2,2,-1))
    xtest[1,0,3]=0.7
    xtestc=bigsplines.mfspline(xtest,degf)
    x0=0.1*reshape(array((1,1,0,-1)*12),(3,2,2,4))
    x0[0,1,1,3]=-0.3
    wp=WarpingProblem(xref,xtestc)
    #wp.getEg(zeros((3,2,2,4)))
    import bigoptimize
    bigoptimize.test_evproblem(wp,x0,h=1e-7)
    #bigoptimize.test_optimizerMLdH(p=wp)

def test_lena():
    degf=3 ; degc=3
    im=Image.open('lena.tiff')
    #im=im.resize((64,64))
    ima=bigtools.ImageToArray(im)
    print "Initialization finished"
    imb=bigtools.timeop(lambda x=ima,y=degf:
                        bigsplines.mfspline(x,y),'Bspline transform')
    w=WarpableImage(ima,degf,degc)
    w.setIdentity((2,4,4))
    c0=array(w.getC())
    c1=array(c0)
    c1[0,2,1]=2 # deform
    w.setC(c1) 
    wi=w.getWarped() 
    print "Lena warped"
    #bigtools.view(wi)
    m=calculateMask(ima,10,degf)
    wp=WarpingProblem(wi,imb,c0.shape,degf,degc,mask=m)
    import bigoptimize
    #bigoptimize.test_evproblem(wp,x=c0)
    #t1=time.clock()
    #E,gr,H=wp.getEgDh(c0)
    #t2=time.clock()
    #print "getEgDh took %f s." % (t2-t1)
    x=c0 ; # x[0,3,2]+=0.05
    t1=time.clock()
    x=bigoptimize.test_optimizerGdes(wp,x)
    import sys
    sys.float_output_precision=3
    sys.float_output_suppress_small=1
    #print x
    #import bigpyramid
    #refp=bigpyramid.BSPyramid(wi,degree=degf)
    #testp=bigpyramid.BSPyramid(ima,degree=degf)
    #p=MultigridableWarpingProblem(refp,testp,c0.shape,degf,degc)
    #p=bigoptimize.CorrectedMultigridableProblem(p)
    #x=bigoptimize.testFMG(p)
    #c1=bigoptimize.test_optimizerMLdH(wp)
    #x=bigoptimize.testMR(p)
    t2=time.clock()
    print "Registration took %f s." % (t2-t1)
    #bigoptimize.testVcycle(p)
    c2=wp.denormalizeC(x)
    print array2string(c2,precision=3,suppress_small=1)
    w.setC(c2)
    wd=w.getWarped()
    print "Lena warped"
    bigtools.view(wd)
    bigtools.view(abs(ima-wi),size=(512,512),scale=(0,185),title='before')
    bigtools.view(abs(wd-wi),size=(512,512),scale=(0,185),title='after')

def time_getEg():
    import sys
    sys.float_output_precision=3
    sys.float_output_suppress_small=0

    degf=3 ; degc=3 ;
    #ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((64,64)))
    ima=bigtools.ImageToArray(Image.open('lena.tiff'))
    #imb=array(ima)
    w=WarpableImage(ima,degf,degc)
    w.setIdentity((2,4,4))
    c0=array(w.getC())
    c0[0,0,0]=10 # deform
    w.setC(c0)
    wi=w.getWarped()
    #imb[60,60]+=100
    #imb[10,11]+=50
    imaw=bigsplines.mfspline(ima,degf)
    #c0=zeros((2,4,4),Float)
    #wp=WarpingProblem(imb,imaw,c0.shape,degf,degc)
    m=identity(ima.shape[0]).astype(UnsignedInt8)
    m=zeros(ima.shape,UnsignedInt8)
    wp=WarpingProblem(wi,imaw,c0.shape,degf,degc,verbose=1,mask=m)
    #wp,w=bigmultigrid.getWarpedLena(degf,degc)
    #c0=wp.getInitialX()
    #E,gr1=bigtools.timeop(lambda p=wp,c=c0:
    #                        p.getEgPython(c),'getEg')
    #print "E=",E #,"gr=",gr1
    #E,gr=bigtools.timeop(lambda p=wp,c=c0:
    #                    p.getEg(c),'getEg')
    #print "E=",E # ,"gr=",gr
    import profile
    l={'wp':wp,'c0':c0}
    prof=profile.Profile()
    prof.runctx('wp.getEg(c0)',l,l)
    prof.print_stats()
    #print "gdif=",bigtools.sumall((gr-gr1)**2)
    #bigoptimize.test_evproblem(wp,x=c0)
    #print wp.w.steps
    
def test_mask():
    degf=3 ; degc=3 ;
    ima=bigtools.ImageToArray(Image.open('lena.tiff').resize((64,64)))
    imaw=bigsplines.mfspline(ima,degf)
    wp=WarpingProblem(ima,imaw,degf=degf,degc=degc)
    wp.w.setIdentity((2,8,8))
    derw=wp.getDerivatives()
    dera=sum(derw**2)
    sdera=sort(ravel(dera)) # sort the derivative amplitudes
    trsh=sdera[-sdera.shape[0]/10] # take 10% of the pixels
    imask=choose(greater(dera,trsh),(0,1))
    bigtools.view(imask,size=(128,128),scale=(0,2))
    #wi=WarpableImage(ima,mask=imask)
    #wi.setIdentity((2,4,4))
    #warped=wi.getWarped()
    #bigtools.view(warped)
    m2=bigpyramid.mbsreduce(imask,0)
    print bigtools.sumall(imask)/64.0/64.0
    m2=choose(greater(m2,0.2),(0,1))
    bigtools.view(m2,size=(128,128),scale=(0,2),title='reduced')
    print bigtools.sumall(m2)/32.0/32.0,m2.shape
    
# ------------------------------------------------------------

    
if __name__=='__main__': # running interactively
    #test_WarpableImage()
    #    test_WarpingProblem()
    test_lena()
