""" Module bigtools provides various auxiliary functions for
    doing numerics and image processing with Python 

    It contains some utility routines, e.g. converting
    to/from PIL images, up/down sampling, ...


    Jan Kybic, October 2000
    $Id: bigtools.py,v 1.16 2004/05/31 13:49:44 kybic Exp $

"""
import Tkinter 
import Image
from Numeric import *  # import Numeric Python extension
import LinearAlgebra, RandomArray
import Gnuplot
import ImageTk
import threading
import tempfile
import time
import warnings
import types 
import operator
import new
import string
import sys
import re
# import bigsplines # applyseparable functions moved here to bigtools

# ----------------------------------------------------------------------

def ArrayToImage(a,byte=1,clip=0):
    """ converts an integer array to greyscale or color image
        Makes 1B/pixel images (default, byte=1), or 2B/pixel (byte=2)

        clip to 1B (0-255) or 2B (0-65535), or no clipping (clip==0)

        See http://www-md.fsl.noaa.gov/eft/developer/python/pythonware/library/pil/handbook/decoder.html for the mode description.
    """    
    if len(a.shape)==2:
        try:
            typec=['b','s'][byte-1]
            decmode=['L','I;16'][byte-1] ; immode=decmode
        except IndexError:
            raise IndexError,"Parameter 'byte' out of range"

        if clip==1:     
            a=choose(greater(a,255.0),(a,255.0))
            a=choose(greater(a,0.0),(0.0,a))
        if clip==2:     
            a=choose(greater(a,65535.0),(a,65535.0))
            a=choose(greater(a,0.0),(0.0,a))

        a=a.astype(Int) # rounding

        if maxelement(a)>256**byte or minelement(a)<0:
            raise ValueError, \
                   "Array elements out of range for the image. (%f,%f)" % \
                   (minelement(a),maxelement(a))
        i=Image.fromstring(immode,(a.shape[1],a.shape[0]),
                       (a.astype(typec)).tostring(),"raw",immode)
    else: # color
        r,g,b=(ArrayToImage(a[0],byte=byte),
               ArrayToImage(a[1],byte=byte),ArrayToImage(a[2],byte=byte))
        i=Image.merge("RGB",(r,g,b))
    return i

def ImageToArray(i):
    """ converts greyscale or color image to an array """
    if i.mode=='RGB': # color image
        r,g,b=i.split()
        a=array((ImageToArray(r),ImageToArray(g),ImageToArray(b)))
    elif i.mode=='RGBA': # color image with alpha channel
        r,g,b,alpha=i.split()
        a=array((ImageToArray(r),ImageToArray(g),ImageToArray(b)))
    else:    # other modes,  'I','L' or 'F' for grayscale images
        a=reshape(array(i.getdata()),(i.im.size[1],i.im.size[0]))
    return a

# ----------------------------------------------------------------------

def downsample(x,n):
    """ Downsample 1D array x by a factor of n, starting at the first
        element """
    return take(x,range(0,x.shape[0],n))

# ----------------------------------------------------------------------

def upsample(x,n):
    """ Upsample 1D array x by a factor of n, starting at the first
        element and filling gaps with zero. Returns a vector of length
        len(x)*n """
    y=zeros((x.shape[0],n),Float)
    y[:,0]=x
    return y.flat

# --------------------------------------------------------------------

def mupsample(x,n):
    """ Upsample a multidimensional array x by a factor n[i] in
    direction i. Uses upsample. """
    
    return applyseparablewithcopy(x,
          lambda inp,n=n,ndim=None: upsample(inp,n[ndim]),passdim=1)

# --------------------------------------------------------------------

def mdownsample(x,n):
    """ Downsample a multidimensional array x by a factor n[i] in
    direction i. Uses downsample. """
    
    return applyseparablewithcopy(x,
          lambda inp,n=n,ndim=None: downsample(inp,n[ndim]),passdim=1)

# --------------------------------------------------------------------

def getindices(x=None,shape=None):
    """ Return a list of all indices of an array x. Usage:

    for i in getindices(x): # print all elements
        print x[i]
    """
    if not shape:
        shape=x.shape
    i=indices(shape)
    i=reshape(i,(i.shape[0],-1)) ; i=transpose(i)
    return i.tolist()

# --------------------------------------------------------------------

def basisv(shape,i, type=Float):
    """ Return an  array of 'type' of shape 'shape', with all zeros except
        one at index i.
    """
    x=zeros(shape,type) ; x[i]=1.0
    return x

# --------------------------------------------------------------------

# for type comparisons such as type(points)!=ArrayType
ArrayType=type(array((1)))

# --------------------------------------------------------------------

def maxelement(x):
    """ Returns the value of the largest element of x, regardless of the
    dimensionality. """
    x=array(x)
    while type(x)==ArrayType and len(x.shape)>0:
        x=maximum.reduce(x)
    return x

# --------------------------------------------------------------------

def minelement(x):
    """ Returns the value of the smallest element of x, regardless of the
    dimensionality. """
    x=array(x)
    while type(x)==ArrayType and len(x.shape)>0:
        x=minimum.reduce(x)
    return x

# --------------------------------------------------------------------

def sumall(x):
    """ Returns the sum of all the elements of x, regardless of the
    dimensionality. """
    x=array(x)
    while type(x)==ArrayType and len(x.shape)>0:
        x=add.reduce(x)
    return x


def prodall(x):
    """ Returns the product of all the elements of x, regardless of the
    dimensionality. """
    x=array(x)
    while type(x)==ArrayType and len(x.shape)>0:
        x=multiply.reduce(x)
    return x

# --------------------------------------------------------------------

def nelems(x):
    """ Return the number of elements of array x """
    return prodall(x.shape)

# --------------------------------------------------------------------

def float_array(m):
    """ Makes a floating array out of m, accepting just about anything
    Taken from Gnuplot.utils by
    Michael Haggerty <mhagger@alum.mit.edu>"""
    
    try:
        # Try Float32 (this will refuse to downcast)
        return asarray(m, Float32)
    except TypeError:
        # That failure might have been because the input array was
        # of a wider data type than Float32; try to convert to the
        # largest floating-point type available:
        return asarray(m, Float)

def int_array(m):
    """ Makes an int array out of m, accepting just about anything
    """
    
    try:
        return asarray(m, Int)
    except TypeError:
        return m.astype(Int) # round to Int


# --------------------------------------------------------------------

def addmirrored(x,n):
    """ Take a 1D vector x and add n points at both ends, respecting the mirror
    on boundary conditions. """
    if n==0: return x
    x=float_array(x)
    y=zeros((x.shape[0]+2*n,),'d')
    y[0:n]=x[n:0:-1]
    y[n:n+x.shape[0]]=x
    y[-1:-n-1:-1]=x[-n-1:-1]
    return y

# --------------------------------------------------------------------

def maddmirrored(x,n):
    """ Apply addmirrored to a multidimensional x, extending
    it in all directions"""
    if n==0: return x
    return applyseparablewithcopy(x,
                            lambda inp,n=n: addmirrored(inp,n))

# --------------------------------------------------------------------

        
def foldmirrored(x,n):
    """ Cut n points at the beginning and end of x, foldover and add
    to the original vector"""
    if n==0: return x
    y=array(x[n:-n]) # make a copy
    y[1:n+1]+=x[n-1::-1]
    y[-2:-n-2:-1]+=x[-n:]
    return y

# --------------------------------------------------------------------

def mfold(k,n):
    """ Calculates the real index corresponding to k mirrored for vector
        length n """
    m=2*(n-1)
    f=k % m
    if f<0: return f+m
    if f>=n: return m-f
    return f

# --------------------------------------------------------------------
    
def mfoldv(k,n):
    """ As mfold but accepts sequences as k and n"""
    f=asarray(k,Int) # make copy
    k=asarray(f,Int) # make copy
    n=asarray(n,Int) # make copy
    for i in getindices(f):
        f[i]=mfold(k[i],n[i])
    return f

# --------------------------------------------------------------------

def rangetoslice(r):
    """ Given the range as returned by the splinerange
    in the form (array(l1,l2,l3,...),array(h1,h2,h3,...)),
    return a list of slice objects to be used for slicing """
    s=[]
    r=transpose(asarray(r,Int))
    for i in range(len(r)):
        s.append(slice(r[i,0],r[i,1]+1))
    return s
    
# --------------------------------------------------------------------

def rangesintersect(r,s):
    """ Given two ranges r,s
    in the form (array(l1,l2,l3,...),array(h1,h2,h3,...)),
    return their intersection """
    return maximum(r[0],s[0]),minimum(r[1],s[1])

# --------------------------------------------------------------------

def rangeisempty(r):
    """ Given a range r, return true if it is empty """
    return not alltrue(greater(r[1]-r[0],0))

# --------------------------------------------------------------------

def dotrange(x,y,s,square=0):
    """ Compute the dot-product of x and y, taking advantage of the
    fact that the only non-zero elements of y are in range s,
    as returned by rangetoslice(r). If square==1, the second argument
    is squared.
    """
    if not square:
        return sumall(x[s]*y[s])
    else:
        ys=y[s]
        return sumall(x[s]*ys*ys)

# --------------------------------------------------------------------

def applyforallrows(x,oper,ndim=None):
    """ Apply an operation for all rows of x. The operation oper
    should take a 1D vector and return another 1D vector.

    If passdim is true, the 'oper' is passed a keyword argument
    'ndim' giving the dimension being processed.
    """
    xshape=x.shape ; dims=len(xshape)
    x=reshape(x,(-1,xshape[dims-1]))
    y=zeros(x.shape,Float)
    for i in range(x.shape[0]): # for all 'rows'
        if ndim is None:
            y[i]=apply(oper,(x[i],)) # transform 'row'
        else:
            y[i]=apply(oper,(x[i],),{'ndim':ndim}) # transform 'row'
    x=reshape(x,xshape) # correct the shape of x
    return reshape(y,xshape)

# --------------------------------------------------------------------

def vectorsumforalldims(x,oper):
    """ Call operation (vector->vector) on all rows of all transposes of x
    and return the sum of the results """
    xshape=x.shape ; dims=len(xshape)
    y=zeros(x.shape,Float) # the resultant vector
    for i in range(dims):
        y+=applyforallrows(x,oper)
        x=transpose(x,[dims-1]+range(dims-1))
        y=transpose(y,[dims-1]+range(dims-1))
    return y

# --------------------------------------------------------------------
        
def sumforallrows(x,oper):
    """ Call oper on all rows (including higher dimensions) of x
    and return the sum of the results """
    E=0.0
    xshape=x.shape ; dims=len(xshape)
    x=reshape(x,(-1,xshape[dims-1])) # flatten all other dimensions
    for i in range(x.shape[0]): # for all rows
        E+=apply(oper,(x[i],))
    x=reshape(x,xshape) # change shape back
    return E

# --------------------------------------------------------------------

def sumforalldims(x,oper):
    """ Call operation on all transposes of x
    and return the sum of the results """
    E=0.0
    xshape=x.shape ; dims=len(xshape)
    for i in range(dims):
        E+=apply(oper,(x,))
        x=transpose(x,[dims-1]+range(dims-1))
    return E

# --------------------------------------------------------------------

def keyboard():
    """ Debugging function. Stops the program  and lets you execute
    commands from the keyboard until empty line is given. """
    ### DOES NOT WORK SO FAR ###
    while 1:
        cmd = raw_input("keyboard: ")
        if not cmd:
            break
        exec(cmd)
    

# --------------------------------------------------------------------

def plot(*set,**kw):
    """ Plot function going through points x,y (Numeric arrays or other).
        Only y can be given.

        Usage: g=plot(y)  or g=plot(x,y)

        Keyword parameters: title, gnupl, with, xlabel, ylabel, cmd
        
        Returns the Gnuplot handle """
    if len(set)==1:
        y=float_array(set[0])
        x=arange(y.shape[-1])
    else:
        x=float_array(set[0])
        y=float_array(set[1])

    if kw.has_key('with'):
        w=kw['with']
    else:
        w='linespoints pointsize 2'

    if kw.has_key('title'):
        t=kw['title']
    else:
        t=None
        
    if kw.has_key('gnupl'):
        g=kw['gnupl']
    else:
        g=Gnuplot.Gnuplot()
    if len(y.shape)==1: # one plot
        d=[ Gnuplot.Data(x,y,with=w,title=t) ]
    else:
        d=[ Gnuplot.Data(x,y[i],with=w,title=t[i]) for i in xrange(y.shape[0]) ]
    if kw.has_key('xlabel'):
        g.xlabel(kw['xlabel'])
    else:
        g.xlabel('x')
    if kw.has_key('ylabel'):
        g.ylabel(kw['ylabel'])
    else:
        g.ylabel('y')
    if kw.has_key('cmd'):
        g(kw['cmd'])
    apply(g.replot,d)
    return g

# --------------------------------------------------------------------
def save_ppm(ppm, fname=None):
    if fname == None:
        fname = tempfile.mktemp('.ppm')
    f = open(fname, 'wb')
    f.write(ppm)
    f.close()
    return fname


def array2ppm(image):
    # scaling
    if len(image.shape) == 2:
        # B&W:
        image = transpose(image)
        return "P5\n#PPM version of array\n%d %d\n255\n%s" % \
               (image.shape[1], image.shape[0], ravel(image).tostring())
    else:
        # color
        image = transpose(image, (1, 0, 2))
        return "P6\n%d %d\n255\n%s" % \
               (image.shape[1], image.shape[0], ravel(image).tostring())


class ImageLabel(Tkinter.Label):

    def __init__(self, master, im):

        if im.mode == "1":
            # bitmap image
            self.image = ImageTk.BitmapImage(im, foreground="white")
            Tkinter.Label.__init__(self, master, image=self.image,
                                   bg="black", bd=0)

        else:
            # photo image
            self.image = ImageTk.PhotoImage(im)
            #print "image=",self.image
            Tkinter.Label.__init__(self, master, image=self.image, bd=0)

class ViewWindow:
    def __init__(self):
        if Tkinter._default_root is None:
            Tkinter.Tk()
            Tkinter._default_root.withdraw()
        self.root=Tkinter.Toplevel()

    def getRoot(self):
        return self.root

    def __del__(self):
        if self.root!=None:
            self.root.withdraw()
            self.root=None

def view(img,size=None,scale='auto',title='view',resample=Image.NEAREST,
         clip=1):
    """ View a 2D array img as image, possibly scaled to 'size'
    which should be a tuple (xsize,ysize). If no scaling is used (scale=None),
    the range is 0..255

    If img is 3D the first coordinate is assumed to be the color index,
    i.e. img[0] is red, img[1] green, and img[2] is blue

    It returns a handle to a ViewWindow object. When it is destroyed, the
    window disappears.

    resample defines the interpolation method, it can be NEAREST, BILINEAR,
    or BICUBIC, see Image.py
    """

    img=img.astype(Float)
    if scale=='auto':
        maxel=maxelement(img)
        minel=minelement(img)
        scale=(minel,maxel)
        #print "maxel=",scale[1],"minel=",scale[0]
    if scale:
        difel=scale[1]-scale[0]
        if difel==0: difel=1.0
        img=(img-scale[0])*255.0/difel
        maxel,minel=(0.0,255.0) # so that the test below works
    if clip==1:     
        img=choose(greater(img,255.0),(img,255.0))
        img=choose(greater(img,0.0),(0.0,img))
    if clip==2:     
        img=choose(greater(img,65535.0),(img,65535.0))
        img=choose(greater(img,0.0),(0.0,img))

    #fname=save_ppm(array2ppm(img))
    #print "fname=",fname
    pimg=ArrayToImage(img)
    if size:
        pimg=pimg.resize(size,resample=resample)
    return viewimg(pimg,title=title)

def viewimg(img,title='view'):
    """ View a PIL image object img. It
    returns a handle to a ViewWindow object. When it is destroyed, the
    window disappears. """
    vw=ViewWindow()
    root=vw.getRoot()
    root.title(title)
    ImageLabel(root,img).pack()
    return vw

    

def printtofile(img,filename,size=None,scale='auto',resample=Image.NEAREST,
                clip=1):
    """ Save to a file a 2D array img, possibly scaled to 'size'
    which should be a tuple (xsize,ysize). The filename should
    already contain the extension. If no scaling is used (scale=None),
    the range is 0..255

    If img is 3D the first coordinate is assumed to be the color index,
    i.e. img[0] is red, img[1] green, and img[2] is blue

    resample defines the interpolation method, it can be NEAREST, BILINEAR,
    or BICUBIC, see Image.py
    If clip==1, clip to 0..255, if clip==2, clip to 0..65535

    Chooses automatically if the resulting file should be 8b or 16b/pixel,
    clip==2 forces 16b/pix.
    """
    img=img.astype(Float)
    maxel=maxelement(img)
    if scale=='auto':
        minel=minelement(img)
        scale=(minel,maxel)
    if scale:
        difel=scale[1]-scale[0]
        if difel==0: difel=1.0
        img=(img-scale[0])*255.0/difel
        maxel,minel=(0.0,255.0) # so that the test below works
    if clip==1:     
        img=choose(greater(img,255.0),(img,255.0))
        img=choose(greater(img,0.0),(0.0,img))
    if clip==2:     
        img=choose(greater(img,65535.0),(img,65535.0))
        img=choose(greater(img,0.0),(0.0,img))
    if maxel>255 or clip==2:
        byte=2
    else: byte=1
    pimg=ArrayToImage(img,byte=byte)
    if size:
        pimg=pimg.resize(size,resample=resample)
    pimg.save(filename)


# --------------------------------------------------------------------

def timeop(oper,comm='Operation'):
    t1=time.clock()
    r=apply(oper)
    t2=time.clock()
    print comm+' took %f s.' % (t2-t1)
    return r

# --------------------------------------------------------------------

def aexp(x):
    """ Works just like an ordinary exp, except it returns zero for
        large negative arguments, not OverflowError :-(

        no longer necessary in Python 2.1
        """
    warnings.warn("aexp no longer necessary, use exp() instead")
    try:
        r=exp(x)
    except OverflowError:
        if x<0: return 0.0
        raise OverflowError, "argument to exp too big"
    return r

# --------------------------------------------------------------------

def superdiagonal(x):
    """ returns a superdiagonal of a matrix x, i.e. another matrix
        y(i)=x(i,i), where i can be a vector index """
    xs=x.shape ; xd=len(xs)
    y=zeros(xs[:xd/2],x.typecode())
    for i in getindices(y):
        y[i]=x[i+i]
    return y

# -------------------------------------------------------------------------


class Parameters:
    """ An object of class Parameters is a collection of parameters.
        Its instantiated as:

        p=Parameters(x)

        where x can be either another object of type Parameters,
        or a dictionary, such as

        p=Parameters({'tol':1e-6, 'startx':(0,0,0)})

        Additional keyword parameters to the constructor include:

        default: a dictionary with default values
        keep:    a list of parameter names to keep. If given, all other
                 parameters are discarded, except those given by
                 'default'
        discard: list of parameters to discard. This is done after
                 'default' and 'keep' are processed.
        override: these parameters are added as the last step

        Once the object is instantiated, the values can be read by
        as 'p.tol' or 'p.startx', they should not be changed this way
        (although this cannot be enforced for types passed by reference)
        For invalid names, KeyError is raised.

        Policy: parameter values should be never changed.

        Warning: Each subordinate module makes its copy
        
    """

    def __init__(self,p,default=None,keep=None,discard=[],override=None):
        self.p={}
        if default is not None:
            self.p.update(default)
        if p is not None:
            if type(p)==types.DictType:
                self.p.update(p)
            else:
                self.p.update(p.p) # p is a Parameters object
        if keep is not None:
            keep+=default.keys() # all keys we want to keep
            p={}
            for i in keep:
                p[i]=self.p[i] # copy what is needed
            self.p=p
        for i in discard:      # delete what we do not want
            del self.p[i]
        if override is not None:
            self.p.update(override)

    def isdef(self,name):
        """ Returns true if a parameter 'name' is defined """
        return self.p.has_key(name)

    def __getattr__(self,name):
        if self.isdef(name):
            return self.p[name]
        raise KeyError, "Unknown attribute: %s" % name

# -------------------------------------------------------------------------

def test_Parameters():
    p=Parameters({'a':0,'b':1,'c':"hmm"})
    print p.p, p.a, p.b, p.c
    q=Parameters(p,default={'a':100,'d':"d"})
    print q.p # should print a=0,b,c,d
    r=Parameters(p,default={'a':100,'d':"d"},keep=['b'])
    print r.p # should print a,b,d
    s=Parameters(r,discard=['a','d'])
    print s.p # should print b

# -------------------------------------------------------------------------

def makeClassBag(classes=[],accumulators={}):
    """ makeClassBag permits to create a bag object containing several
        client objects. When a message is sent to the bag object, it is
        resent to all clients. The response from clients is aggregated
        using a user supplied accumulating functions.

        Instantiation:
        classbag=makeClassBag([classA,classB,classC],
                     accumulators={'result': operator.add})

        b=classbag(1,2,3) # instantiates all clients

        Calling b.fun(1,2) then returns: classA(1,2)+classB(1,2)
        Default operator is returning the list of all results.

        Note that changing data members of the client objects is not supported
        """
    name=string.join(map(lambda x: x.__name__,classes),'+')
    c=new.classobj('ClassBag+'+name,(ClassBag,),{}) # new class
    c.ClassBagClassList=classes
    c.ClassBagAccumulators=accumulators
    return c

class ClassBag:

    def __init__(*set,**kw):
        """ Initialize the ClassBag object """
        result=[] ; self=set[0]
        for i in self.ClassBagClassList:
            r=apply(i,set[1:],kw) # instantiate a client
            result.append(r)
        self.ClassBagObjectList=result
        return None

    def __getattr__(self,name):
        """ Dispatching routines """
        #print "Calling __getattr__ from ClassBag with `%s'" % name
        #print "objects=",self.ClassBagObjectList
        a=self.ClassBagAccumulators.get(name,None)
        d=ClassBagDispatcher(self.ClassBagObjectList,name,a)
        return d.doCall


class ClassBagDispatcher:
    def __init__(self,objects,fnname,accumulator):
        self.objects=objects
        self.fnname=fnname
        self.accumulator=accumulator
        
    def doCall(*set,**kw):
        # print "doCall called with ",set," and ", kw
        result=[] ; self=set[0]
        for i in self.objects:
            r=apply(getattr(i,self.fnname),set[1:],kw) # call client function
            result.append(r)
        if self.accumulator is not None:
            result=reduce(self.accumulator,result) # accumulate
        return result

def test_ClassBag():
    class A:
        def __init__(self,a):
            print "__init__ of A called with ",a

        def r(self,w):
            print "r of A called with ",w
            return 2*w
        
    class B:
        def __init__(self,a):
            print "__init__ of B called with ",a

        def r(self,w):
            print "r of B called with ",w
            return 3*w

    classbag=makeClassBag([A,B],accumulators={'r':operator.add})
    print classbag
    b=classbag(7) # instantiate
    print "objects=",b.ClassBagObjectList
    print b.r(10) # dispatch calls

# -------------------------------------------------------------------------

class CombinedBase:
    """ A mixin class for objects derived from several base classes. It
    can discover its base classes and call its methods """
    
    def getBaseNames(self):
        """ Finds out base names (as strings) at puts them into
            self.basenames. If a getNickname_<BaseName> method exists,
            the nickname returned is used insted """
        self.basenames=[]
        for x in self.__class__.__bases__: # x is a base class
            name=x.__name__ 
            try:
                nicknamemethod=getattr(x,'getNickname_'+name)
            except AttributeError: # getNickname is not implemented
                self.basenames.append(name)
            else:                  # if it is implemented, call it
                self.basenames.append(apply(nicknamemethod,self))
        return self

    def callNames(self,prefix,args=(),kwds={}):
        """ calls all methods with names prefix_BaseName, where BaseName is
        taken from self.basenames. Returns a list of all results. """
        return [ apply(gettattr(self,prefix+'_'+b),args,kwds)
                 for b in self.basenames ]

def test_CombinedBase():
    import bigfelreg
    class t(bigfelreg.WarpableImageS,CombinedBase):
        pass
    to=t(array((0,0,)))
    print to.getBaseNames().basenames


# -------------------------------------------------------------------------

def tensorprod(v):
    """ Calculates a tensor product of all elements of v, v is a list
    of vectors """
    lv=len(v)
    if lv<1: return v
    if lv==1: return v[0]
    v01=outerproduct(v[0],v[1])
    if lv==2: return v01
    v01shp=list(v01.shape)
    v2=tensorprod([reshape(v01,(-1,))]+v[2:]) ; v2shp=list(v2.shape)
    return(reshape(v2,v01shp+v2shp[1:]))

def try_tensor():
    q=arange(3)
    r=tensorprod([q,q,arange(2)])
    print r,r.shape
    
         

# -------------------------------------------------------------------------

def makeGrid(size,step=(20,20)):
    """ Returns an array with a grid pattern """
    g=zeros(size,Float)
    def addgrid(g,step):
        """ adds lines with step along one dimension """
        line=ones((g.shape[1],),Float)
        for i in range(step/2,g.shape[0],step): g[i,:]=line
    addgrid(g,step[0])
    g=transpose(g)
    addgrid(g,step[1])
    return transpose(g)

# ----------------------------------------------------------------------

def showDif(img1,img2,size=None):
    """ Shows a difference between two images. The two images must be
    of the same size. Currently we put one in red and the other one in the
    green channels. Returns an image suitable for bigtools.view. """
    imtot=zeros((3,)+img1.shape,Float)
    imtot[0]=img1 ; imtot[1]=img2
    return imtot

# ----------------------------------------------------------------------

def showContours(img,cont,size=None,scale='auto',title='showDif'):
    """ Shows superposed contours cont in red over an image img in gray.
    The two images must be of the same size.
    Contours are plotted if bigger than 0.
    Returns an image suitable for bigtools.view. """
    imtot=zeros((3,)+img.shape,Float)
    col=array((1.0,0.0,0.0))*maxelement(img) # red
    img=img.astype(Float)
    for i in xrange(3):
        imtot[i]=choose(cont>0,(img,col[i]))
    return imtot

# ----------------------------------------------------------------------

def showImgWithLandmarks(img,lnd,size=None,scale='auto',title='Landmarks',
                         lndsize=2):
    """ Returns an image suitable for bigtools.view
        with superimposed landmarks in red.
    WARNING: We do not check if the landmarks are too close to borders.
    """
    imtot=zeros((3,)+img.shape,Float)
    val=maxelement(img)
    imtot[:]=img
    for i in list(lnd):
        i=int_array(i)
        imtot[0,i[0]-lndsize:i[0]+lndsize,i[1]-lndsize:i[1]+lndsize]=val
        imtot[1,i[0]-lndsize:i[0]+lndsize,i[1]-lndsize:i[1]+lndsize]=0.0
        imtot[2,i[0]-lndsize:i[0]+lndsize,i[1]-lndsize:i[1]+lndsize]=0.0
        
    return imtot

# ----------------------------------------------------------------------

def calchist(x,h=1.0,minx=None,maxx=None):
    """ Given an array (or vector x), returns pair of vectors (t,cnt)
    suitable for showing the histogram of x with plot. h is the spacing
    and minx/maxx the lower and upper limits """
    x=reshape(float_array(x),(-1,))
    if minx is None: minx=minelement(x)
    if maxx is None: maxx=maxelement(x)
    nslots=math.ceil((maxx-minx)/h)
    cnt=zeros((nslots,),Int)
    for i in xrange(x.shape[0]):
        nsl=int(math.floor((x[i]-minx)/h))
        if nsl<nslots and nsl>=0: cnt[nsl]+=1
    t=arange(nslots)*h+0.5*h+minx
    return t,cnt

# ----------------------------------------------------------------------

def applyseparable(x,oper,passdim=None):
    """ Apply operation oper on all dimensions of an 
        arrays x. Oper should take one argument - a 1D double array and return
        an array of the same size. applyseparable returns a double array

        If passdim is true, the 'oper' is passed a keyword argument
        'ndim' giving the dimension being processed.
        """
    dims=len(x.shape)
    x=array(x,'d') # make a copy and upcast to double if needed
    for d in range(dims): # for all dimensions
        if passdim:
            x=bigtools.applyforallrows(x,oper,ndim=d)
        else:
            x=bigtools.applyforallrows(x,oper)
        x=transpose(x,range(1,dims)+[0]) # shift dimensions
    return x

# ----------------------------------------------------------------------

def applyseparablewithcopy(x,oper,targetlen=None,passdim=None):
    """ Apply operation oper on all dimensions of an 
        arrays x. Oper should take one argument - a 1D double array and return
        another 1D array not necessarily of the same size.
        applyseparablewithcopy returns a double array.

        If targetlen is not None, it is supposed to be the list of target
        dimensions, the respective dimension will be passed to 'oper'.

        If passdim is true, the 'oper' is passed a keyword argument
        'ndim' giving the dimension being processed.
        Currently targetlen and ndim are mutually exclusive.
        """
    dims=len(x.shape)
    x=array(x,'d') # make a copy and upcast to double if needed
                   # y is going to be the output copy
    t=targetlen
    for d in xrange(dims): # for all dimensions
        xshape=list(x.shape)
        x=reshape(x,(xshape[0],-1)) # flatten all other dimensions
        yshapew=list(x.shape)       # flattened shape
        yshapew[0]=-1               # force allocating y
        for i in xrange(x.shape[1]): # for all 'rows'
            if targetlen:
                trrow=apply(oper,(x[:,i],),{'tlen': t[d]}) # transform 'row'
            if passdim:
                trrow=apply(oper,(x[:,i],),{'ndim': d}) # transform 'row'
            else:
                trrow=apply(oper,(x[:,i],)) # transform 'row'
            if trrow.shape[0]!=yshapew[0]: # dimensions differ
                yshapew[0]=trrow.shape[0] ; xshape[0]=yshapew[0]               
                y=zeros(yshapew,Float)            # change them
            y[:,i]=trrow
        y=reshape(y,xshape) # return to stored shape
        # attention, shifting in another direction then applyseparable
        y=transpose(y,range(1,dims)+[0]) # shift dimensions
        x=y # in case we want to repeat the cycle
    return y

# ----------------------------------------------------------------------

def crossproduct(a,b):
    """ Calculates a cross product of two 3 element float vectors """
    c=zeros((3,),Float)
    c[0]= a[1]*b[2]-a[2]*b[1] 
    c[1]=-a[0]*b[2]+a[2]*b[0] 
    c[2]= a[0]*b[1]-a[1]*b[0]
    return c

# ----------------------------------------------------------------------

def norm(x):
    """ l2 norm of a vector """
    return sqrt(dot(x,x))

# ----------------------------------------------------------------------

def norm2(x):
    """ l2 norm of a vector squared """
    return dot(x,x)

# ----------------------------------------------------------------------

def norm2m(x):
    """ l2 norm of a vector or matrix squared """
    return dot(x.flat,x.flat)


# ----------------------------------------------------------------------

def mean(x):
    """ a mean of a vector """
    return sumall(x)/reshape(x,(-1,)).shape[0]

# ----------------------------------------------------------------------

def normalize(x):
    """ Normalize a vector x """
    return x/norm(x)

# ----------------------------------------------------------------------

def scaletozeroone(x):
    """ Scale a vector x """
    maxe=maxelement(x) ; mine=minelement(x) ;
    return (x-mine)/(maxe-mine)

# ----------------------------------------------------------------------

def saveForMatlab(x,filename):
    """ saves a 2D matrix into a 'file' so that it can be read by Matlab
    """
    f=open(filename,'w')
    x=reshape(x,(x.shape[0],-1)) # make it 2D
    for i in xrange(x.shape[0]):
        s=array2string(x[i],max_line_width=sys.maxint)[1:-1]+'\n'
        # add zeros after points so that Haskell can read it
        f.write(re.sub(r'\.([^0-9])',r'.0\1',s,sys.maxint)) 
    f.close()
    
# ----------------------------------------------------------------------

def floatForHaskell(x):
    return re.sub(r'\.([^0-9])',r'.0\1',`x`,sys.maxint)

def vectorForHaskell(x):
    s=array2string(x,max_line_width=sys.maxint)[1:-1]
    return re.sub(r'\.([^0-9])',r'.0\1',s,sys.maxint) 

# ----------------------------------------------------------------------

def readFromMatlabAsciiFile(filename):
    """ Returns a 2D array from a text file such as produced by Matlab's
    save -ascii command """
    f=open(filename,'r')
    x=[]
    while 1:
        l=string.split(f.readline())
        if len(l)==0: break    # no more lines to read
        try:
            x.append(array([ float(num) for num in l ],Float))
        except ValueError:
            print "Could not parse ",l
            raise ValueError
    return array(x,Float)

# ----------------------------------------------------------------------

def linRegr(x,y):
    """ Returns a,b so that ax+b=y best in the LS sense """
    n=x.shape[0] ; xs=add.reduce(x) ; ys=add.reduce(y) ;
    xys=add.reduce(x*y) ; x2s=add.reduce(x*x) ;
    a=(n*xys-ys*xs)/float(n*x2s-xs*xs)
    b=float(ys-a*xs)/n
    return a,b

# ----------------------------------------------------------------------

def quadRegr(x,y):
    """ Returns a,b so that ax^2+bx+c=y best in the LS sense """
    Ft=array((x**2,x,x*0+1))
    FFt=matrixmultiply(Ft,transpose(Ft))
    Fty=matrixmultiply(Ft,y)
    z=LinearAlgebra.solve_linear_equations(FFt,Fty)
    return z[0],z[1],z[2]

# ----------------------------------------------------------------------

def cond(a):
    """ Returns the condition number of   a matrix a.
    The method is crude, as SVD in LinearAlgebra seems to be broken """
    inva=LinearAlgebra.inverse(a)
    return sumall(inva**2)/sumall(a**2)

# ----------------------------------------------------------------------

# ----------------------------------------------------------------------

def PythonFileToAscci(in_file,out_file):
    """Given a python shelve file it converts it into ASCII and writes it in
       a new file out_file"""
    
    fn=shelve.open(in_file,"r")
    list= fn.keys()
    #list_var= list 
    for i in xrange(len(list)):
        var=fn[list[i]]
        if type(var) == ArrayType:
            varstring=array2string(var)
        else:
            varstring=str(var)
        fout=open(out_file,"w")
        fout.write(varstring)
        fout.close()
        

def readpointsfromfile(fn):  # filename -> array
    """ Reads an ASCII file with a header and converts the rest
        to a Numerical Python 2d array"""
    fd=open(fn) 
    s=fd.read()                 # string con el contenido
    s1=re.sub('\n',' ',s)       # converts \n -> <space>
    s2=re.sub('^.*?\\[','[',s1) # skip the header
    l=eval(s2)                  # convert to list
    a=array(l,Float)            # convert to an array (can be Int)
    return a

def long_ax_projection(pts, v, v1):
    """This function gets a set of displacements (2xnpoints) as
    an array and computes the projection of those vector x,y
    onto v and the perpendicular vector to v. v (UNITARY VECTOR) defines
    the long axis of the LV, v1 is perpendicular to v. Output represents
    the logitudinal projection on [0,i] and the axial on [1,i] projection""" 
    npoints=pts.shape[0]
    pts_new=zeros(pts.shape,Float)
    for i in range(npoints):
        u=pts[i,:]
        if i==0:
            print u
            print v
        pts_new[i,0]=innerproduct(u,v)
        pts_new[i,1]=innerproduct(u,v1)#norm(innerproduct(u,v)*v-u) Esto poda ser pero no tena signo
        if i==0:
            print u
            print pts_new[i,:]
    return pts_new
    
#def norm(v):
#    """Calculates the norm of a vector"""
#    return sqrt(v[0]*v[0]+v[1]*v[1])


def norm_displ(pts):
    """Returns the absolute value of the displacement for the
    given vector"""
    pts_new=zeros(pts.shape[0])
    npoints=pts.shape[0]
    for i in range(npoints):
        pts_new[i]=norm(pts[i,:])

    return pts_new

def long_displ(pts):
    """Returns the longitudinal displacement from the given"""
    pts_new=zeros(pts.shape[0])
    npoints=pts.shape[0]
    for i in range(npoints):
        pts_new[i]=pts[i,0]

    return pts_new

def ax_displ(pts):
    """Returns axial displacement from the given vector"""
    pts_new=zeros(pts.shape[0])
    npoints=pts.shape[0]
    for i in range(npoints):
        pts_new[i]=pts[i,1]

    return pts_new


def unitary_vector(pt1,pt2):
    """Returns the unitary vector defined by
    the two points pt1,pt2"""
    v=array([pt1[0]-pt2[0],pt1[1]-pt2[1]],'float')
    
    return v/norm(v)

# -----------------------------------------------------------------

def lgammln(xx):
    """
Returns the logarithm gamma function of xx.
    Gamma(z) = Integral(0,infinity) of t^(z-1)exp(-t) dt.
(Adapted from: Numerical Recipies in C.)

   taken from scipy/stat/stat.py

Usage:   lgammln(xx)
"""

    coeff = [76.18009173, -86.50532033, 24.01409822, -1.231739516,
	     0.120858003e-2, -0.536382e-5]
    x = xx - 1.0
    tmp = x + 5.5
    tmp = tmp - (x+0.5)*math.log(tmp)
    ser = 1.0
    for j in range(len(coeff)):
	x = x + 1
	ser = ser + coeff[j]/x
    return -tmp + math.log(2.50662827465*ser)

def gamma(x):
    """ Returns the gamma function of x """
    return exp(lgammln(x))


# -----------------------------------------------------------------

def shuffle(x):
    """ Randomly permute array x along the first component """
    n=x.shape[0]
    return take(x,RandomArray.permutation(n))

# -----------------------------------------------------------------

def randomexp(rg,fl,n):
    """ Generate n-times random data using a generator rg and apply
    scalar functions in list fl to it. Return array [(mean,stdev)]
    of the results """
    m=len(fl)
    res=zeros((n,m),Float)
    for i in xrange(n):
        x=rg()
        for j in xrange(m):
            res[i,j]=(fl[j])(x)
    stats=zeros((m,2),Float)        
    for j in xrange(m):
        stats[j,0]=mean(res[:,j])
        stats[j,1]=sample_std(res[:,j])
    return stats

# -----------------------------------------------------------------

def sample_std(x):
    """ Sample standard deviation for a 1D vector x """
    return sqrt(norm2m(x-mean(x))/nelems(x))

# -----------------------------------------------------------------

def correlation_matrix(x):
    """ Correlation matrix for x[n,d] """
    d=x.shape[1]
    c=zeros((d,d),Float)    
    for i in xrange(d): # calculate the correlation matrix
        for j in xrange(i,d):
            corr=dot(x[:,i],x[:,j])
            c[i,j]=corr ; c[j,i]=corr
    return c

# ------------------------------------------------------------------

def timefun(f):
    """ Measure the execution time of a result-less function f.
        Repeat the measurement several times if necessary for precision """
    def rawtimefun(f):
        t1=time.clock()
        apply(f)
        t2=time.clock()
        return t2-t1
    def postimefun(f):
        t=-1.0
        while t<0:
            t=rawtimefun(f)
        return t
    def reptimefun(f,reps):
        t=0.0
        for i in xrange(reps):
            t+=postimefun(f)
        # print "reps=",reps, "t=",t
        return t
    t=0.0 ; reps=1
    while t<0.1: # measurement time should be longer than 0.1s
        t=reptimefun(f,reps)
        reps*=10
    return t/float(reps/10)
