#!/usr/bin/env python

""" This script provides a command line interface to my registration
    algorithm implemented essentially in bigfelreg.py

    Invocation:

    jkregister.py [options] <refimgname> <testimgname> 

    To see a help text, call jkregister without arguments

    

    Jan Kybic, July 2001.  Jan.Kybic@epfl.ch
    $Id: jkregister.py,v 1.6 2003/11/27 14:51:38 kybic Exp $
"""

import getopt,sys
import bigtools
import bigfelreg
import Image
import bigoptimize

from Numeric import *

usage_info="""
This script provides a command line interface to my registration
algorithm.

Usage:

    jkregister.py [options] <refimg> <testimg> 

It finds a deformation deforming <testimg> to be as close as possible to
<refimg>. Options can be:

    --degc n       model spline degree (0-3,default 2)
    --degf n       image spline degree (0-3,default 3)
    --verbose n    verbosity           (0-10, default 6)
    --xtol eps     tolerance           (default 0.1)
    --crate (x,y)  pixels per knot     (default (64,64))
    --warped fn    saves warped testimg to file fn
    --wgrid  fn    saves warped grid image to file fn
    --difb   fn    saves difference before to file fn
    --difa   fn    saves difference before to file fn
    --ralpha  a     use a `Laplacian' warping regularizer with a (>0)
    --deff   fn    saves the deformation field to file fn (in ASCII)
"""

def main(argv):
    optlist=["xtol=","crate=","degc=","degf=","verbose=",
             "warped=","wgrid=","difa=","difb=","ralpha=","deff=" ]
    try:
        opts,args=getopt.getopt(argv,"",optlist)
    except getopt.GetoptError:
        print usage_info
        return
    par=bigtools.Parameters({'verbose':6,'xtol':0.1,'degf':3,
                             'degc':2,'crate':(64,64)})
    otheropts={}
    for o,a in opts:
        if o[:2]=='--':
            o=o[2:]
            if o in ('degc', 'degf', 'verbose','xtol','crate','ralpha'):
                par=bigtools.Parameters(par,override={o:eval(a,{},{})})
            elif o in ('warped','wgrid','difb','difa','deff'):
                otheropts[o]=a
        else:
            raise "Unhandled option"

    if len(args)!=2:
        print usage_info
        return
        
    try:
        refimg=bigtools.ImageToArray(Image.open(args[0]))
        testimg=bigtools.ImageToArray(Image.open(args[1]))
    except:
        print "Could not read images: ",sys.exc_info()[:2]
        return

    if otheropts.has_key('difb'):
        difimg=bigtools.showDif(refimg,testimg)
        bigtools.printtofile(difimg,otheropts['difb'])
    par=bigtools.Parameters(par,override={'refimg':refimg,'testimg':testimg})
    if par.isdef('ralpha'):
        wp=bigfelreg.MultigridableWarpingProblemRegularized(par=par)
    else:
        wp=bigfelreg.MultigridableWarpingProblemBase(par=par)
    mrs=bigoptimize.MROptimizerState(wp,par=par)
    print "Registration started"
    mrs.solveMR()
    print "Registration finished"
    x=mrs.getProblem().getXstate() # get the deformation parameters
    if (otheropts.has_key('warped') or otheropts.has_key('difa') or
        otheropts.has_key('deff')):
        wimg=bigfelreg.WarpableImageS(testimg,par=par).setDefS(x)
        testw=wimg.getWarped()
    if otheropts.has_key('warped') :
        bigtools.printtofile(testw,otheropts['warped'])
    if otheropts.has_key('wgrid'):
        grid=bigtools.makeGrid(testimg.shape)
        gridw=bigfelreg.WarpableImageS(grid,par=par).setDefS(x).getWarped()
        bigtools.printtofile(gridw,otheropts['wgrid'])
    if otheropts.has_key('deff'):
        df=wimg.getDeformation()
        df=reshape(df,(df.shape[1]*df.shape[0],df.shape[2]))
        bigtools.saveForMatlab(df,otheropts['deff'])
    if otheropts.has_key('difa'):
        difimg=bigtools.showDif(refimg,testw)
        bigtools.printtofile(difimg,otheropts['difa'],clip=1)
    print "jkregister finished"

# ----------------------------------------------------------------------
    
if __name__ == "__main__":  # running interactively
    main(sys.argv[1:])
