""" Module assembles code to work with and register
    the ultrasound images of MJL

    Maria J. Ledesma-Carbayo, January 2001
    $Id: sequence_model.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
    """

from Numeric import *
from RandomArray import *
import math
import Image
import shelve
import Gnuplot
import bigtools
import bigpyramid
import bigregister
import bigmultigrid
import regultrasound

# ------------------------------------------------------------------------
im_model_dir="images_chus/2C-Model/results-grid6x6/"
rootfilename=im_model_dir+"2C-model-"
original_image=im_model_dir+"2C-model-0.tiff"
nimages=22
x0=195 #x coordinate that defines aproximatly the midial axis.
x1=151 #x coordinate that defines the maximum relax distance of the epicardium
       #  to the midial axis
sizeim=(288,384)
# ------------------------------------------------------------------------



# ------------------------------------------------------------------------
def Model2CDeformationAtPoint(point,time, T=1.0,Maxdespx=6.0,Maxdespy=8.0):
    """ Evaluates the simulated deformation with point with time
    """
    #T=1.0 #The period is 1 second. t should be provided also in senconds
    #Maxdespx=6.0
    #Maxdespy=8.0
    PI=math.pi
    desp=zeros(2,Float)
    x_center=float(x0)
    x_wall=float(x1)
    #IMPORTANT the Deformation table that we are creating gives the
    #coordinate transformation to be done to find the
    #coordinates in the original image to get the new pixel values. If
    #we want to apply the transformation to the original image
    #coordinates the sign of the transformation that we want to impose 
    #should be changed.
    desp[1]=-1.0*Maxdespx*sin(time*PI/T)**2*sin(((x_center-point[1])/abs(x_wall-x_center))*PI/2.0)
    desp[0]=-1.0*-Maxdespy*sin(time*PI/T)**2
    
    
    return desp# turn upside down

# ------------------------------------------------------------------------

def test_Model2CDeformationAtPoint():
    """ Tests Model 2CDeformationAtPoint
    """
    T=1.0 #The period is 1 second.

    point=asarray((147,151), Float)
    pointdef=zeros((nimages,2),Float)
    for i in range(nimages):
       time=i*T/nimages
       pointdef[i]=Model2CDeformationAtPoint(point,time)


    myplotcum_x=bigtools.plot(pointdef[:,1],title='X Cumulative Deformation')
    myplotcum_y=bigtools.plot(pointdef[:,0],title='Y Cumulative Deformation')
    #raw_input('Press any key')
  
# ------------------------------------------------------------------------

def Model2CDeformationAtPointThroughtime(point, T=1.0, num_images=nimages):
    """ Makes 2CDeformationAtPoint at point for every time and returns
    the displacement vector in the direction from image test coordinates
    to image ref coordinates. IMPORTANT the sign is opposite to the deformation
    table we apply to the algorithm to creat the model. 
    """
    #T=1.0 #The period is 1 second.
    #point=asarray((147,151), Float)
    #point=asarray((147,151), Float)
    pointdef=zeros((num_images,2),Float)
    for i in range(num_images):
       time=i*T/(num_images-1)
       #IMPORTANT the Deformation table that we are creating gives the
       #coordinate transformation to be done to find the
       #coordinates in the original image to get the new pixel values. If
       #we want to apply the transformation to the original image
       #coordinates the sign of the transformaiton that we want to impose 
       #should be changed.
       pointdef[i]=-1.0*Model2CDeformationAtPoint(point,time)

    return pointdef
  

# ------------------------------------------------------------------------

def Model2CDefTableAtTime(time):
    """ Creates the deformation table for a given time point.
    """
    deftable=zeros((2,sizeim[0],sizeim[1]),Float)
    desp=zeros(2, Float)
    point=zeros(2, Float)

    for y in range(sizeim[0]):
        for x in range(sizeim[1]):
            point=[y,x]
            desp=Model2CDeformationAtPoint(point,time)
            deftable[1,:,x]=desp[1]
            deftable[0,y,:]=desp[0]

    return deftable
        
# ------------------------------------------------------------------------

def test_Model2CDefTableAtTime():
   """Tests Model2CDefTableAtTime().
   """
   T=1.0 #the period is set to 1 second.
   time=5*T/nimages
   deftable=Model2CDefTableAtTime(time)
   bigtools.view(deftable[0])
   bigtools.view(deftable[1])
   
# ------------------------------------------------------------------------

def Model2C_CreateSequence():
    """Makes a simulated sequence applying to an image the deformation
    model.
    """
    T=1.0 #The period is 1 second.
    im1=Image.open(original_image)
    #im2=Image.open(imname_test)
    img=bigtools.ImageToArray(im1)
    wim=bigregister.WarpableImage(img)
    for i in range(nimages+1):
       time=i*T/nimages
       deftable=Model2CDefTableAtTime(time)
       deftable=deftable + indices(img.shape,'d')
       wim.setDeformation(deftable)
       imwarped=wim.getWarped()
       #imwarped=bigtools.ArrayToImage(imwarped)
       fileout=rootfilename+`i`+"_s.tiff"
       bigtools.printtofile(imwarped,fileout,scale=(0,255))
       #imwarped.save(fileout)
       print 'Finish round',i



# ------------------------------------------------------------------------

def Create_Image_RayleighNoise(sigma, filein, fileout,rayl=1,gauss=1):
    """Takes an image and adds multiplicative Rayleigh
    distributed Noise and Additive gaussian Noise.
    """
    #rootf=rootfilename
    #nim=nimages
    im1=Image.open(filein)
    img=asarray(bigtools.ImageToArray(im1),Float)
    #bigtools.view(img,scale=(0,255))
    seed(x=0,y=0)
    yuniform=uniform(0.0,0.99999,img.shape)
    rayleigh=sigma*sqrt(-2.0*log(1-yuniform))
    yuniform2=uniform(0.0,0.99999,img.shape)
    gaussnoise=sqrt(-2*log(yuniform))*cos(2*math.pi*yuniform2)
    #other models
    #newimage=img*rayleigh
    #newimage2=(img*rayleigh+img)/2.0
    if rayl:
     img=(sqrt(img)*rayleigh+img)
    if gauss:
     img=img+20*gaussnoise
    
    #newim=bigtools.ArrayToImage(newimage)
    bigtools.printtofile(img,fileout)
    
# ------------------------------------------------------------------------
def Model2C_CreateNoisySequence(rootfileout,rootfilein=rootfilename):
    """Makes a simulated sequence applying to an image the deformation
    model.
    """
    for i in range(nimages):
       fileim=rootfilein+`i`+"_s.tiff"
       fileimout=rootfileout+`i`+"_s.tiff"
       sigma=0.8
       Create_Image_RayleighNoise(sigma, filein=fileim, fileout=fileimout)
       print 'Finish round',i
       
# ------------------------------------------------------------------------

def Model2C_RegisterSequence(rootf=rootfilename,nim=nimages,cshape=(2,6,6), tol=1e-3, reverse=1, no_withfirst=0):
    """Register the  simulated sequence.
    """
    #rootf=rootfilename
    #nim=nimages
    if no_withfirst==0:
        regultrasound.register_sequence_with_first(rootf,nim,cshape, tol)

    regultrasound.register_sequence(rootf,nim,cshape,tol,reverse=reverse)



# ------------------------------------------------------------------------

def Model2C_DeformationSequence():
    """findes the deformation for a given point in the simulated sequece. From
    the found deftables. 
    """
    rootf=rootfilename
    nim=nimages
    point=asarray((150,147), Float)
    pointdef1=regultrasound.deformation_sequence(rootf,nim, point)
    pointdef2=regultrasound.deformation_sequence_with_first(rootf,nim,point)
    pointdef3=Model2CDeformationAtPointThroughtime(point)
    gx=Gnuplot.Gnuplot()
    gy=Gnuplot.Gnuplot()
    
    dx1=Gnuplot.Data(arange(nimages+1),pointdef1[:,0], with='lines',title='X Cumulative Deformation (ref. previous image)')
    dx2=Gnuplot.Data(arange(nimages+1),pointdef2[:,0], with='linespoints',title='X Cumulative Deformation (ref. first image)')
    dx3=Gnuplot.Data(arange(nimages+1),pointdef3[:,0], with='linespoints',title='Modeled Deformation')
    gx.replot(dx1,dx2,dx3)

    dy1=Gnuplot.Data(arange(nimages+1),pointdef1[:,1], with='lines',title='Y Cumulative Deformation (ref. previous image)')
    dy2=Gnuplot.Data(arange(nimages+1),pointdef2[:,1], with='linespoints',title='Y Cumulative Deformation (ref. first image)')
    dy3=Gnuplot.Data(arange(nimages+1),pointdef3[:,1], with='linespoints',title='Modeled Deformation')
    gy.replot(dy1,dy2, dy3)
     
    gx.hardcopy('gx-model.eps')
    gy.hardcopy('gy-model.eps')
    
    #raw_input('Press any key')

# ------------------------------------------------------------------------

def Model2C_DeformationError(points, nim=nimages, rootf=rootfilename):
    """Calculates the mean absolute error of the given points through the time.
    """
    #rootf=rootfilename
    nim=nimages
    point=zeros(2, Float)
    npoints=points.shape[0]
    error1=0.0
    error2=0.0
    
    for i in range(npoints):
       point=points[i]
       pointdef1=regultrasound.deformation_sequence(rootf,nim, point)
       pointdef2=regultrasound.deformation_sequence_with_first(rootf,nim,point)
       pointdef3=Model2CDeformationAtPointThroughtime(point)
       error1=error1+bigtools.sumall(abs(pointdef1-pointdef3)) 
       error2=error2+bigtools.sumall(abs(pointdef2-pointdef3))

    error1=error1/(npoints*nim)
    error2=error2/(npoints*nim)
    return error1, error2

# ------------------------------------------------------------------------

def Model2C_DeformationDrawings(points,nim=nimages, rootf=rootfilename):

    gx=Gnuplot.Gnuplot()
    gy=Gnuplot.Gnuplot()
    dxl=[]
    dyl=[]
    npoints=points.shape[0]
    f=open((rootf+'cumerr.txt'),'w')
    
    for i in range(npoints):
       point=points[i]
       pointdef1=regultrasound.deformation_sequence(rootf,nim, point)
       pointdef2=regultrasound.deformation_sequence_with_first(rootf,nim,point)
       pointdef3=Model2CDeformationAtPointThroughtime(point)
       
       dx1=Gnuplot.Data(arange(nim),pointdef1[:,0], with='lines',title=('P '+`i`+' Y Cum Def (ref. prev)'))
       dx2=Gnuplot.Data(arange(nim),pointdef2[:,0], with='linespoints',title='P '+`i`+' Y Cum Def (ref 1')
       dx3=Gnuplot.Data(arange(nim),pointdef3[:,0], with='linespoints',title='P '+`i`+' Y Model Def')           
       
       apply(gx.plot,(dx1,dx2,dx3))
       
       dy1=Gnuplot.Data(arange(nim),pointdef1[:,1], with='lines',title=('P '+`i`+' X Cum Def (ref. prev)'))
       dy2=Gnuplot.Data(arange(nim),pointdef2[:,1], with='linespoints',title='P '+`i`+' X Cum Def (ref 1)')
       dy3=Gnuplot.Data(arange(nim),pointdef3[:,1], with='linespoints',title='P '+`i`+' X Model Def')           
       
       apply(gy.plot,(dy1,dy2,dy3))
       
       gx.hardcopy((rootf+'P'+`i`+'gx-model.eps'))
       gy.hardcopy((rootf+'P'+`i`+'gy-model.eps'))
       
       #We make a displacement on the curves to display all at the same time.
       pointdef1=10*i+pointdef1
       pointdef2=10*i+pointdef2
       pointdef3=10*i+pointdef3
       
       dx1=Gnuplot.Data(arange(nim),pointdef1[:,0], with='lines',title=('P '+`i`+' Y Cum Def (ref. prev)'))
       dx2=Gnuplot.Data(arange(nim),pointdef2[:,0], with='linespoints',title='P '+`i`+' Y Cum Def (ref 1')
       dx3=Gnuplot.Data(arange(nim),pointdef3[:,0], with='linespoints',title='P '+`i`+' Y Model Def')           
       dxl+=[dx1,dx2,dx3]
       
       dy1=Gnuplot.Data(arange(nim),pointdef1[:,1], with='lines',title=('P '+`i`+' X Cum Def (ref. prev)'))
       dy2=Gnuplot.Data(arange(nim),pointdef2[:,1], with='linespoints',title='P '+`i`+' X Cum Def (ref 1)')
       dy3=Gnuplot.Data(arange(nim),pointdef3[:,1], with='linespoints',title='P '+`i`+' X Model Def')           
       dyl+=[dy1,dy2,dy3]
       
       f.write(('P '+`i`+' Cumulative error after one cycle (errx, erry) = ('+`abs(pointdef1[nim-1,1]-pointdef2[nim-1,1])`+','+\
               `abs(pointdef1[nim-1,0]-pointdef2[nim-1,0])`+')\n'))


    f.close() 
    gx('set size 1,8')
    gy('set size 1,8')
    apply(gx.plot,dxl)
    apply(gy.plot,dyl)

    gx.hardcopy(rootf + 'gx-model.eps')
    gy.hardcopy(rootf + 'gy-model.eps')
    
    #raw_input('Press any key')

# ------------------------------------------------------------------------
def test_Model2C_DeformationError(rootf=rootfilename):
    points=[[208,174],[189,162],[180,160],[162,157],[158,155],\
            [140,150],[130,151],[118,152],[109,153],[95,157],[78,177],\
            [78,186],[82,205],[93,212],[106,223],[126,239],[139,245],\
            [151,254],[173,261],[185,263]]
    points=array(points, Float)        
    errors=Model2C_DeformationError(points, rootf=rootf)
    #rootf=rootfilename
    f=open((rootf+'AbsErr.txt'),'w')
    f.write(('Mean absolute error of the given points through the time (MErrX, MerrY) = '+`errors[0]`+','+`errors[1]`))
    f.close()

# ------------------------------------------------------------------------
def test_Model2C_DeformationDrawings(rootf=rootfilename):
    points=[[208,174],[162,157],[130,151],[78,177],[93,212],[126,239]]
    points=array(points, Float)        
    Model2C_DeformationDrawings(points,nimages,rootf)


# ------------------------------------------------------------------------

def Model2C_ForwRevError(points, rootfforward, rootfreverse,nim=nimages):
    """Calculates the mean absolute error of the given points through the time.
    """
    #rootf=rootfilename
    nim=nimages
    point=zeros(2, Float)
    npoints=points.shape[0]
    error1=0.0
    error2=0.0
    
    for i in range(npoints):
       point=points[i]
       pointdef1=regultrasound.deformation_sequence(rootfforward,nim, point)
       pointdef2=regultrasound.deformation_sequence(rootfreverse,nim,point, reverse=1)
       pointdef3=Model2CDeformationAtPointThroughtime(point)
       error1=error1+bigtools.sumall(abs(pointdef1-pointdef3)) 
       error2=error2+bigtools.sumall(abs(pointdef2-pointdef3))

    error1=error1/(npoints*nim)
    error2=error2/(npoints*nim)
    return error1, error2

# ------------------------------------------------------------------------

def Model2C_ForwRevDrawings(points, rootfforward, rootfreverse,nim=nimages):

    gx=Gnuplot.Gnuplot()
    gy=Gnuplot.Gnuplot()
    dxl=[]
    dyl=[]
    npoints=points.shape[0]
    f=open((rootfreverse+'cumerr.txt'),'w')
    
    for i in range(npoints):
       point=points[i]
       pointdef1=regultrasound.deformation_sequence(rootfforward,nim, point)
       pointdef2=regultrasound.deformation_sequence(rootfreverse,nim,point,reverse=1)
       pointdef3=Model2CDeformationAtPointThroughtime(point)
       
       dx1=Gnuplot.Data(arange(nim),pointdef1[:,0], with='lines',title=('P '+`i`+' Y Cum Disp (Forw.)'))
       dx2=Gnuplot.Data(arange(nim),pointdef2[:,0], with='linespoints',title='P '+`i`+' Y Cum Disp (Rev.)')
       dx3=Gnuplot.Data(arange(nim),pointdef3[:,0], with='linespoints',title='P '+`i`+' Y Model Disp')           
       
       apply(gx.plot,(dx1,dx2,dx3))
       
       dy1=Gnuplot.Data(arange(nim),pointdef1[:,1], with='lines',title=('P '+`i`+' X Cum Disp (Forw.)'))
       dy2=Gnuplot.Data(arange(nim),pointdef2[:,1], with='linespoints',title='P '+`i`+' X Cum Disp (Rev.)')
       dy3=Gnuplot.Data(arange(nim),pointdef3[:,1], with='linespoints',title='P '+`i`+' X Model Disp')           
       
       apply(gy.plot,(dy1,dy2,dy3))
       
       gx.hardcopy((rootfreverse+'P'+`i`+'gx-model.eps'))
       gy.hardcopy((rootfreverse+'P'+`i`+'gy-model.eps'))
       
       #We make a displacement on the curves to display all at the same time.
       pointdef1=10*i+pointdef1
       pointdef2=10*i+pointdef2
       pointdef3=10*i+pointdef3
       
       dx1=Gnuplot.Data(arange(nim),pointdef1[:,0], with='lines',title=('P '+`i`+' Y Cum Disp. (Forw.)'))
       dx2=Gnuplot.Data(arange(nim),pointdef2[:,0], with='linespoints',title='P '+`i`+' Y Cum Disp. (Rev.)')
       dx3=Gnuplot.Data(arange(nim),pointdef3[:,0], with='linespoints',title='P '+`i`+' Y Model Disp.')           
       dxl+=[dx1,dx2,dx3]
       
       dy1=Gnuplot.Data(arange(nim),pointdef1[:,1], with='lines',title=('P '+`i`+' X Cum Def (Forw.)'))
       dy2=Gnuplot.Data(arange(nim),pointdef2[:,1], with='linespoints',title='P '+`i`+' X Cum Def (Rev.)')
       dy3=Gnuplot.Data(arange(nim),pointdef3[:,1], with='linespoints',title='P '+`i`+' X Model Disp.')           
       dyl+=[dy1,dy2,dy3]
       
       f.write(('P '+`i`+'Cum Forward Error (cycle) (errx, erry) = ('+`abs(pointdef1[nim-1,1]-pointdef2[nim-1,1])`+','+\
               `abs(pointdef1[nim-1,0]-pointdef2[nim-1,0])`+')\n'))
       f.write(('P '+`i`+' Cum Reverse Error (cycle) (errx, erry) = ('+`abs(pointdef1[0,1]-pointdef2[0,1])`+','+\
               `abs(pointdef1[0,0]-pointdef2[0,0])`+')\n'))

    f.close() 
    gx('set size 1,8')
    gy('set size 1,8')
    apply(gx.plot,dxl)
    apply(gy.plot,dyl)

    gx.hardcopy(rootfreverse + 'gx-model.eps')
    gy.hardcopy(rootfreverse + 'gy-model.eps')
    
    #raw_input('Press any key')

# ------------------------------------------------------------------------
def test_Model2C_ForwRevError(rootfforward,rootfreverse):
    points=[[208,174],[189,162],[180,160],[162,157],[158,155],\
            [140,150],[130,151],[118,152],[109,153],[95,157],[78,177],\
            [78,186],[82,205],[93,212],[106,223],[126,239],[139,245],\
            [151,254],[173,261],[185,263]]
    points=array(points, Float)        
    errors=Model2C_ForwRevError(points, rootfforward, rootfreverse)
    #rootf=rootfilename
    f=open((rootfreverse+'AbsErr.txt'),'w')
    f.write(('Mean absolute error of the given points through the time (MErrX, MerrY) = '+`errors[0]`+','+`errors[1]`))
    f.close()

# ------------------------------------------------------------------------
def test_Model2C_ForwRevDrawings(rootfforward,rootfreverse):
    points=[[208,174],[162,157],[130,151],[78,177],[93,212],[126,239]]
    points=array(points, Float)        
    Model2C_ForwRevDrawings(points,rootfforward, rootfreverse, nimages)
    
# ------------------------------------------------------------------------


#test_Model2CDeformationAtPoint()
#test_Model2CDefTableAtTime()
#Model2C_CreateSequence()
#Model2C_DeformationSequence()

#imdir="images_chus/2C-Model-Noise/results-grid8x8/"
#rootf=imdir+"2C-model-"
#nim=nimages
#cshape=(2,8,8)
#tol=1e-3
#Model2C_RegisterSequence(rootf,nim,cshape,tol)
#test_Model2C_DeformationError(rootf)
#test_Model2C_DeformationDrawings(rootf)

#imdir="images_chus/2C-Model-Noise/results-grid12x12/"
#rootf=imdir+"2C-model-"
#nim=nimages
#cshape=(2,12,12)
#tol=1e-3
#Model2C_RegisterSequence(rootf,nim,cshape, tol)
#test_Model2C_DeformationError(rootf)#test_Model2C_DeformationDrawings(rootf)

#Creates Noisy Sequence
#rootfilein="images_chus/2C-Model/2C-model-"
#rootfileout="images_chus/2C-Model-Noise/2C-model-"
#Model2C_CreateNoisySequence(rootfileout,rootfilein=rootfilein)
#
#
#
#We do the reverse registration for the model with tol=1e-3 and
#grid=8x8.
#


#imdir_forward= "images_chus/2C-Model/results-grid8x8/"
#imdir_reverse= "images_chus/2C-Model/results-grid-8x8-reverse/"
#rootfforward= imdir_forward + "2C-model-"
#rootfreverse= imdir_reverse + "2C-model-"
#nim=nimages
#cshape=(2,8,8)
#tol=1e-3
#Model2C_RegisterSequence(rootfreverse,nim,cshape,tol,reverse=1)
#test_Model2C_ForwRevError(rootfforward, rootfreverse)
#test_Model2C_ForwRevDrawings(rootfforward, rootfreverse)

#raw_input('Press any key')

if 0:
    imdir_forward="images_chus/2C-Model-Noise/results-grid8x8/"
    imdir_reverse= "images_chus/2C-Model-Noise/results-grid-8x8-reverse/"
    rootfforward= imdir_forward + "2C-model-"
    rootfreverse= imdir_reverse + "2C-model-"
    nim=nimages
    cshape=(2,8,8)
    tol=1e-3
    Model2C_RegisterSequence(rootfreverse,nim,cshape,tol,reverse=1,no_withfirst=1)
    test_Model2C_ForwRevError(rootfforward, rootfreverse)
    test_Model2C_ForwRevDrawings(rootfforward, rootfreverse)


#__________________

















