""" Module showultrasound assembles routines
    to present the results of the registration
    of ultrasound images of Chus.

    Jan Kybic, February 2001
    $Id: showultrasound.py,v 1.2 2002/01/21 16:33:19 jkybic Exp $
"""

import Image
import os, glob, re
import traceback, sys
import time
import bigregister
import bigtools
from Numeric import *
import Gnuplot
import shelve
import regultrasound
import sequence_model

# ----------------------------------------------------------------------

class UnusedFileName:
    """ gives a new, unused name for a link or file
        To be called from the working directory. """

    def __init__(self,format="figure%04d.eps"):
        self.format=format
        self.cnt=0
        while os.path.exists(self.format % self.cnt):
            self.cnt+=1

    def getUnused(self):
        while os.path.exists(self.format % self.cnt):
            self.cnt+=1
        name=self.format % self.cnt
        self.cnt+=1
        return name

    def makeLink(self,fn):
        success=0
        while not success:
            name=self.getUnused()
            try:
                os.symlink(fn,name)
                success=1
            except:
                print "Could not symlink", fn, ' -> ',name
        return name

def sanitizeTeX(f):
    """ transforms path string f not to contain any illegal characters """
    f=re.compile('_').sub("\\_",f)
    f=re.compile('\^').sub("\\\^",f)
    return f

def createTeXfromDisplacementCurves(
    filename=None,
    workdir="tex",
    whichfiles="../images_chus/Im-2c-Clip0013-masked/results-*/*P*-real.eps",
    suffixes=("gx-real.eps","gy-real.eps")):
    """ First goes into workdir. Looks for all files
        corresponding to whichfiles and creates a TeX file
        with each page files corresponding to the two suffixes side-by-side.
        It writes the result into a filename.
        It also puts to workdir all the necessary links
    """

    cwd=os.getcwd()
    try:
        os.chdir(workdir)
        u=UnusedFileName(format="figure%04d.eps")
        f=open(filename,'w')
        # Write header
        f.write("""%% Generated automatically at %s
\\documentclass{article}
\\usepackage{epsfig,pstricks,moreverb}
\\begin{document}
    """ % time.strftime('%H:%M %d.%m.%Y',time.localtime(time.time())))

        pat=re.compile('.*'+suffixes[0])
        reppat=re.compile(suffixes[0]+'$')
        files=glob.glob(whichfiles)
        files.sort()
        for i in files:
            if pat.search(i): # i ends with suffixes[0]
                i2=reppat.sub(suffixes[1],i) # get the other filename
                if not os.path.exists(i2):
                    print "Warning: file %s does not exist." % i2
                    i2=i 
                # create links
                link1=u.getUnused() ; os.symlink(i,link1)
                link2=u.getUnused() ; os.symlink(i2,link2)
                # Make a page
                f.write("""
\\begin{center}
\\mbox{\\rotateright{\\epsfig{file=%s,height=0.7\\textwidth}}}\\\\
\\mbox{\\rotateright{\\epsfig{file=%s,height=0.7\\textwidth}}}\\\\
  %s and %s \\
\\end{center}
""" % (link1,link2,sanitizeTeX(i),sanitizeTeX(i2)))

        # Write tailer
        f.write("""
\\end{document}
%%%% End %%%%%
""")
    finally:
        os.chdir(cwd)

def createTeXfromDisplacementCurvesTable(
    filename=None,
    workdir="tex",
    filesformat="../images_chus/Im-2c-Clip0013-masked/results-grid%s/2C-Clip0013-768x576-P%d",
    suffixes=("gx-real.eps","gy-real.eps"),
    controlsizes=('6x6','8x8','10x10','12x12','14x14','16x16'),
    summaryfileformats=["../images_chus/Im-2c-Clip0013-masked/results-grid%s/2C-Clip0013-768x576-cumerr.txt"],
    pointnum=6):
    """ First goes into workdir. Looks for all files
        corresponding to whichfiles and creates a TeX file
        with each page files corresponding to the two suffixes side-by-side.
        It writes the result into a filename.
        It also puts to workdir all the necessary links
    """

    cwd=os.getcwd()
    try:
        os.chdir(workdir)
        u=UnusedFileName(format="figure%04d.eps")
        f=open(filename,'w')
        # Write header
        f.write("""%% Generated automatically at %s
\\documentclass{article}
\\usepackage{epsfig,pstricks,moreverb}
\\sloppy\\emergencystretch=10cm
{\\catcode`\\^^M=13 %% these lines must end with %%
  \\gdef\\obeylines{\\catcode`\\^^M13 \\let^^M\\par}%%
  \\global\\let^^M\\par} %% this is in case ^^M appears in a \\write

\\begin{document}
    """ % time.strftime('%H:%M %d.%m.%Y',time.localtime(time.time())))

        for pt in range(pointnum):
            for suf in suffixes:
                # Make a page
                f.write("""
Point %d, suf=%s

\\begin{center}
\\begin{tabular}{cc}

""" % (pt,suf))
                fnlist=[]
                figlist=[]
                for sz in controlsizes:
                        fnlist.append(filesformat % (sz,pt)+suf)
                for fn in fnlist:
                        if not os.path.exists(fn):
                               print "Warning: file %s does not exist." % fn
                        link1=u.makeLink(fn) 
                        figlist.append(link1)
                i=0
                while i<len(fnlist):
                    if i<len(fnlist)-1:
                        f.write("""
\\rotateright{\\epsfig{file=%s,height=0.45\\textwidth}} &
\\rotateright{\\epsfig{file=%s,height=0.45\\textwidth}} \\\\
%s & %s\\\\""" % (fnlist[i],fnlist[i+1],controlsizes[i],controlsizes[i+1]))
                        i+=2
                    else:
                        f.write("""
\\rotateright{\\epsfig{file=%s,height=0.45\\textwidth}} & \\\\
%s & \\\\""" % (fnlist[i],controlsizes[i]))
                        i+=1
                        
                f.write("""
\\end{tabular}
\end{center}
\clearpage
""")

        
        if summaryfileformats!=None:
            for sff in summaryfileformats:
                for sz in controlsizes:
                    sumfn=sff % sz
                    f.write("""
Control size %s

{\\small\\noindent\obeylines                    
\\input %s

}
                    """ % (sz,sumfn))
        # Write tailer
        f.write("""
\\end{document}
%%%% End %%%%%
""")
    finally:
        os.chdir(cwd)


# ----------------------------------------------------------------------

class RegultraImageSequence:
    """ represents a sequence of images as in regultrasound

        last image pair to be considered will be imcnt-1, imcnt
        where the last image imcnt is the same as image 0
    """

    def __init__(self,
                 basedir="images_chus/Im-2c-Clip0013-masked/results-grid8x8/",
                 imnameformat="2C-Clip0013-768x576-%d_s.tiff",
                 datnameformat="2C-Clip0013-768x576-%d-%d-coeff.dat",
                 imcnt=21,
                 imsize=None):
        self.basedir=basedir
        self.imnameformat=imnameformat
        self.datnameformat=datnameformat
        self.imcnt=imcnt
        if imsize==None: # Automatically determine imsize if not given
            self.imsize=Image.open(self.getImageName(0)).size
            self.imsize=(self.imsize[1],self.imsize[0])
            # print "imsize=",self.imsize
        else:
            self.imsize=imsize
        self.mask=None # mask to be filled when needed
        self.allc=None # all c coefficients
        self.forwallc=None # forward accumulated c coefficients
        self.defaultEvalfnc=self.getPairDefCoefs

            
    def getPointList(self):
        """ Returns the list of points we are interested in. """
        pts=[[208,174],[162,157],[130,151],[78,177],[93,212],[126,239]]
        return array(pts,Float)

    def getPointListNorb2C(self):
        """ Returns the list of points we are interested in. """
        """Equivalent points with getPointsListH2C """
        """(Basal Inferior wall, Medium Inferior wall, Basal Lateral Wall)"""
        pts=[[156,171],[131,169],[160,260]]
        return array(pts,Float)

    def getPointListH2C(self):
        """ Returns the list of points we are interested in. """
        """Equivalent points with getPointsListH2C """
        """(Basal Inferior wall, Medium Inferior wall, Basal Lateral Wall)"""
        pts=[[245,89],[183,98],[252,233]]
        return array(pts,Float)
    
    def getImageName(self,i):
        """ get the name of image i """
        return self.basedir+self.imnameformat % i

    def getImage(self,i):
        """ get the image i as matrix"""
        return self.getImageFromName(self.getImageName(i))

    def getImageFromName(self,fn):
        """ get the image in file fn as matrix"""
        try:
            return bigtools.ImageToArray(Image.open(fn))
        except:
            raise "Could not read image from file `%s'" % fn

    def getPairDefFileName(self,i):
        """ get the name of the file representing a deformation
        between images i and i+1 """
        return self.basedir+self.datnameformat % (i+1,i)

    def readPairDefCoefs(self,i):
        """ Reads from the file the coefficients c for deformation between
        images i and i+1 """
        return self.readCoefsFromName(self.getPairDefFileName(i))
        
    def readCoefsFromName(self,fn):
        try:
            f=shelve.open(fn)
            c=array(f['coeff']) # Make an explicit copy
            f.close()
            return c
        except:
            print traceback.print_exc(file=sys.stdout)
            raise "Could not read coefs from file `%s'" % fn
        

    def getPairDefCoefs(self,i):
        """ Returns the coefficients c for deformation between images
        i and i+1 """
        self.fetchCoefs()
        return self.allc[i]

    def fetchCoefs(self):
        """ Reads all the c coefficients into one big array self.allc """
        if self.allc==None:
            c=self.readPairDefCoefs(0) # read the first one to get shape
            self.allc=zeros((self.imcnt+1,)+c.shape,Float)
            self.allc[0]=c
            for i in range(1,self.imcnt):
                self.allc[i]=self.readPairDefCoefs(i)

    def calculateForwardAccumulatedDefCoefs(self):
        """TO CONSIDER: Take into account that the points move during the sequence  """

        if self.forwallc==None:
            self.fetchCoefs() # make sure self.allc exists
            self.forwallc=zeros(self.allc.shape,Float)
            for i in range(1,self.imcnt+1):
                self.forwallc[i]=self.forwallc[i-1]+self.allc[i]

    def getForwardAccumulatedDefCoefs(self,i):
        """ Return coefficients explaining a deformation from image 0
        to image i, calculated by forward accumulation """
        self.calculateForwardAccumulatedDefCoefs()
        return self.forwallc[i]

    def setDefaultEvalfnc(self,fnc):
        """ Default function to get the c values for getDefAtPoints """
        self.defaultEvalfnc=fnc

    def getDefAtPoints(self,i,points,evalfnc=None,flag=0):
        """ By default, evaluates the deformation between images i and i+1 at
            `points'. The deformation returned has already the sign
            reverted with respect to the deformation function used
            for warping.

            Optional parameter fnc is a method of this object, accepting
            the image index i as an argument and returns the corresponding c
            """
        wim=bigregister.WarpableImage(zeros(self.imsize),imageiscoefs=1)
        if evalfnc==None:
                evalfnc=self.defaultEvalfnc
        wim.setC(apply(evalfnc,[i]))
        displ=-wim.getDisplacementAtPoints(points,flag)
        return displ

    def showArrowsAtPointsWithDisplacement(self,i,points,displ):
        """ Makes a Gnuplot graphics with arrows representing vectors
            'displ' at 'points'. Returns a black&white image (matrix)
            with the arrows """
        g=Gnuplot.Gnuplot()
        d=zeros((points.shape[0],4),Float)
        d[:,:2]=points[:,::-1]
        d[:,2:]=reshape(displ[:,::-1],(points.shape[0],2))
        gd=Gnuplot.Data(d,with='vector lw 3') # linewidth
        g('set xrange [0:%d]' % self.imsize[1])
        g('set yrange [%d:0] ' % self.imsize[0])
        g('set noborder')
        g('set noxtics') ;  g('set noytics') ; 
        #g.plot(gd)
        g('set terminal gif size %d,%d' % (self.imsize[1],self.imsize[0]))
        tmpname=os.tmpnam()+'.gif'
        g('set output "%s"' % tmpname)
        g.plot(gd)
        del g # hopefully forces caches to flush, otherwise insert a delay
        imfig=bigtools.ImageToArray(Image.open(tmpname))
        os.unlink(tmpname)
        return imfig

    def showImageWithArrowsAtPointsWithDiplacement(self,i,points,displ,arrowcolor=(0,255,0)):
        """ Returns a [3,:,:] matrix with arrows defined by 'points'
        and 'displ' superimposed over image i """
        imarr=self.showArrowsAtPointsWithDisplacement(i,points,displ)
        imfig=bigtools.ImageToArray(Image.open(self.getImageName(i)))
        arrowcolor=array(arrowcolor,Float)
        imtot=zeros((3,)+imfig.shape,Float)
        for i in range(3):
            imtot[i]=choose(greater(imarr,0.5),(imfig,arrowcolor[i]))
        return imtot


    def showImageWithDeformation(self,i,points,scale=20.0,j=None):
        """ Returns a [3,:,:] matrix with arrows showing deformation
            returned by getDefAtPoints(i) image j (defaults to i)
        """
        if j==None:
            j=i
        return self.showImageWithArrowsAtPointsWithDiplacement(
            j,points,scale*self.getDefAtPoints(i,points))

    def getPointGrid(self,spacing=10):
        """ Creates a list of points, like getPointList,
        spaced regularly with 'spacing' pixels apart, in the whole image """
        npts=array(self.imsize,Int)/spacing-1 # how many points, integer division
        pttbl=indices(npts,'d')*spacing+spacing/2
        return transpose(reshape(pttbl,(2,-1)))

    def getMask(self,i=0,threshold=30,degs=5):
        """ returns a mask of where the object(heart) is, based on image i """
        if self.mask==None:
            img=self.getImage(i)
            self.mask=bigregister.calculateThresholdedMask(img,threshold,degs)
        return self.mask

    def resetMask(self):
        self.mask=None

    def getMaskedPointGrid(self,spacing=10):
        """ Calls getPointGrid and returns only arrows where the mask is
        """
        mask=self.getMask()
        pts=list(self.getPointGrid(spacing).astype(Int))
        pts=filter(lambda x,m=mask: m[x],pts)
        return array(pts)

    def getAllDisplacements(self,points):
        """ Returns a vector [imcnt,n,2] of displacements at points
            with shape [n,2] """

        displ=zeros((self.imcnt+1,)+points.shape,Float)
        for i in range(self.imcnt+1):
            displ[i]=self.getDefAtPoints(i,points)
        return displ

    def getStrainAtMask(self,i,evalfnc=None):
        """ Returns a matrix [2,imsize] with the strain values in percents
            for image i where mask zero elsewhere """
        wim=bigregister.WarpableImage(zeros(self.imsize),imageiscoefs=1,mask=self.getMask())
        if evalfnc==None:
                evalfnc=self.defaultEvalfnc
        wim.setC(apply(evalfnc,[i]))
        strain=zeros((2,)+self.imsize,Float)
        strain[0]=-wim.getDeformationDerivative(flag=1)[0]/self.imsize[0]
        strain[1]=-wim.getDeformationDerivative(flag=4)[1]/self.imsize[1]
        return strain
        
    def getStrainAtPointsCorreg(self,points):
        """ Returns the strain at points for all images in the sequence as a matrix
        [imcnt,n,2].
        Chus.
        NOTE: getAllDisplacements() does not take this movement into account
        either. Jan
        """
        s=zeros((self.imcnt,points.shape[0],2),Float)
        points=points.astype(Int)
        disp=self.getAllDisplacements(points)
        for i in range(self.imcnt):
            strain=self.getStrainAtMask(i,evalfnc=None)
            for j in range(points.shape[0]):
                s[i,j]=strain[:,int(points[j,0]+disp[i,j,0]),
                              int(points[j,1]+disp[i,j,1])]
        return s
    
    def getStrainAtPoints(self,points):
        """ Returns the strain at points for all images in the sequence as a matrix
        [imcnt,n,2].

        TO CONSIDER: Take into account that the points move during the sequence  
        """
        s=zeros((self.imcnt,points.shape[0],2),Float)
        points=points.astype(Int)
        for i in range(self.imcnt):
            strain=self.getStrainAtMask(i,evalfnc=None)
            for j in range(points.shape[0]):
                s[i,j]=strain[:,points[j,0],points[j,1]]
        return s
# ----------------------------------------------------------------------

class RegultraImageBidirSequence(RegultraImageSequence):
    """ Bidirectional subclass of RegultraImageSequence, capable of using
        the results from registering in both directions.
    """

    def __init__(self,
                 basedir="images_chus/Im-2c-Clip0013-masked/results-grid8x8/",
                 imnameformat="2C-Clip0013-768x576-%d_s.tiff",
                 datnameformat="2C-Clip0013-768x576-%d-%d-coeff.dat",
                 imcnt=21,
                 imsize=None,
                 basedirbackward=
                 "images_chus/Im-2c-Clip0013-masked/results-grid-8x8-reverse/",
):

        # call super (father's) init
        RegultraImageSequence.__init__(self,basedir=basedir,
        imnameformat=imnameformat,datnameformat=datnameformat,imcnt=imcnt,
        imsize=imsize)
        self.basedirbackward=basedirbackward
        self.allbc=None     # all c coefficients from the backward registration
        self.backwallc=None # backward accumulated c coefficients
        self.comballc=None  # combined accumulated c coefficients
        self.setDefaultEvalfnc(self.getCombinedAccumulatedDefCoefs)
        
    def getBackwPairDefFileName(self,i):
        """ get the name of the file representing a deformation
        between images i+1 and i """
        return self.basedirbackward+self.datnameformat % (i,i+1)

    def readBackwPairDefCoefs(self,i):
        """ Reads from the file the coefficients c for deformation between
        images i+1 and i """
        return self.readCoefsFromName(self.getBackwPairDefFileName(i))

    def fetchBackwCoefs(self):
        """ Reads all the backward c coefficients into one
        big array self.allbc, where allbc[i] is the deformation
        from i+1 to i """
        if self.allbc==None:
            c=self.readBackwPairDefCoefs(0) # read the first one to get shape
            self.allbc=zeros((self.imcnt+1,)+c.shape,Float)
            self.allbc[0]=c
            for i in range(1,self.imcnt):
                self.allbc[i]=self.readBackwPairDefCoefs(i)

    def getBackwPairDefCoefs(self,i):
        """ Returns the coefficients c for deformation between images
        i+1 and i """
        self.fetchBackwCoefs()
        return self.allbc[i]


    def calculateBackwardAccumulatedDefCoefs(self):
        """TO CONSIDER: Take into account that the points move during the sequence
        """

        if self.backwallc==None:
            self.fetchBackwCoefs() # make sure self.allbc exists
            self.backwallc=zeros(self.allc.shape,Float)
            for i in range(self.imcnt-1,-1,-1):
                self.backwallc[i]=self.backwallc[i+1]+self.allbc[i]

    def getBackwardAccumulatedDefCoefs(self,i):
        """ Return coefficients explaining a deformation from image 0
        to image i, calculated by backward accumulation """
        self.calculateBackwardAccumulatedDefCoefs()
        return self.backwallc[i]

    def calculateCombinedAccumulatedDefCoefs(self):
        if self.comballc==None:
            self.calculateForwardAccumulatedDefCoefs()  # make sure they exist
            self.calculateBackwardAccumulatedDefCoefs()
            self.comballc=zeros(self.forwallc.shape,Float)
            for i in range(self.imcnt+1):
                fweight=(self.imcnt-i)/float(self.imcnt)
                bweight=i/float(self.imcnt)
                self.comballc[i]=fweight*self.forwallc[i]+\
                                  bweight*self.backwallc[i]

    def getCombinedAccumulatedDefCoefs(self,i):
        """ Return coefficients explaining a deformation from image 0
        to image i, calculated by combining forward and backward
        accumulation """
        self.calculateCombinedAccumulatedDefCoefs()
        return self.comballc[i]
        
# ----------------------------------------------------------------------


def tryCreateTeX():
    #createTeXfromDisplacementCurvesTable('dispcur.tex')
    #createTeXfromDisplacementCurvesTable('dispcurmod.tex',
    #                                     filesformat="../images_chus/2C-Model/results-grid%s/2C-model-P%d",
    #                                    
    #                                   suffixes=("gx-model.eps","gy-model.eps"),
    #summaryfileformats=["../images_chus/2C-Model/results-grid%s/2C-model-cumerr.txt","../images_chus/2C-Model/results-grid%s/2C-model-AbsErr.txt"])

    #createTeXfromDisplacementCurvesTable('dispcurtol.tex',
    #                                     filesformat="../images_chus/Im-2c-Clip0013-masked/results-grid-%s-t1e-4/2C-Clip0013-768x576-P%d",
    #summaryfileformats=["../images_chus/Im-2c-Clip0013-masked/results-grid-%s-t1e-4/2C-Clip0013-768x576-cumerr.txt"],controlsizes=('8x8','12x12'))

    createTeXfromDisplacementCurvesTable('dispcurmodforwRevnoise.tex',
                                         filesformat="../images_chus/2C-Model-Noise/results-grid-%s-reverse/2C-model-P%d",
                                        controlsizes=(['8x8']),
                                       suffixes=("gx-model.eps","gy-model.eps"),
    summaryfileformats=["../images_chus/2C-Model-Noise/results-grid-%s-reverse/2C-model-cumerr.txt","../images_chus/2C-Model-Noise/results-grid-%s-reverse/2C-model-AbsErr.txt"])

    createTeXfromDisplacementCurvesTable('dispcurmodforwRev.tex',
                                         filesformat="../images_chus/2C-Model/results-grid-%s-reverse/2C-model-P%d",
                                        controlsizes=(['8x8']),
                                       suffixes=("gx-model.eps","gy-model.eps"),
    summaryfileformats=["../images_chus/2C-Model/results-grid-%s-reverse/2C-model-cumerr.txt","../images_chus/2C-Model/results-grid-%s-reverse/2C-model-AbsErr.txt"])


def test_RegultraImageSequence():
#    ims=RegultraImageSequence(basedir="images_chus/2C-Model/results-grid8x8/",
#                 imnameformat="2C-model-%d_s.tiff",
#                 datnameformat="2C-model-%d-%d-coeff.dat",
#                 imcnt=21
# )
    ims=RegultraImageSequence()
    #d=ims.getDefAtPoints(8,ims.getPointList())
    #print d
    #g=ims.showArrowsAtPoints(8,ims.getPointList())
    #bigtools.view(ims.showImageWithArrows(8,ims.getPointList()),scale=(0,255))
    ll=[]
    for i in range(0,20,3):
        ll.append(bigtools.view(ims.showImageWithDeformation(i,
                                ims.getMaskedPointGrid()),
                      title='Image %d' % i))
    raw_input('Press any key')
    
def testStrain():
    #ims=RegultraImageSequence(basedir="images_chus/2C-Model/results-grid8x8/",
    #             imnameformat="2C-model-%d_s.tiff",
    #             datnameformat="2C-model-%d-%d-coeff.dat",
    #             imcnt=21)
    ims=RegultraImageBidirSequence()
    i=5
    strain=ims.getStrainAtMask(i)
    sc=(-0.02,0.020)
    v1=bigtools.view(strain[0],title='strain y',scale=sc)
    bigtools.printtofile(strain[0],filename='miccai2001/strain-r-dispx-5.tiff',
                         scale=sc)
    print "Strain Y max", bigtools.maxelement(strain[0])," min ",-bigtools.maxelement(-strain[0])
    v2=bigtools.view(strain[1],title='strain x',scale=sc)
    bigtools.printtofile(strain[1],filename='miccai2001/strain-r-dispy-5.tiff',
                         scale=sc)

    print "Strain X max", bigtools.maxelement(strain[1])," min ",-bigtools.maxelement(-strain[1])
    raw_input('Press any key')

def testStrain2():
    #ims=RegultraImageSequence(basedir="images_chus/2C-Model/results-grid8x8/",
    #             imnameformat="2C-model-%d_s.tiff",
    #             datnameformat="2C-model-%d-%d-coeff.dat",
    #             imcnt=21)
    #ims=RegultraImageBidirSequence()
    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8/",
                     imnameformat="Clip0007-768x576-%d_s.tiff",
                     datnameformat="Clip0007-768x576-%d-%d-coeff.dat",
                     imcnt=35,
                     basedirbackward=
                 "images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8-reverse/" )
    i=5
    strain=ims.getStrainAtMask(i)
    sc=(-0.03,0.03)
    v1=bigtools.view(strain[0],title='strain y',scale=sc)
    bigtools.printtofile(strain[0],filename='miccai2001/strain-n2C-dispx-5.tiff',
                         scale=sc)
    print "Strain Y max", bigtools.maxelement(strain[0])," min ",-bigtools.maxelement(-strain[0])
    v2=bigtools.view(strain[1],title='strain x',scale=sc)
    bigtools.printtofile(strain[1],filename='miccai2001/strain-n2C-dispy-5.tiff',
                         scale=sc)

    print "Strain X max", bigtools.maxelement(strain[1])," min ",-bigtools.maxelement(-strain[1])
    
    ims=RegultraImageBidirSequence(basedir="images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6/",
                     imnameformat="Clip0023-384x288-%d_s.tiff",
                     datnameformat="Clip0023-384x288-%d-%d-coeff.dat",
                     imcnt=9,
                     basedirbackward=
                 "images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6-reverse/" )
    i=5
    strain=ims.getStrainAtMask(i)
    sc=(-0.03,0.03)
    v3=bigtools.view(strain[0],title='strain y',scale=sc)
    bigtools.printtofile(strain[0],filename='miccai2001/strain-h2C-dispx-5.tiff',
                         scale=sc)
    print "Strain Y max", bigtools.maxelement(strain[0])," min ",-bigtools.maxelement(-strain[0])
    v4=bigtools.view(strain[1],title='strain x',scale=sc)
    bigtools.printtofile(strain[1],filename='miccai2001/strain-h2C-dispy-5.tiff',
                         scale=sc)

    print "Strain X max", bigtools.maxelement(strain[1])," min ",-bigtools.maxelement(-strain[1])
    raw_input('Press any key')
    
def testMasks():
    #ims=RegultraImageSequence()
    ims=RegultraImageBidirSequence()
    #img=bigtools.ImageToArray(Image.open(ims.getImageName(5)))
    #q1=bigtools.view(img)
    #mask=bigregister.calculateThresholdedMask(img)
    #q2=bigtools.view(mask)
    #print ims.getMaskedPointGrid()
    i=10 ; sc=2.0
    q1=bigtools.view(ims.showImageWithDeformation(i,ims.getMaskedPointGrid(),
                                                  scale=sc,j=0),
                      title='Image %d - combined' % i)
    ims.setDefaultEvalfnc(ims.getForwardAccumulatedDefCoefs)
    q2=bigtools.view(ims.showImageWithDeformation(i,ims.getMaskedPointGrid(),
                                                  scale=sc,j=0),
                      title='Image %d - forward' % i)
    ims.setDefaultEvalfnc(ims.getBackwardAccumulatedDefCoefs)
    q3=bigtools.view(ims.showImageWithDeformation(i,ims.getMaskedPointGrid(),
                                                  scale=sc,j=0),
                      title='Image %d - backward' % i)
    raw_input('Press any key')

def plotDeformation():
#    ims=RegultraImageSequence(basedir="images_chus/2C-Model/results-grid8x8/",
#                 imnameformat="2C-model-%d_s.tiff",
#                 datnameformat="2C-model-%d-%d-coeff.dat",
#                 imcnt=21,
# )
    ims=RegultraImageBidirSequence()
    ims.setDefaultEvalfnc(ims.getForwardAccumulatedDefCoefs)
    #ims.setDefaultEvalfnc(ims.getPairDefCoefs)
    displf=ims.getAllDisplacements(ims.getPointList())
    #print ims.imcnt,ims.allc.shape,ims.forwallc.shape,displf.shape
    ims.setDefaultEvalfnc(ims.getBackwardAccumulatedDefCoefs)
    #ims.setDefaultEvalfnc(ims.getBackwPairDefCoefs)
    displb=ims.getAllDisplacements(ims.getPointList())
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    displc=ims.getAllDisplacements(ims.getPointList())
    d1=Gnuplot.Data(arange(displf.shape[0]),reshape(displf[:,3,0],(-1,)),
                    with='lines',title='forward')
    d2=Gnuplot.Data(arange(displb.shape[0]),reshape(displb[:,3,0],(-1,)),
                    with='lines',title='backward')
    d3=Gnuplot.Data(arange(displc.shape[0]),reshape(displc[:,3,0],(-1,)),
                    with='lines',title='combined')
    g=Gnuplot.Gnuplot()
    g.plot(d1,d2,d3)
    raw_input('Press any key')

def plotStrain():
    #ims=RegultraImageBidirSequence(basedir="images_chus/2C-Model/results-grid8x8/",
    #                 imnameformat="2C-model-%d_s.tiff",
    #                 datnameformat="2C-model-%d-%d-coeff.dat",
    #                 imcnt=21,
    #                 basedirbackward=
    #             "images_chus/2C-Model/results-grid-8x8-reverse/" )
    ims=RegultraImageBidirSequence()
    pts=ims.getPointList()
    s=ims.getStrainAtPoints(pts)*100.0 # to get percents
    trange=arange(s.shape[0])*timeperframe

    d1=Gnuplot.Data(trange,reshape(s[:,4,1],(-1,)),
                    with='lines',title='P3 x')
    d2=Gnuplot.Data(trange,reshape(s[:,4,0],(-1,)),
                    with='lines',title='P3 y')
    g=Gnuplot.Gnuplot()
    g.plot(d1,d2)
    d3=Gnuplot.Data(trange,reshape(s[:,2,1],(-1,)),
                    with='lines',title='Axial')
    d4=Gnuplot.Data(trange,reshape(s[:,2,0],(-1,)),
                    with='lines',title='Longitudinal')
    g2=Gnuplot.Gnuplot()
    g2('set title "Langrangian strain"')
    g2.plot(d3,d4)
    g2.hardcopy("miccai2001/strain-curve-r-2.eps")
    d5=Gnuplot.Data(trange,reshape(s[:,1,1],(-1,)),
                    with='lines',title='P1 x')
    d6=Gnuplot.Data(trange,reshape(s[:,1,0],(-1,)),
                    with='lines',title='P1 y')
    g3=Gnuplot.Gnuplot()
    g3.plot(d5,d6)
    raw_input('Press any key')

# constants of our images, to be eventually moved elsewhere
timeperframe=0.040
mmperpixel=1.000


def makeFiguresForMICCAI():
    """ This function makes figures for the MICCAI article """

    # first the initialize the model sequence
    ims=RegultraImageBidirSequence(basedir="images_chus/2C-Model/results-grid8x8/",
                     imnameformat="2C-model-%d_s.tiff",
                     datnameformat="2C-model-%d-%d-coeff.dat",
                     imcnt=21,
                     basedirbackward=
                 "images_chus/2C-Model/results-grid-8x8-reverse/" )
    imnum=3 # show image 3
    ims.setDefaultEvalfnc(ims.getPairDefCoefs) # forward speed data
    ims.getMask(i=imnum) # set mask according to imnum
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(),scale=10.0)
    v1=bigtools.view(img,title='Model velocity at t=%f ms' %
                     (imnum*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-m-vel-%d.tiff' % imnum)
    # now the accumulated displacement at image 8 with respect to image 0
    imnumacc=8
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(),scale=4.0,j=0)
    v2=bigtools.view(img,title='Model displacement at t=%f ms, scale 4' %
                     (imnumacc*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-m-acc-%d.tiff' % imnumacc)

    # Now the same thing for the real images
    ims=RegultraImageBidirSequence()
    imnum=3 # show image 3
    ims.setDefaultEvalfnc(ims.getPairDefCoefs) # forward speed data
    ims.getMask(i=imnum) # set mask according to imnum
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(),scale=10.0)
    v3=bigtools.view(img,title='Velocity at t=%f ms' %
                     (imnum*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-r-vel-%d.tiff' % imnum)
    # now the accumulated displacement at image 8 with respect to image 0
    imnumacc=8
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(),scale=2.0,j=0)
    v4=bigtools.view(img,title='Displacement at t=%f ms, scale 2' %
                     (imnumacc*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-r-acc-%d.tiff' % imnumacc)
    raw_input('Press any key')
    
def makeFiguresForMICCAI2():
    """ This function makes figures of a
hypokinetic patient and healthy Norberto for the MICCAI article """

    # first the initialize the 2C-hipokinetic sequence
    ims=RegultraImageBidirSequence(basedir="images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6/",
                     imnameformat="Clip0023-384x288-%d_s.tiff",
                     datnameformat="Clip0023-384x288-%d-%d-coeff.dat",
                     imcnt=9,
                     basedirbackward=
                 "images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6-reverse/" )
    
    imnum=4 # show image 3
    ims.setDefaultEvalfnc(ims.getPairDefCoefs) # forward speed data
    ims.getMask(i=imnum, threshold=55) # set mask according to imnum
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=10),scale=5.0)
    v1=bigtools.view(img,title='Hypokinetic velocity at t=%f ms' %
                     (imnum*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-h2C-vel-%d.tiff' % imnum)
    # now the accumulated displacement at image 8 with respect to image 0
    imnumacc=9
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=10),scale=1.5,j=0)
    v2=bigtools.view(img,title='Hypokinetic displacement at t=%f ms, scale 1.5' %
                     (imnumacc*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-h2C-acc-%d.tiff' % imnumacc)

    # Now the same thing for the real images
    #ims=RegultraImageBidirSequence()
    #Normal 2C Norb
    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8/",
                     imnameformat="Clip0007-768x576-%d_s.tiff",
                     datnameformat="Clip0007-768x576-%d-%d-coeff.dat",
                     imcnt=35,
                     basedirbackward=
                 "images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8-reverse/" )
    imnum=5 # show image 5 (middle systole)
    ims.setDefaultEvalfnc(ims.getPairDefCoefs) # forward speed data
    ims.getMask(i=imnum, threshold=55) # set mask according to imnum
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=8),scale=5.0)
    v3=bigtools.view(img,title='Velocity at t=%f ms' %
                     (imnum*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-2Cn-vel-%d.tiff' % imnum)
    # now the accumulated displacement at image 11(end systole) with respect to image 0
    imnumacc=11
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=8),scale=1.5,j=0)
    v4=bigtools.view(img,title='Displacement at t=%f ms, scale 1.5' %
                     (imnumacc*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-2Cn-%d.tiff' % imnumacc)
    raw_input('Press any key')

def makeFiguresForMICCAI3():
    """ This function makes figures of Chus's heart for the MICCAI article """

    # first the initialize the sequence of Chus
    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Chus-4C-Clip0009/results-grid-8x8/",
                     imnameformat="Clip0009-768x576-%d_s.tiff",
                     datnameformat="Clip0009-768x576-%d-%d-coeff.dat",
                     imcnt=23,
                     basedirbackward=
                 "images_chus/Imn-Chus-4C-Clip0009/results-grid-8x8-reverse/" )
    
    imnum=5 # show image 3
    ims.setDefaultEvalfnc(ims.getPairDefCoefs) # forward speed data
    ims.getMask(i=imnum, threshold=20) # set mask according to imnum
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=10),scale=5.0)
    v1=bigtools.view(img,title='Model velocity at t=%f ms' %
                     (imnum*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-n4C-vel-%d.tiff' % imnum)
    # now the accumulated displacement at image 8 with respect to image 0
    imnumacc=11
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=10),scale=1.5,j=0)
    v2=bigtools.view(img,title='Model displacement at t=%f ms, scale 4' %
                     (imnumacc*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-n4C-acc-%d.tiff' % imnumacc)

    # Now the same thing for the real images
    #ims=RegultraImageBidirSequence()
    #Normal 2C Norb
    ims=RegultraImageBidirSequence()
    imnum=5# show image 5 (middle systole)
    ims.setDefaultEvalfnc(ims.getPairDefCoefs) # forward speed data
    ims.getMask(i=imnum, threshold=20) # set mask according to imnum
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=10),scale=5.0)
    v3=bigtools.view(img,title='Velocity at t=%f ms' %
                     (imnum*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-4Ch-vel-%d.tiff' % imnum)
    ## now the accumulated displacement at image 11(end systole) with respect to image 0
    imnumacc=10
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    img=ims.showImageWithDeformation(imnum,ims.getMaskedPointGrid(spacing=10),scale=1.5,j=0)
    v4=bigtools.view(img,title='Displacement at t=%f ms, scale 2' %
                     (imnumacc*timeperframe))
    bigtools.printtofile(img,'miccai2001/arrows-4Ch-%d.tiff' % imnumacc)
    raw_input('Press any key')

def makeGraphsForMICCAI():
    """ This function shows the movement curves for MICCAI.
        Code taken from Chus's sequence_model """
    # first get the point where to show it
    ims=RegultraImageBidirSequence() 
    pointnum=1
    point=ims.getPointList()[pointnum]
    # set paths for noiseless model
    imdir_forward= "images_chus/2C-Model/results-grid8x8/"
    imdir_reverse= "images_chus/2C-Model/results-grid-8x8-reverse/"
    rootfforward= imdir_forward + "2C-model-"
    rootfreverse= imdir_reverse + "2C-model-"
    nim=ims.imcnt+1
    pointdef1=regultrasound.deformation_sequence(rootfforward,nim, point)
    pointdef2=regultrasound.deformation_sequence(rootfreverse,nim,point,reverse=1)
    pointdef3=sequence_model.Model2CDeformationAtPointThroughtime(point)
    
    trange=arange(nim)*timeperframe
    dx1=Gnuplot.Data(trange,pointdef1[:,0], with='lines',title='forward')
    dx2=Gnuplot.Data(trange,pointdef2[:,0], with='lines',title='backward')
    dx3=Gnuplot.Data(trange,pointdef3[:,0], with='lines',title='true')
    gx=Gnuplot.Gnuplot()
    apply(gx.plot,(dx3,dx2,dx1))
    gx.hardcopy('miccai2001/curves-m-dispy-2.eps')

    # show x
    dx1=Gnuplot.Data(trange,pointdef1[:,1], with='lines',title='forward')
    dx2=Gnuplot.Data(trange,pointdef2[:,1], with='lines',title='backward')
    dx3=Gnuplot.Data(trange,pointdef3[:,1], with='lines',title='true')
    gx=Gnuplot.Gnuplot()
    gx('set title "Model, Axial displacement"')
    apply(gx.plot,(dx3,dx2,dx1))
    gx.hardcopy('miccai2001/curves-m-dispx-2.eps')

    # with noise
    imdir_forward= "images_chus/2C-Model-Noise/results-grid8x8/"
    imdir_reverse= "images_chus/2C-Model-Noise/results-grid-8x8-reverse/"
    rootfforward= imdir_forward + "2C-model-"
    rootfreverse= imdir_reverse + "2C-model-"
    nim=ims.imcnt+1
    pointdef1=regultrasound.deformation_sequence(rootfforward,nim, point)
    pointdef2=regultrasound.deformation_sequence(rootfreverse,nim,point,reverse=1)
    pointdef3=sequence_model.Model2CDeformationAtPointThroughtime(point)

    weights=arange(nim)/nim
    pointdef4=array(pointdef1)
    for i in range(nim):
                fweight=(nim-i)/float(nim)
                bweight=i/float(nim)
                pointdef4[i]=fweight*pointdef1[i]+bweight*pointdef2[i]

    #pointdef4=pointdef1*(1-weights[:,NewAxis])+pointdef2*weights[:,NewAxis]
    
    trange=arange(nim)*timeperframe
    dx1=Gnuplot.Data(trange,pointdef1[:,0], with='lines lw 1',title='forward')
    dx2=Gnuplot.Data(trange,pointdef2[:,0], with='lines lw 1',title='backward')
    dx3=Gnuplot.Data(trange,pointdef3[:,0], with='lines lw 10',title='true')
    dx4=Gnuplot.Data(trange,pointdef4[:,0], with='lines lw 10',title='combined')
    gy=Gnuplot.Gnuplot()
    gy('set title "Noisy model, Longitudinal displacement"')
    gy('set key bottom')
    gy('set noxtics')
    apply(gy.plot,(dx3,dx2,dx1,dx4))
    gy("set terminal postscript 20")
    gy.hardcopy('miccai2001/curves-mn-dispy-2.eps')

    # show x
    dx1=Gnuplot.Data(trange,pointdef1[:,1], with='lines lw 1',title='forward')
    dx2=Gnuplot.Data(trange,pointdef2[:,1], with='lines lw 1',title='backward')
    dx3=Gnuplot.Data(trange,pointdef3[:,1], with='lines lw 10',title='true')
    dx4=Gnuplot.Data(trange,pointdef4[:,1], with='lines lw 10',title='combined')

    gx=Gnuplot.Gnuplot()
    gx('set title "Noisy model, Axial displacement"')
    apply(gx.plot,(dx3,dx2,dx1,dx4))
    gx("set terminal postscript 20")
    gx('set noxtics')
    gx.hardcopy('miccai2001/curves-mn-dispx-2.eps')

    # now the real images
    imdir_forward= "images_chus/Im-2c-Clip0013-masked/results-grid8x8/"
    imdir_reverse= "images_chus/Im-2c-Clip0013-masked/results-grid-8x8-reverse/"
    rootfforward= imdir_forward + "2C-Clip0013-768x576-"
    rootfreverse= imdir_reverse + "2C-Clip0013-768x576-"
    nim=ims.imcnt+1
    pointdef1=regultrasound.deformation_sequence(rootfforward,nim, point)
    pointdef2=regultrasound.deformation_sequence(rootfreverse,nim,point,reverse=1)
    weights=arange(nim)/nim
    pointdef4=array(pointdef1)
    for i in range(nim):
                fweight=(nim-i)/float(nim)
                bweight=i/float(nim)
                pointdef4[i]=fweight*pointdef1[i]+bweight*pointdef2[i]

    #pointdef4=pointdef1*(1-weights[:,NewAxis])+pointdef2*weights[:,NewAxis]
    
    trange=arange(nim)*timeperframe
    dx1=Gnuplot.Data(trange,pointdef1[:,0], with='lines',title='forward')
    dx2=Gnuplot.Data(trange,pointdef2[:,0], with='lines',title='backward')
    dx4=Gnuplot.Data(trange,pointdef4[:,0], with='lines',title='combined')
    gy=Gnuplot.Gnuplot()
    gy('set title "Real data, Longitudinal displacement"')
    apply(gy.plot,(dx2,dx1,dx4))
    gy("set terminal postscript 20")
    gy.hardcopy('miccai2001/curves-r-dispy-2.eps')

    # show x
    dx1=Gnuplot.Data(trange,pointdef1[:,1], with='lines',title='forward')
    dx2=Gnuplot.Data(trange,pointdef2[:,1], with='lines',title='backward')
    dx4=Gnuplot.Data(trange,pointdef4[:,1], with='lines',title='combined')

    gx=Gnuplot.Gnuplot()
    gx('set title "Real data, Axial displacement"')
    apply(gx.plot,(dx2,dx1,dx4))
    gx("set terminal postscript 20")
    gx.hardcopy('miccai2001/curves-r-dispx-2.eps')
 
    raw_input('Press any key')



def makeGraphsForMICCAI2():
    """ This function shows the movement curves comparing and isquemic patient Imh-2C-... and normal  patient Imn-Norb-2C for MICCAI """
    
    pointnum=1 # point to consider
    # first the initialize the 2C-hipokinetic sequence
    ims=RegultraImageBidirSequence(basedir="images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6/",
                     imnameformat="Clip0023-384x288-%d_s.tiff",
                     datnameformat="Clip0023-384x288-%d-%d-coeff.dat",
                     imcnt=9,
                     basedirbackward=
                 "images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6-reverse/" )
    pixpermm=2.02 ; imsys=10
    # displacement from the combined forward/backward run
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    displc=ims.getAllDisplacements(ims.getPointListH2C())[:imsys]/pixpermm
    dy1=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,0],(-1,)),
                    with='lines lw 10',title='Ischemic')
    dx1=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,1],(-1,)),
                    with='lines lw 10',title='Ischemic')
    # healthy subject - Norberto
    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8/",
                     imnameformat="Clip0007-768x576-%d_s.tiff",
                     datnameformat="Clip0007-768x576-%d-%d-coeff.dat",
                     imcnt=35,
                     basedirbackward=
                 "images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8-reverse/" )

    pixpermm=1.03 ; imsys=11
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    displc=ims.getAllDisplacements(ims.getPointListNorb2C())[:imsys]/pixpermm
    dy2=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,0],(-1,)),
                    with='lines lw 10',title='Normal')
    dx2=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,1],(-1,)),
                    with='lines lw 10',title='Normal')

    # healthy subject - Chus
    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Chus-4C-Clip0009/results-grid-8x8/",
                     imnameformat="Clip0009-768x576-%d_s.tiff",
                     datnameformat="Clip0009-768x576-%d-%d-coeff.dat",
                     imcnt=23,
                     basedirbackward=
                 "images_chus/Imn-Chus-4C-Clip0009/results-grid-8x8-reverse/" )
    pixpermm=1.5 ; imsys=11
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    displc=ims.getAllDisplacements(ims.getPointList())[:imsys]/pixpermm
    dy3=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,0],(-1,)),
                    with='lines',title='Chus')
    dx3=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,1],(-1,)),
                    with='lines',title='Chus')

    # Plotting
    gy3=Gnuplot.Gnuplot()
    gy3('set title "Accumulated longitudinal displacement"')
    gy3('set ylabel "Displacement in mm"')
    gy3('set noxtics')
    gy3.plot(dy1,dy2)
    gy3("set terminal postscript 20")
    gy3.hardcopy('miccai2001/curves-r-HypoNorb-dispy-%d.eps' % pointnum)
    gx3=Gnuplot.Gnuplot()
    gx3('set title "Accumulated axial displacement"')
    gx3('set ylabel "Displacement in mm"')
    gx3('set noxtics')
    gx3.plot(dx1,dx2)
    gx3("set terminal postscript 20")
    gx3.hardcopy('miccai2001/curves-r-HypoNorb-dispx-%d.eps' % pointnum)

    raw_input('Press any key')

def plotStrain2():
    
    pointnum=0
    ims=RegultraImageBidirSequence(basedir="images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6/",
                     imnameformat="Clip0023-384x288-%d_s.tiff",
                     datnameformat="Clip0023-384x288-%d-%d-coeff.dat",
                     imcnt=9,
                     basedirbackward=
                 "images_chus/Imh-2C-Clip0023-384x288/results-grid-6x6-reverse/" )
    
    pixpermm=2.02 ; imsys=10
    # displacement from the combined forward/backward run
    #ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    #displc=ims.getAllDisplacements(ims.getPointListNorb2C())[:imsys]/pixpermm
    pts=ims.getPointListH2C()
    sh=(ims.getStrainAtPointsCorreg(pts)[:imsys])*100.0 # to get percents
    
    sxh=Gnuplot.Data(arange(sh.shape[0])/float(imsys),reshape(sh[:,pointnum,1],(-1,)), with='lines',title='Ischemic')
    syh=Gnuplot.Data(arange(sh.shape[0])/float(imsys),reshape(sh[:,pointnum,0],(-1,)), with='lines',title='Ischemic')

    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8/",
                     imnameformat="Clip0007-768x576-%d_s.tiff",
                     datnameformat="Clip0007-768x576-%d-%d-coeff.dat",
                     imcnt=35,
                     basedirbackward=
                 "images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8-reverse/" )

    pixpermm=1.03 ; imsys=11
    pts=ims.getPointListNorb2C()
    sn=(ims.getStrainAtPointsCorreg(pts)[:imsys])*100.0 # to get percents
    
    sxn=Gnuplot.Data(arange(sn.shape[0])/float(imsys),reshape(sn[:,pointnum,1],(-1,)), with='lines',title='Normal')
    syn=Gnuplot.Data(arange(sn.shape[0])/float(imsys),reshape(sn[:,pointnum,0],(-1,)), with='lines',title='Normal')

    g=Gnuplot.Gnuplot()
    g('set title "Langrangian Axial Strain"')
    g.plot(sxh,sxn)
    g.hardcopy("miccai2001/strain-curve-nh-axial.eps")
    
    g2=Gnuplot.Gnuplot()
    g2('set title "Langrangian Longitudinal Strain"')
    g2.plot(syh,syn)
    g2.hardcopy("miccai2001/strain-curve-nh-long.eps")
    
    raw_input('Press any key')






def tryDifPointNor():
    pointnum=1 # point to consider
    # healthy subject - Norberto
    ims=RegultraImageBidirSequence(basedir="images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8/",
                     imnameformat="Clip0007-768x576-%d_s.tiff",
                     datnameformat="Clip0007-768x576-%d-%d-coeff.dat",
                     imcnt=35,
                     basedirbackward=
                 "images_chus/Imn-Norb-2C-Clip0007/results-grid-8x8-reverse/" )

    pixpermm=1.03 ; imsys=11
    ims.setDefaultEvalfnc(ims.getCombinedAccumulatedDefCoefs)
    displc=ims.getAllDisplacements(ims.getPointListNorb2C())[:imsys]/pixpermm
    dy2=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,0],(-1,)),
                    with='lines',title='Norberto')
    dx2=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,1],(-1,)),
                    with='lines',title='Norberto')
    displc=ims.getAllDisplacements(ims.getPointListNorb2C()+
                                   reshape(array((-10,4),Float),(1,2)))[:imsys]/pixpermm
    dy1=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,0],(-1,)),
                    with='lines',title='Norberto2')
    dx1=Gnuplot.Data(arange(displc.shape[0])/float(imsys),
                    reshape(displc[:,pointnum,1],(-1,)),
                    with='lines',title='Norberto2')
    gy3=Gnuplot.Gnuplot()
    gy3('set title "Accumulated longitudinal displacement"')
    gy3('set xlabel "Time relative to systole length"')
    gy3('set ylabel "Displacement in mm"')
    gy3.plot(dy1,dy2)
    gx3=Gnuplot.Gnuplot()
    gx3('set title "Accumulated axial displacement"')
    gx3('set xlabel "Time relative to systole length"')
    gx3('set ylabel "Displacement in mm"')
    gx3.plot(dx1,dx2)
    raw_input('Press any key')

