/* Interpolation of a volume described by B-spline coefficients
   at arbitrary points.

   Jan Kybic, 1999
   $Id: splninterp.c,v 1.2 2002/01/21 16:33:19 jkybic Exp $
*/

#ifdef BIGSPLINES
#include "BIGsplines.h"
#include <math.h>
#endif


extern int mfoldmirroronbound(int k,int n)
/* having a signal 0..n-1, fold k using mirror on boundary conditions, 
   i.e. k=n-1 gives n-1, k=n gives n-2, k=-1 gives 1 etc. */
{ int m ;
  if (n<=1) return 0 ;
  m=2*(n-1) ;
  k=(k<0) ? k%m+m : k%m ;
  return k>=n ? m-k : k ;
}

extern double mfolddmirroronbound(double k,int n)
/* having a signal 0..n-1, fold k using mirror on boundary conditions, 
   i.e. k=n-1 gives n-1, k=n gives n-2, k=-1 gives 1 etc. */
{ int m,q ;
  m=2*(n-1) ;
  q=floor(k/m) ; k-=q*m ;
  return k>n-1 ? m-k : k ;
}

extern int mfoldmirroroffbound(int k,int n)
/* having a signal 0..n-1, fold k using mirror off boundary conditions, 
   i.e. k=n-1 gives n-1, k=n gives n-1, k=-1 gives 0 etc. */
{ int m ;
  if (n<=1) return 0 ;
  m=2*n ;
  k=(k<0) ? k%m+m : k%m ;
  return k>=n ? m-1-k : k ;
}

extern double mfolddmirroroffbound(double k,int n)
/* having a signal 0..n-1, fold k using mirror off boundary conditions, 
   i.e. k=n-1 gives n-1, k=n gives n-1, k=-1 gives 0 etc. */
{ int m,q ;
  m=2*n ;
  q=floor(k/m) ; k-=q*m ;
  return k>n-1 ? m-1-k : k ;
}

extern int evalbspln(double *x,double *y, int n, int degree)
/* Evaluates B-spline of degree 'degree' at n-points x[0],...,x[n-1] */
/* results are put into y[0],..,y[n-1] */
{ int i,supp ;
  double (*evsplnx)(double) ;    /* pointer to the evaluation functions */
  if (choosespln(degree,&evsplnx,&supp,0)) {
    myErrMsg("Unsupported degree.") ;
    return 1 ;
  }

  for (i=0;i<n;i++) y[i]=(*evsplnx)(x[i]) ;
  return 0 ;
}




extern int    splinterp
/* Takes an input matrix of size nxi*nyi*nzi, containing B-spline
    coefficients of degree `degree'. Samples the resulting function at
    nxc*nyc*nzc points given by coord, each point described by ndc 
    coordinates. Uses mirror boundary conditions. Everything should be 
    allocated in advance.
    In this routine, x is simply the index that changes fastest and 
    z the index which changes slowest. C convention is applied for indexing,
    i.e. the first element of `input' is assumed to correspond to 
    point (0,0,0). If coord gives multidimensional coordinates, they are laid
    consecutively, i.e., as the fastest changing (sub)index - even faster 
    than x.
*/
(
 double	*input,                 /* B-spline coefficients */
 double	*coord,		        /* 3D coordinates */
 double *output,                /* output values, one for each coord */ 
 int	nxi, int nyi,int nzi, 		/* input size */
 int    nxc,int nyc,int nzc,            /* output size */ 
 int    ndc,                    /* coordinate range dimensionality */
 int	degree,                 /* 0 - Haar, 1 - Linear etc. */
 int    dflag,                  /* 0 - value, 1,4,16 - first derivative with 
                                   respect to x,y,z ; 2,8,32 - second der. */
 int bcond 
 )
{ 
  int ix,iy,iz,ofs,jx,jy,jz,kz,lx,ly,lz ;
  double (*evsplnx)(double) ;    /* pointer to the evaluation functions */
  double (*evsplny)(double),(*evsplnz)(double) ;  
  int supp; double hsupp ;      /* spline support */
  double x,y,z,sum,sumt ;
  double *tabx,*taby,*ptr ;
  int *foldx,*foldy ;
  int (*mfold)(int,int) ; double (*mfoldd)(double,int) ;

  switch(bcond) {
      case MirrorOffBounds: 
	mfold=mfoldmirroroffbound ;
	mfoldd=mfolddmirroroffbound ;
	break ;
      case MirrorOnBounds: 
	mfold=mfoldmirroronbound ;
	mfoldd=mfolddmirroronbound ;
	break ;
  default:
        myErrMsg("Unsupported boundary conditions.") ;
	return 1 ;
  }

    

  //  mexPrintf("splninterp called with nxi=%d nyi=%d nzi=%d nxc=%d nyc=%d nzc=%d ndc=%d degree=%d\n", nxi, nyi, nzi, nxc, nyc, nzc, ndc, degree) ;
  if (choosespln(degree,&evsplnx,&supp,dflag & 3) | 
      choosespln(degree,&evsplny,NULL,(dflag & 12)>>2) | 
      choosespln(degree,&evsplnz,NULL,(dflag & 48)>>4)) {
    myErrMsg("Unsupported degree,dflag.") ;
    return 1 ;
  }
  hsupp=supp/2.0 ;

  if (ndc==2 && degree==0 && dflag==0) 
    { /* special case - 2D nearest neighbor */
      for (iy=0;iy<nyc;iy++)        /* loop through all elements of coord */
	for (ix=0;ix<nxc;ix++) {
	  ofs=ndc*(ix+nxc*iy) ;
	  x=coord[ofs] ; y=coord[ofs+1] ;
	  output[ix+nxc*iy]=
	    input[nxi*mfold(rint(y),nyi)+mfold(rint(x),nxi)] ;
	}
    } else { /* normal case */
	
      myMalloc(tabx,supp,1) 
      myMalloc(taby,supp,1) 
      myMallocInt(foldx,supp,1) 
      myMallocInt(foldy,supp,1) 

    for (iz=0;iz<nzc;iz++)        /* loop through all elements of coord */
      for (iy=0;iy<nyc;iy++)
	for (ix=0;ix<nxc;ix++) {
	  ofs=ndc*(ix+nxc*(iy+nyc*iz)) ;
	  x=coord[ofs] ;
	  y=ndc>1 ? coord[ofs+1] : 0.0 ;
	  z=ndc>2 ? coord[ofs+2] : 0.0 ;
	  lx=ceil(x-hsupp) ; ly=ceil(y-hsupp) ; lz=ceil(z-hsupp) ;
	  sum=0.0 ;
	  /* precalculate B-spline values and folded indexes */
	  for(jx=0;jx<supp;jx++) { tabx[jx]=(*evsplnx)(x-jx-lx) ; 
	                           foldx[jx]=mfold(jx+lx,nxi) ; }
	  if (nzi>1) { /* 3D case. 3rd dimension is nondegenerate */
	    for(jy=0;jy<supp;jy++) { taby[jy]=(*evsplny)(y-jy-ly) ;
	                             foldy[jy]=mfold(jy+ly,nyi) ; }
	    for(jz=lz;jz<lz+supp;jz++) { /* pass through all influential coefs */
	      kz=mfold(jz,nzi) ; 
	      for (jy=0;jy<supp;jy++) {
	        ptr=input+nxi*(foldy[jy]+nyi*kz) ;
		for (sumt=0.0,jx=0;jx<supp;jx++) sumt+=ptr[foldx[jx]]*tabx[jx] ;
		sum+=sumt*taby[jy]*(*evsplnz)(z-jz) ;
	      }
	    }
	  } else { /* 2D case */
	    for (jy=ly;jy<ly+supp;jy++) {
	      ptr=input+nxi*mfold(jy,nyi) ;
	      for (sumt=0.0,jx=0;jx<supp;jx++) sumt+=ptr[foldx[jx]]*tabx[jx] ;
	      sum+=sumt*(*evsplny)(y-jy) ;
	    } /* end of 2D case */
	  } /* if-else 2D case */
	  output[ix+nxc*(iy+nyc*iz)]=sum ;
	} /* for ix */

  myFree(taby) ; myFree(tabx) ; myFree(foldx) ; myFree(foldy) ;
    } /* if-else nearest neighbor */

  return 0 ;
}

/* --------------------------------------------------------------- */

extern int    splinterpmask
/* Takes an input matrix of size nxi*nyi*nzi, containing B-spline
    coefficients of degree `degree'. Samples the resulting function at
    nxc*nyc*nzc points given by coord, each point described by ndc 
    coordinates. Uses mirror boundary conditions. Everything should be 
    allocated in advance. Accepts a mask of the size coord, to
    determine, where to calculate.
    In this routine, x is simply the index that changes fastest and 
    z the index which changes slowest. C convention is applied for indexing,
    i.e. the first element of `input' is assumed to correspond to 
    point (0,0,0). If coord gives multidimensional coordinates, they are laid
    consecutively, i.e., as the fastest changing (sub)index - even faster 
    than x.
*/
(
 double	*input,                 /* B-spline coefficients */
 double	*coord,		        /* 3D coordinates */
 double *output,                /* output values, one for each coord */ 
 int	nxi, int nyi,int nzi, 		/* input size */
 int    nxc,int nyc,int nzc,            /* output size */ 
 int    ndc,                    /* coordinate range dimensionality */
 int	degree,                 /* 0 - Haar, 1 - Linear etc. */
 int    dflag,                   /* 0 - value, 1,4,16 - first derivative with 
                                   respect to x,y,z ; 2,8,32 - second der. */
 unsigned char *mask,
 int bcond 
 )
{ 
  int ix,iy,iz,ofs,jx,jy,jz,kz,lx,ly,lz ;
  double (*evsplnx)(double) ;    /* pointer to the evaluation functions */
  double (*evsplny)(double),(*evsplnz)(double) ;  
  int supp; double hsupp ;      /* spline support */
  double x,y,z,sum,sumt ;
  double *tabx,*taby,*ptr ;
  int *foldx,*foldy ;
  int (*mfold)(int,int) ; double (*mfoldd)(double,int) ;
  /* printf("splninterp called with nxi=%d nyi=%d nzi=%d nxc=%d nyc=%d nzc=%d ndc=%d degree=%d\n", nxi, nyi, nzi, nxc, nyc, nzc, ndc, degree) ;
   */


  switch(bcond) {
      case MirrorOffBounds: 
	mfold=mfoldmirroroffbound ;
	mfoldd=mfolddmirroroffbound ;
	break ;
      case MirrorOnBounds: 
	mfold=mfoldmirroronbound ;
	mfoldd=mfolddmirroronbound ;
	break ;
  default:
        myErrMsg("Unsupported boundary conditions.") ;
	return 1 ;
  }


  if (choosespln(degree,&evsplnx,&supp,dflag & 3) | 
      choosespln(degree,&evsplny,NULL,(dflag & 12)>>2) | 
      choosespln(degree,&evsplnz,NULL,(dflag & 48)>>4)) {
    myErrMsg("Unsupported degree,dflag.") ;
    return 1 ;
  }
  hsupp=supp/2.0 ;

  myMalloc(tabx,supp,1) 
  myMalloc(taby,supp,1) 
  myMallocInt(foldx,supp,1) 
  myMallocInt(foldy,supp,1) 

    for (iz=0;iz<nzc;iz++)        /* loop through all elements of coord */
      for (iy=0;iy<nyc;iy++)
	for (ix=0;ix<nxc;ix++) {
	  ofs=ndc*(ix+nxc*(iy+nyc*iz)) ;
	  /* printf("ix=%d iy=%d iz=%d mask=%d\n",ix,iy,iz,
	     (int)mask[ix+nxc*(iy+nyc*iz)]) ;*/
	  /* skip points with zero mask */
	  if (!mask[ix+nxc*(iy+nyc*iz)]) continue ; 
	  x=coord[ofs] ;
	  y=ndc>1 ? coord[ofs+1] : 0.0 ;
	  z=ndc>2 ? coord[ofs+2] : 0.0 ;
	  lx=ceil(x-hsupp) ; ly=ceil(y-hsupp) ; lz=ceil(z-hsupp) ;
	  sum=0.0 ;
	  /* precalculate B-spline values and folded indexes */
	  for(jx=0;jx<supp;jx++) { tabx[jx]=(*evsplnx)(x-jx-lx) ; 
	                           foldx[jx]=mfold(jx+lx,nxi) ; }
	  if (nzi>1) { /* 3D case. 3rd dimension is nondegenerate */
	    for(jy=0;jy<supp;jy++) { taby[jy]=(*evsplny)(y-jy-ly) ;
	                             foldy[jy]=mfold(jy+ly,nyi) ; }
	    for(jz=lz;jz<lz+supp;jz++) { /* pass through all influential coefs */
	      kz=mfold(jz,nzi) ; 
	      for (jy=0;jy<supp;jy++) {
	        ptr=input+nxi*(foldy[jy]+nyi*kz) ;
		for (sumt=0.0,jx=0;jx<supp;jx++) sumt+=ptr[foldx[jx]]*tabx[jx] ;
		sum+=sumt*taby[jy]*(*evsplnz)(z-jz) ;
	      }
	    }
	  } else { /* 2D case */
	    for (jy=ly;jy<ly+supp;jy++) {
	      ptr=input+nxi*mfold(jy,nyi) ;
	      for (sumt=0.0,jx=0;jx<supp;jx++) sumt+=ptr[foldx[jx]]*tabx[jx] ;
	      sum+=sumt*(*evsplny)(y-jy) ;
	    } /* end of 2D case */
	  } /* if-else 2D case */
	  output[ix+nxc*(iy+nyc*iz)]=sum ;
	} /* for ix */

  myFree(taby) ; myFree(tabx) ; myFree(foldx) ; myFree(foldy) ;

  return 0 ;
}

/* --------------------------------------------------------------- */

extern int    splinterpone
/* Samples a B-spline of degree `degree' centered at (bx,by,bz) at
    nxc*nyc*nzc points given by coord, each point described by ndc 
    coordinates. Uses mirror boundary conditions. Everything should be 
    allocated in advance.
    In this routine, x is simply the index that changes fastest and 
    z the index which changes slowest. C convention is applied for indexing,
    i.e. the first element of `input' is assumed to correspond to 
    point (0,0,0). If coord gives multidimensional coordinates, they are laid 
    consecutively, i.e., as the fastest changing (sub)index - even faster 
    than x.
*/
(
 int    bx, int by, int bz,     /* B-spline center */       
 double	*coord,		        /* 3D coordinates */
 double *output,                /* output values, one for each coord */ 
 int    nxi,int nyi,int nzi,    /* input matrix size */
 int    nxc,int nyc,int nzc,    /* output size */ 
 int    ndc,                    /* coordinate range dimensionality */
 int	degree,                 /* 0 - Haar, 1 - Linear etc. */
 int    dflag,                   /* 0 - value, 1,4,16 - first derivative with 
                                   respect to x,y,z ; 2,8,32 - second der. */
 int bcond
 )
{ 
  int ix,iy,iz,ofs ;
  double (*evsplnx)(double) ;    /* pointer to the evaluation functions */
  double (*evsplny)(double),(*evsplnz)(double) ;  
  int supp; double hsupp ;      /* spline support */
  double x,y,z,sum ;
  int (*mfold)(int,int) ; double (*mfoldd)(double,int) ;

  //  mexPrintf("splninterpone called with nxi=%d nyi=%d nzi=%d nxc=%d nyc=%d nzc=%d ndc=%d degree=%d\n", nxi, nyi, nzi, nxc, nyc, nzc, ndc, degree) ;


  switch(bcond) {
      case MirrorOffBounds: 
	mfold=mfoldmirroroffbound ;
	mfoldd=mfolddmirroroffbound ;
	break ;
      case MirrorOnBounds: 
	mfold=mfoldmirroronbound ;
	mfoldd=mfolddmirroronbound ;
	break ;
  default:
        myErrMsg("Unsupported boundary conditions.") ;
	return 1 ;
  }


  if (choosespln(degree,&evsplnx,&supp,dflag & 3) | 
      choosespln(degree,&evsplny,NULL,(dflag & 12)>>2) | 
      choosespln(degree,&evsplnz,NULL,(dflag & 48)>>4)) {
    myErrMsg("Unsupported degree,dflag.") ;
    return 1 ;
  }
  hsupp=supp/2.0 ;
  for (iz=0;iz<nzc;iz++)        /* loop through all elements of coord */
    for (iy=0;iy<nyc;iy++)
      for (ix=0;ix<nxc;ix++) {
	  sum=0.0 ;
	  ofs=ndc*(ix+nxc*(iy+nyc*iz)) ; 
	  x=coord[ofs] ; x=mfoldd(x-bx,nxi) ; 
       	  if (fabs(x)<hsupp) sum=(*evsplnx)(x) ; 
	  if (ndc>=2) {
	    y=coord[ofs+1] ; y=mfoldd(y-by,nyi) ; 
	    if (fabs(y)<hsupp) sum*=(*evsplny)(y) ; else sum=0.0 ;
	    if (ndc>=3) {
	      z=coord[ofs+2] ; z=mfoldd(z-bz,nzi) ; 
	      if (fabs(z)<hsupp) sum*=(*evsplnz)(z) ; else sum=0.0 ;
	    } /* if (ndc>=3) */
	  } /* if (ndc>=2) */
	  output[ix+nxc*(iy+nyc*iz)]=sum ;
	} /* for ix,iy,iz */
  return 0 ;
}
